/*
 * alloc.c - 
 */

/*
 * $Id: alloc.c,v 1.6 2014/10/28 02:22:55 komiya Exp $
 *
 *       Memory allocator (alloc.c)
 *    written by T.Komiya  Sep 27 1991
 */

static char *RCSid="$Id: alloc.c,v 1.6 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"

static HEADER base;
static HEADER *freep = NULL;

HEADER *morecore();

extern int  message_level;
extern int  memget_mode;
extern int  extend_heap();
extern unsigned a4, a5;
extern int sharp_eq_context_max;
extern SHARP_EQ_CONTEXT_STRUCT sharp_eq_context[];

int init_system_heap = 0;

#ifndef CTHREAD
#ifndef PVM3
#ifndef TK
#ifndef __CYGWIN32__

void    *malloc(nbytes)
size_t    nbytes;
{
    HEADER * p, *prevp;
    unsigned    nunits;

    if (nbytes == 0) return NULL;
    if (!init_system_heap) {
	return((char *)sbrk(nbytes));
    }
    nunits = (nbytes + sizeof(HEADER) - 1) / sizeof(HEADER) + 1;
    if ((prevp = freep) == NULL) {
        base.ptr = freep = prevp = &base;
        base.size = 0;
        free((char *)(sheap_start + 1));
        prevp = freep;
    }
    for (p = prevp->ptr; ; prevp = p, p = p->ptr) {
        if (p->size >= nunits) {
            if (p->size == nunits)
                prevp->ptr = p->ptr;
            else {
                p->size -= nunits;
                p += p->size;
                p->size = nunits;
            }
            freep = prevp;
            return((char *)(p + 1));
        }
        if (p == freep)
            if ((p = morecore(nunits)) == NULL) {
                print_con("morecore fail.\n");
                return(NULL);
            } 
            else
                print_con("morecore success.\n");
    }
}


#define NALLOC 1024

HEADER *morecore(nu)
unsigned    nu;
{
    HEADER * up;
    unsigned *old_fin;
    int i, dist, vacancy;
    extern unsigned *compaction();

    print_con("morecore...\n");
    if (memget_mode == COPY_GC) {
        copygc(0);
        if (entity_start != entity_base) {
            move_up_entity_area(entity_start, last_entity, entity_base);
            last_entity = last_entity - (entity_start - entity_base);
        }
        goto_mark_and_sweep();
    }
    vacancy = mark_and_sweep_gc();
    last_entity = (unsigned *)compaction(vacancy, 0);
    gctype = memget_mode = COMPACTION;
    for (i = 0; i < BLOCK_TABLE_SIZE; i++)
        block_table[i] = 0;
    big_block = 0;
    if (vacancy >= 4)
        *last_entity = LTAG_BLOCK2 | (vacancy - 4);
    if (vacancy > 4) {
        big_block = last_entity;
        *(last_entity + 1) = 0;
    }
    old_fin = entity_fin;
    if (big_block != 0)
        vacancy = *big_block;
    else
        vacancy = 0;
    extend_heap((int)(nu * sizeof(HEADER))); /* Feb 8, 1994 */
    dist = entity_fin - old_fin;
    if (vacancy != 0) {
        *big_block = vacancy;
        *(big_block + 1) = 0;
    } else
        big_block = 0;
    move_down_entity_area(entity_base, old_fin, entity_base + dist);
    if (big_block != 0)
        big_block += dist;
    up = (HEADER * )entity_base;
    nu = (dist * sizeof(int)) / sizeof(HEADER);
    up->size = nu;
    free((char *)(up + 1));
    entity_base += dist;
    entity_area_size = (char *)entity_end - (char *)entity_base;
    max_entity_area_size = (char *)entity_fin - (char *)entity_base;
    return(freep);
}


void free(ap_arg)
void    *ap_arg;
{
    HEADER * bp, *p;
    char *ap = ap_arg;

    if (!init_system_heap || ap == NULL)
        return;

    bp = (HEADER * )ap - 1;
    for (p = freep; !(bp > p && bp < p->ptr); p = p->ptr)
        if (p >= p->ptr && (bp > p || bp < p->ptr))
            break;

    if (bp + bp->size == p->ptr) {
        bp->size += (p->ptr)->size;
        bp->ptr = (p->ptr)->ptr;
    } else
        bp->ptr = p->ptr;
    if (p + p->size == bp) {
        p->size += bp->size;
        p->ptr = bp->ptr;
    } else
        p->ptr = bp;
    freep = p;
}


void    *realloc(oldptr_arg, size)
void    *oldptr_arg;
size_t    size;
{
    HEADER * bp;
    unsigned    nunits;
    char    *p;
    int i;
    char *oldptr = oldptr_arg;

#if defined(FREEBSD) || defined(LINUX)
    if (oldptr == NULL)
        return(malloc(size));
#endif

    bp = (HEADER * )oldptr - 1;
    nunits = (size + sizeof(HEADER) - 1) / sizeof(HEADER) + 1;
    if (bp->size == nunits)
        return(oldptr);
    else if (bp->size > nunits) {
        (bp + nunits)->size = bp->size - nunits;
        bp->size = nunits;
        free((char *)(bp + nunits + 1));
        return((char *)(bp + 1));
    } else {
        p = malloc(size);
        for (i = 0; i < ((bp->size - 1) * sizeof(HEADER)); i++)
            p[i] = oldptr[i];
        return(p);
    }
}


void    *calloc(nelem, elsize)
size_t    nelem, elsize;
{
    char    *ptr;
    int i;

    ptr = malloc(i = nelem * elsize);
    while (--i >= 0)
        ptr[i] = 0;
    return(ptr);
}


cfree(ptr)
char    *ptr;
{
    free(ptr);
}


move_down_entity_area(start, end, to)
unsigned start;
unsigned end;
unsigned to;
{
    int dist;
    int i;
    unsigned a2, a3, d3, d4;

    a2 = (unsigned)a2_val();
    a3 = (unsigned)active_val();
    a4 = (unsigned)base_val();
    a5 = (unsigned)top_val();
    d3 = (unsigned)reg_val();
    d4 = (unsigned)d4_val();
    dist = (int)-(to - start);
    if (message_level)
        print_con("moving down a entity area from high to low.\n");
    add_dis(&quotesym, dist);
    add_dis(&qquotesym, dist);
    add_dis(&atsym, dist);
    add_dis(&commasym, dist);

    /*
  add_dis(&loadsym, dist);
  add_dis(&dumpsym, dist);
  add_dis(&setsym, dist);
  add_dis(&endsym, dist);
  add_dis(&breakfunsym, dist);
  add_dis(&breaknumsym, dist);
  add_dis(&tracesym, dist);
  add_dis(&untracesym, dist);
  add_dis(&gctracesym, dist);
  add_dis(&ungctracesym, dist);
  add_dis(&pctracesym, dist);
  add_dis(&unpctracesym, dist);
  add_dis(&settracesym, dist);
  add_dis(&breakpointsym, dist);
  add_dis(&dumppointsym, dist);
  add_dis(&stacksym, dist);
*/

    add_dis(&definesym, dist);
    add_dis(&macrosym, dist);
    add_dis(&xreturnsym, dist);
    add_dis(&lreturnsym, dist);
    add_dis(&xcallsym, dist);
    add_dis(&c_f_sym, dist);
    add_dis(&mkncsym, dist);
    add_dis(&sctopsym, dist);
    add_dis(&errsym,   dist);		/* By N.Watanabe 92/11/19 */
    add_dis(&recomsym, dist);		/* By A243 */
    add_dis(&recomlevel, dist);		/* By A243 */
#ifdef PVM3
    add_dis(&bctran_exec_form_sym, dist);   /* By A243 */
    add_dis(&sharp_closure_pvm, dist);      /* By A243 */
    add_dis(&pvm_recv_function, dist);      /* By A243 */
    add_dis(&pvm_reduce_function, dist);      /* By A243 */
    add_dis(&pvm_get_unbound_obj, dist);      /* By A243 */
    add_dis(&pvm_put_unbound_obj, dist);      /* By A243 */
#endif
    add_dis(&debugsym, dist);
    add_dis(&user_exec_sym, dist);
    add_dis(&sys_debug_sym, dist);
    add_dis(&bcloadsym, dist);
    add_dis(&bc_trace_proc_sym, dist);
    add_dis(&sysfun_tracer_body, dist);
    add_dis(&max_back_trace_depth_sym, dist);
    add_dis(&gc_message_level_sym, dist);
    add_dis(&msgc_to_copygc_sym, dist);
    add_dis(&struct_sym, dist);
    add_dis(&l_t_e_sym, dist);
#ifdef PVM3
    add_dis(&pvm_l_t_e_sym, dist);    /* By A243 */
#endif
    add_dis(&local_fun_sym, dist);
    add_dis(&beginsym, dist);
    add_dis(&bcload_with_cip_sym, dist);

    add_dis(&unbound_obj, dist);
    add_dis(&eof_obj, dist);
    add_dis(&reg1, dist);
    add_dis(&reg2, dist);
    add_dis(&reg3, dist);
    add_dis(&stdin_port, dist);
    add_dis(&stdout_port, dist);
    add_dis(&c_in_port, dist);
    add_dis(&c_out_port, dist);
    add_dis(&bcconst, dist);
    add_dis(&xfunconst, dist);
    add_dis(&bcraw, dist);
    add_dis(&bcraw2, dist);
    add_dis(&bcbody, dist);
    add_dis(&acode, dist);
    add_dis(&tempp, dist);
    add_dis(&cdrbody, dist);
    add_dis(&bcf, dist);
    add_dis(&bcfhead, dist);
    add_dis(&call_cc_arg, dist);
    add_dis(&top_of_aflk, dist);
    add_dis(&end_of_aflk, dist);
    add_dis(&af_tmp, dist);
    add_dis(&more, dist);
#ifdef CTHREAD
    add_dis(&optcnt, dist);
#endif
    add_dis(&portlist, dist);
    add_dis(&ncclist, dist);
    add_dis(&wplist, dist);

#ifdef NIHONGO
    add_dis(&default_kanji_input_code_sym, dist);
    add_dis(&default_kanji_output_code_sym, dist);
#ifdef TCPIP
    add_dis(&default_network_kanji_input_code_sym, dist);
    add_dis(&default_network_kanji_output_code_sym, dist);
#endif
    add_dis(&euc_sym, dist);
    add_dis(&jis_sym, dist);
    add_dis(&sjis_sym, dist);
    add_dis(&noc_sym, dist);
#endif

#ifdef TCPIP
    add_dis (&conn_timeout, dist);
    add_dis (&tcp_timeout, dist);
    add_dis (&udp_timeout, dist);
    add_dis (&accept_timeout, dist);
    add_dis (&localhostname, dist);
#endif

    add_dis(&print_circle_sym, dist);	/* By ushijima */

    if (system_function_loading_flag == ON)
        add_dis(&protect_symbol_listsym, dist);

    for (i = 0; i < sharp_eq_context_max; i++) {
	add_dis(&(sharp_eq_context[i].sharp_index), dist);
	add_dis(&(sharp_eq_context[i].sharp_eq), dist);
	add_dis(&(sharp_eq_context[i].sharp_sharp), dist);
    }

    for (i = 0; i < hash_table_size; ++i)
        add_dis(hash_table + i, dist);
    for (i = 0; i < bchashsize; ++i)
        add_dis( bchash + i , dist);

#ifdef ASMGC
    add_dist_stack(dist);
#else
# ifdef CALLIOC
    {
        unsigned *sb;

        for (sb = active_stacks; sb != NULL; sb = (unsigned *)(sb[3]))
            add_dist_stack(dist, sb);
    }
# else
    add_dist_stack(dist, (unsigned *)a4, (unsigned *)a5);
# endif
#endif

    block_move_down((unsigned *)start, (unsigned *)end, (unsigned *)to, -dist);
    if (a2 != LTAG_NIL)
        a2 = a2 - dist;
    if (c_strbuf != strbuf)
        c_strbuf = c_strbuf - dist;
    set_a2(a2);
    add_dis(&d3, dist);
    set_active(a3);
    set_base(a4);
    set_top(a5);
    set_reg(d3);
    set_pc(d4);
}


block_move_down(start, end, to, dist)
unsigned *start, *end, *to;
int dist;
{
    int content, count, j;
    unsigned *i, *k;

    for (i = start; i < end; )
        switch (GETtag(*i)) {
        case LTAG_BLOCK2 :
            count = IMM(*i) >> 2;
            block2_ex(i, count);
            i += (count + 1);
            break;
        case LTAG_BLOCK1 :
            k = i + (IMM(*i) >> 2) + 1;
            while (++i < k)
                if (GETtag(*i) == LTAG_BLOCK2)
                    break;
            break;
        default :
            i += 2;
        }
    to += (end - start);
    while (start != end) {
        content = *--end;
        if (GETtag(content) == 0 || content & 0x80000000)
            *--to = content + dist;
        else {
            *--to = content;
            if (GETtag(content) == LTAG_BLOCK2) {
                j = count = IMM(content) >> 2;
                while (--j >= 0) 
                    *--to = *--end;
                block2_ex(to + count, -count);
            }
        }
    }
    wp_add_dist(dist);
}


block2_ex(p, q_size)
unsigned *p;
int q_size;
{
    int tmp;

    tmp = *p;
    *p = *(p + q_size);
    *(p + q_size) = tmp;
}

#endif /* CYGWIN */
#endif /* TK */
#endif /* PVM3 */
#endif /* CTHREAD */

wp_add_dist(dist)
int dist;
{
    unsigned *wpl, *wp_p;
    int content;

    wpl = (unsigned *)get_sval(wplist.obj);
    while ((unsigned)wpl != LTAG_NIL) {
        wp_p = (unsigned *)ADDRESS(wpl);
        content = *(wp_p + 3);
        if (GETtag(content) == 0 || content & 0x80000000)
            *(wp_p + 3) = content + dist;
        content = *(wp_p + 4);
        if (content != LTAG_NIL)
            *(wp_p + 4) = content + dist;
        wpl = (unsigned *) * (wp_p + 4);
    }
}


