/*
 * bci.c - 
 */

/*
 * $Id: bci.c,v 1.23 2014/10/28 02:22:55 komiya Exp $
 *
/*    byte code interpreter (C version)
 *        M.Yamamoto   1990.10
 *
 */

static char *RCSid="$Id: bci.c,v 1.23 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"

#ifdef THINK_C
#include <signal.h>
#endif

#ifdef CTHREAD
mutex_t cthread_fork_lock;
#endif

#ifdef LTC
#include "ltc.h"
#endif

/*
 *    ------------------
 *    primitive function
 *    ------------------
 */

#define  FETCH0(offset)     (*(base_reg+(int)(offset)))
#define  STORE(val,offset)      *(base_reg+offset)=(unsigned)(val)
#define  SCPUSH(val)        *(--sp_reg)=(unsigned)(val)
#define  SCPOP()        (*(sp_reg++))
#define  GETbcarg(n)        (*(bcbody_reg+bcpc_reg+(n)))
#define  PCUP(n)        bcpc_reg+=(n)
#define  GETconst(n) *(((unsigned *)ADDRESS( *((unsigned *)ADDRESS(*(active_reg - 1)) + 1) )) + GETbcarg(n))
#define GET_SVAL(sym_obj) (((ENTITYP)ADDRESS(sym_obj))->sym.sym_val.obj)

fetch(baseoffset, level, offset)
int baseoffset;
int level;
int offset;
{
    unsigned *p;
    int i;

    p = (unsigned *) * (base_reg - baseoffset + 1);
    for (i = 1; i < level; i++) {
        p = (unsigned *)*(p+1);
    }
    return(*(p + offset));
}


void iretaddrx();
void iretaddrl();
void s_iretaddrx();
void top();
#if defined(__CYGWIN32__) || defined(linux)
sigjmp_buf exit_bc_interpreter;
#else
jmp_buf exit_bc_interpreter;
#endif
void recover_set_gval();       /* By A243  Dec, 1992 */

void error_invalid_arg1(char *);
void error_invalid_arg2(char *, unsigned);
void error_invalid_arg3(char *, unsigned, unsigned, unsigned);

/*
 *   ---------------------
 *   BYTE CODE INTERPRETER
 *   ---------------------
 */

#ifdef CTHREAD
REGS saved_context;
#endif

/*
** for Lazy Task Cration
**   06 Aug 1997
*/
#ifdef LTC
thread_data_t **pthrs;
#endif

void bc_interpreter()
{
    void xcall();
#ifdef CONDITIONAL_FUTURE
    int counter = 0;
    unsigned int **cstk;
#endif
#ifdef LTC
    init_ltc_scheduler ();
#endif

    xcall();
#if defined(__CYGWIN32__) || defined(linux)
    switch (sigsetjmp(exit_bc_interpreter, 1)) {
#else
    switch (setjmp(exit_bc_interpreter)) {
#endif
    case 2 :
#ifdef CTHREAD
# ifdef USE_REG
#  if defined(LUNA)
        base_reg = saved_context._base_reg;
        sp_reg = saved_context._sp_reg;
        reg_reg = saved_context._reg_reg;
        bcpc_reg = saved_context._bcpc_reg;
        bcbody_reg = saved_context._bcbody_reg;
#  elif defined(SUN)
        *main_thread_regs = saved_context;
#  endif
# endif
#endif
    case 0 :
        while (1) {
#ifdef CONDITIONAL_FUTURE
            /*
             * executing main
             */
            while (1) {
                ((int(*)())(*( bcbody_reg + bcpc_reg )))();
                counter++;
                /* execute condition hook */
                if (counter > PRI_BC_INTERVAL) {
                    counter = 0;
                    break;
                }
            }
            
            /*
             * executing condition
             */
            cstk = pthrs[0]->cstack;
            if (cstk[COND_RESULT] == NULL) {
                mutex_t m = (mutex_t)(pthrs[0]->mutex);
                mutex_lock(m);
                pthrs[0]->condition = INVALID;
                pthrs[0]->cstack[COND_PARENT] = (unsigned int *)LTAG_T;
                pthrs[0]->cstack[COND_CHILD_LIST] = 0;
#ifdef DEBUG_MESG
                fprintf(stderr, "task canceled.\n");
#endif
                mutex_unlock(m);
                primary_steal();
                cstk[COND_RESULT] = (unsigned *)LTAG_FALSE; /* *** */
            } else if (pthrs[0]->condition == VALID) {
                extern void iretaddrl_cond_primary();
                extern void iretaddrl_cond_secondary();
                extern void primary_steal();
                extern void secondary_steal();
                extern waitthr_list_t *fwlp;
                
                if ((cstk[COND_PARENT] == (unsigned int *)LTAG_T)
                    && (cstk[COND_BCBODY] != (unsigned int *)LTAG_NIL)
                    && (! find_act_cstack_q(pthrs[0]))) {
                    enq_act_cstack_q(pthrs[0]);
/*                    goto gogo;  *//* *** */
                } else {
                gogo:
                    while (cstk[COND_PARENT] != (unsigned int *)LTAG_T)
                        cstk = (unsigned int **)cstk[COND_PARENT];
                    if (cstk[COND_RESULT] == (unsigned int *)LTAG_FALSE) {
                        mutex_t m = (mutex_t)(pthrs[0]->mutex);
#ifdef DEBUG_MESG
                        fprintf(stderr, "result = FALSE\n");
#endif
                        mutex_lock(m);
                        pthrs[0]->cstack[COND_RESULT] = (unsigned int *)LTAG_FALSE;
                        if (*((unsigned int *)stack_bottom - 1) != (unsigned int)primary_steal)
                            fprintf(stderr, "?????????\n");
                        *((unsigned int *)stack_bottom - 1)
                            = (unsigned int)secondary_steal;
                        if (pthrs[0]->cstack[COND_BCBODY] != LTAG_NIL) {
                            /* the task is not ordinary task in conditional
                               task group, but conditional task */
                            if (((unsigned int **)(pthrs[0]->cregs->_stack_top))[COND_SPREG][2] != (unsigned int)iretaddrl_cond_primary)
                                fprintf(stderr, "?????**\n");
                            ((unsigned int **)(pthrs[0]->cregs->_stack_top))[COND_SPREG][2] = (unsigned int)iretaddrl_cond_secondary;
                        }

                        put_waitthr(fwlp, pthrs);
                        threaddata_alloc(pthrs, 0);
                        mutex_unlock(m);
                        primary_steal();
                    }
                }
            }
#endif

#ifdef SHE
	    start_bci_pri_aux(0);
#else
            ((void(*)())(*( bcbody_reg + bcpc_reg )))();
#endif

        }
    case 1 :
        return;                 /* stack underflow */
#ifdef CALLIOC
    case 3 :                   /* stack underflow by invalid IOC invocation  */
        bokkan_stack_underflow();
#endif
    }
}




/*
 *   -------
 *   OP-CODE
 *   -------
 */
void open_frame()
{
#ifdef LTC
    if (((unsigned int **)stack_top)[LTQ_TAIL] + CHILD_GETA_OF_STACK > sp_reg)
#else
    if ((unsigned *)stack_top > sp_reg)
#endif
        bokkan_stack_overflow();
    sp_reg[-1] = LTAG_STRING | SSERDDA(bcbody_reg);
    sp_reg[-2] = GETbcarg(1);
    sp_reg[-3] = (unsigned)iretaddrx;
    sp_reg[-4] = (unsigned)active_reg;
    sp_reg[-5] = (unsigned)base_reg;
    sp_reg -= 5;
    base_reg = sp_reg;
    PCUP(2);
}


void open_lframe()
{
#ifdef LTC
    if (((unsigned int **)stack_top)[LTQ_TAIL] + CHILD_GETA_OF_STACK > sp_reg)
#else
    if ((unsigned *)stack_top > sp_reg)
#endif
        bokkan_stack_overflow();
    sp_reg[-1] = GETbcarg(3);
    sp_reg[-2] = (unsigned)iretaddrl;
    if (GETbcarg(2) == 0)
        sp_reg[-3] = FETCH0(-GETbcarg(1) + 1);
    else
        sp_reg[-3] = fetch(GETbcarg(1), GETbcarg(2), 1);
    sp_reg[-4] = (unsigned)base_reg;
    sp_reg -= 4;
    base_reg = sp_reg;
    PCUP(4);
}


void open_lframe_1()
{
    sp_reg[-1] = GETbcarg(2);
    sp_reg[-2] = (unsigned)iretaddrl;
    sp_reg[-3] = (unsigned)(base_reg - GETbcarg(1));
    sp_reg[-4] = (unsigned)base_reg;
    sp_reg -= 4;
    base_reg = sp_reg;
    PCUP(3);
}

#ifdef CALLIOC

int update_marker_s(new_stack)
unsigned *new_stack;
{
    unsigned *active;
    int n;

    active = (unsigned *)new_stack[1];
    n = 0;

    while (active != 0) {
        if (GETtag(*(active + 4)) == LTAG_IOC) {
            n++;
#ifdef SPECIAL_CONS_TAG
            ((unsigned *)ADDRESS(*(active+4)))[2] =(unsigned)mk_fix(new_stack);
#else
            ((unsigned *)ADDRESS(*(active+4)))[2] =(unsigned)mk_fix(SSERDDA(new_stack));
#endif
        }
        active = (unsigned *)*(active + 1);
    }
    return(new_stack[5] = n);
}

int update_marker_h(to, new_stack)
unsigned *to, *new_stack;
{
    SOBJ my_more;
    unsigned *epframep, *ret;
    int n, m;

    my_more.obj = new_stack[0];
    n = 0;

    while (my_more.obj != (unsigned)to) {
/*        epframep = (unsigned *)ADDRESS(CARobj(CARobj(my_more.obj))); */
        /* 961204 */
        epframep = (unsigned *)ADDRESS(CDRobj(CARobj(my_more.obj)));
        ret = (unsigned *)*(epframep + ((IMM(*epframep)>>2) - 2));
        if (GETtag(ret) == LTAG_IOC) {
            n++;
#ifdef SPECIAL_CONS_TAG
            ((unsigned *)ADDRESS(ret))[2] = (unsigned)mk_fix(new_stack);
#else
            ((unsigned *)ADDRESS(ret))[2] = (unsigned)mk_fix(SSERDDA(new_stack));
#endif
        }
        my_more.obj = CDRobj(my_more.obj);
    }
    m = n;
    while (my_more.obj != LTAG_NIL) {
/*        epframep = (unsigned *)ADDRESS(CARobj(CARobj(my_more.obj))); */
        /* 961204 */
        epframep = (unsigned *)ADDRESS(CDRobj(CARobj(my_more.obj)));
        ret = (unsigned *)*(epframep + ((IMM(*epframep)>>2) - 2));
        if (GETtag(ret) == LTAG_IOC) {
            n++;
#ifdef SPECIAL_CONS_TAG
            ((unsigned *)ADDRESS(ret))[2] = (unsigned)mk_fix(new_stack);
#else
            ((unsigned *)ADDRESS(ret))[2] = (unsigned)mk_fix(SSERDDA(new_stack));
#endif
        }
        my_more.obj = CDRobj(my_more.obj);
    }
    new_stack[5] += n;
    return(m);
}

int update_all_marker_h(to, new_stack)
unsigned *to, *new_stack;
{
    SOBJ my_more;
    unsigned *epframep, *ret;
    int n;

    my_more.obj = new_stack[0];
    n = 0;

    while (my_more.obj != (unsigned)to) {
/*        epframep = (unsigned *)ADDRESS(CARobj(CARobj(my_more.obj))); */
        /* 961204 */
        epframep = (unsigned *)ADDRESS(CDRobj(CARobj(my_more.obj)));
        ret = (unsigned *)*(epframep + ((IMM(*epframep)>>2) - 2));
        if (GETtag(ret) == LTAG_IOC) {
            n++;
#ifdef SPECIAL_CONS_TAG
            ((unsigned *)ADDRESS(ret))[2] = (unsigned)mk_fix(new_stack);
#else
            ((unsigned *)ADDRESS(ret))[2] = (unsigned)mk_fix(SSERDDA(new_stack));
#endif
        }
        my_more.obj = CDRobj(my_more.obj);
    }
    new_stack[5] += n;
    return(n);
}

#define dest_copy(n)    *(dest++) = (unsigned)(n)

void copy_stack(from, to, dest, delta)
unsigned *from, *to, *dest;
int delta;
{
    unsigned *active, *base;

    dest -= 3 + to - from;
    while (from < to) {
        base = (unsigned *)*from++;
        active = (unsigned *)*from++;
        dest_copy((int)base + delta);
        dest_copy((int)active + delta);
        while (base > from)
            dest_copy(*from++);
    }
}

void migrate_to_new_stack_ss(from, to, bottom, new_stack)
unsigned *from, *to, *bottom, *new_stack;
{
    int delta;

    delta = (int)new_stack - (int)to - 12;
    copy_stack(from, to, new_stack, delta);
    new_stack[0] = LTAG_NIL;
    new_stack[1] = (unsigned)((int)from + delta);
    bottom[5] -= update_marker_s(new_stack);
    /* DEBUG */
/*
{
    unsigned *u, *v;

    printf("SOURCE              DESTINATION\n");
    for (u = from, v = new_stack - (3 + to - from); u < to; u++, v++)
        printf("%x %x           %x %x\n", u, *u, v, *v);
}
*/
}

void migrate_to_new_stack_sh(from, to, bottom, m, new_stack)
unsigned *from, *to, *bottom, m, *new_stack;
{
    int delta;

    delta = (int)new_stack - (int)(bottom - 3) - 12;
    new_stack[0] = m;
    copy_stack(from, bottom - 3, new_stack, delta);
    new_stack[1] = (unsigned)((int)from + delta);
    bottom[5] -= update_marker_s(new_stack);
    bottom[5] -= update_marker_h(to, new_stack);
}

void migrate_to_new_stack_hh(from, to, bottom, new_stack)
unsigned *from, *to, *bottom, *new_stack;
{
    new_stack[0] = (unsigned)from;
    new_stack[1] = (unsigned)(new_stack - 3);
    new_stack[5] = 0;
    bottom[5] -= update_marker_h(to, new_stack);
}

void migrate_all_to_new_stack(new_stack)
unsigned *new_stack;
{
    SOBJ my_more;
    unsigned *active, *epframep;
    int delta;

    active = active_reg;
    while (active != 0)
        if (GETtag(*(active + 4)) == LTAG_IOC)
            break;
        else
            active = (unsigned *)*(active + 1);
    if (active != 0) {
        delta = (int)new_stack - (int)(stack_bottom - 12) - 12;
        copy_stack(active, stack_bottom - 12, new_stack, delta);
        new_stack[0] = more.obj;
        new_stack[1] = (unsigned)((int)active + delta);
        new_stack[5] = 0;
        update_marker_s(new_stack);
        update_marker_h(LTAG_NIL, new_stack);
    } else {
        my_more.obj = more.obj;
        while (my_more.obj != LTAG_NIL) {
/*            epframep = (unsigned *)ADDRESS(CARobj(CARobj(my_more.obj))); */
            /* 961204 */
            epframep = (unsigned *)ADDRESS(CDRobj(CARobj(my_more.obj)));
            active = (unsigned *)*(epframep + ((IMM(*epframep)>>2) - 2));
            if (GETtag(active) == LTAG_IOC)
                break;
            else
                my_more.obj = CDRobj(my_more.obj);
        }
        if (my_more.obj != LTAG_NIL) {
/*            new_stack[0] = more.obj; */
            new_stack[0] = my_more.obj; /* 961204 */
            new_stack[1] = (unsigned)(new_stack - 3);
            new_stack[5] = 0;
            update_marker_h(LTAG_NIL, new_stack);
        } else {
            fprintf(stderr, "Internal error in migrate_all_to_new_stack().\n");
            exit(1);
        }
    }
/*    new_stack[5] = ((unsigned *)stack_bottom)[5]; *//* 961204 */
}

#endif

#ifdef THREAD
extern int processors;
# ifdef CONDITIONAL_FUTURE
  extern int sprocessors;
# endif
extern thread_data_t **sthrs;
extern Sthe_num_of_current_thread();
#endif

void xreturn();

void xcall()
{
    SOBJ fun;
    ENTITYP funp;
    int arginfo, req_arg, argnum;
    SOBJ rest_list;
    unsigned *p;
    int i;
#ifdef THREAD
    mtus_thread_t thread_id;
#endif

#ifdef PTHREAD
# if (!defined(LTC) && !defined(SHE))
    pthread_testcancel();
# endif
#endif

#ifdef THINK_C
        KeyMap mykey;
        static int trap_count = 1000;
        
        if (!--trap_count) {
                trap_count = 1000;
                GetKeys(mykey);
                if ((mykey[1] & 0x8000) && (mykey[1] & 0x800000))
                        raise(SIGINT);
        }
#endif


#ifdef LTC
#ifdef THREAD
    /* get my thread id */
    thread_id = mtus_thread_self();
    /*
     * paranoia check (threaddata list check)
     */
    /* compare thread id */
    for(i = 0 ; i < processors ; i++) {
        if (thread_id == (pthrs[i]->thr))
            break;
    }
#endif
#ifdef CONDITIONAL_FUTURE
    if (i == processors) {
        for(i = 0 ; i < sprocessors ; i++) {
            if (thread_id == (sthrs[i]->thr))
                break;
        }
    }
#endif
#endif

    active_reg = base_reg;
#if 0
    fprintf(stderr, "xcall[%d]: active=0x%x, base=0x%x\n", i, active_reg, base_reg);
    fprintf(stderr, "xcall[%d]: fun=0x%x\n", i,*(active_reg-1));
#endif 
    fun.obj = *(active_reg - 1);
#if 0
    fprintf(stderr, "xcall[%d]: gettag=0x%x\n", i, GETtag(fun.obj));
#endif
    switch (GETtag(fun.obj)) {
    case LTAG_SFUNCTION :
    case LTAG_MSFUNCTION :
        funp = (ENTITYP)ADDRESS(fun.obj);
/*        if (funp->bcc.env.obj == LTAG_NIL) { *//* T.Komiya */
        arginfo = funp->bcc.arg_info;
        req_arg = 0xfffffffc & arginfo;
        if ( 0x3 & arginfo ) {   /* arg check with rest */
            argnum = (unsigned)base_reg - (unsigned)sp_reg;
            if (req_arg > argnum) {
                err_illegal_number_of_arguments("Too few arguments.", arginfo);
                return;
            }
            if (funp->bcc.env.obj != LTAG_NIL)
                SCPUSH(funp->bcc.env.obj);
            bcpc_reg = funp->bcc.start_addr;
            bcbody_reg = (unsigned *)ADDRESS(funp->bcc.body.obj);
        } else {                 /* arg check without rest */
            argnum = (unsigned)base_reg - (unsigned)sp_reg;
            if (argnum > req_arg) {
                err_illegal_number_of_arguments("Too many arguments.", arginfo);
                return;
            } else if (argnum < req_arg) {
                err_illegal_number_of_arguments("Too few arguments.", arginfo);
                return;
            }
            if (funp->bcc.env.obj != LTAG_NIL)
                SCPUSH(funp->bcc.env.obj);
            bcpc_reg = funp->bcc.start_addr;
            bcbody_reg = (unsigned *)ADDRESS(funp->bcc.body.obj);
        }
        break;
/*        } *//* T.Komiya */

    case LTAG_BCLOSURE :
    case LTAG_MBCLOSURE :
        funp = (ENTITYP)ADDRESS(fun.obj);
        arginfo = funp->bcc.arg_info;
        req_arg = 0xfffffffc & arginfo;
#if 0
        fprintf(stderr, "xcall[%d]: hoge: arginfo=0x%x\n", i, arginfo);
#endif
        if ( 0x3 & arginfo ) {   /* arg check with rest */
            argnum = (unsigned)base_reg - (unsigned)sp_reg;
#if 0
            fprintf(stderr, "xcall[%d]: hoge: req_arg=%d, argnum=%d\n", i, req_arg, argnum);
#endif
            if (req_arg == argnum) {
                SCPUSH(LTAG_NIL);
            } else if (req_arg < argnum) {   /* make rest */
                rest_list.obj = LTAG_NIL;
                for (p=(unsigned *)((unsigned)base_reg-req_arg); sp_reg<p; ) {
                    rest_list.obj = mk_cons(SCPOP(), rest_list.obj);
                }
                SCPUSH(rest_list.obj);
                funp = (ENTITYP)ADDRESS(*(active_reg - 1)); /* for GC */
            } else {
#ifdef THREAD
                fprintf(stderr, "xcall[%d]: hoge: active=0x%x, base=0x%x, sp_reg=0x%x fun=0x%x\n", i, active_reg, base_reg, sp_reg, Sthe_num_of_current_thread);
#endif
                err_illegal_number_of_arguments("Too few arguments.", arginfo);
                return;
            }

            bcpc_reg = funp->bcc.start_addr;
            bcbody_reg = (unsigned *)ADDRESS(funp->bcc.body.obj);
            if (funp->bcc.env.obj != LTAG_NIL)
                SCPUSH(funp->bcc.env.obj);
        } else {                 /* arg check without rest */
            argnum = (unsigned)base_reg - (unsigned)sp_reg;
            if (argnum > req_arg) {
                err_illegal_number_of_arguments("Too many arguments.", arginfo);
                return;
            } else if (argnum < req_arg) {
                err_illegal_number_of_arguments("Too few arguments.", arginfo);
                return;
            }
            bcpc_reg = funp->bcc.start_addr;
            bcbody_reg = (unsigned *)ADDRESS(funp->bcc.body.obj);
            if (funp->bcc.env.obj != LTAG_NIL)
                SCPUSH(funp->bcc.env.obj);
        }
        break;

    case LTAG_NCLOSURE :        /* T.Komiya */
    case LTAG_MNCLOSURE :
        funp = (ENTITYP)ADDRESS(fun.obj);
        arginfo = funp->ncc.arg_info;
        req_arg = 0xfffffffc & arginfo;
        if ( 0x3 & arginfo ) {   /* arg check with rest */
            argnum = (unsigned)base_reg - (unsigned)sp_reg;
            if (req_arg == argnum) {
                SCPUSH(LTAG_NIL);
            } else if (req_arg < argnum) {   /* make rest */
                rest_list.obj = LTAG_NIL;
                for (p=(unsigned *)((unsigned)base_reg-req_arg); sp_reg<p; ) {
                    rest_list.obj = mk_cons(SCPOP(), rest_list.obj);
                }
                SCPUSH(rest_list.obj);
                funp = (ENTITYP)ADDRESS(*(active_reg - 1)); /* for GC */
            } else {
                err_illegal_number_of_arguments("Too few arguments.", arginfo);
                return;
            }

            bcpc_reg = 1;
            bcbody_reg = (unsigned *)ADDRESS(funp->ncc.body.obj);
            if (funp->ncc.env.obj != LTAG_NIL)
                SCPUSH(funp->ncc.env.obj);
        } else {                 /* arg check without rest */
            argnum = (unsigned)base_reg - (unsigned)sp_reg;
            if (argnum > req_arg) {
                err_illegal_number_of_arguments("Too many arguments.", arginfo);
                return;
            } else if (argnum < req_arg) {
                err_illegal_number_of_arguments("Too few arguments.", arginfo);
                return;
            }
            bcpc_reg = 1;
            bcbody_reg = (unsigned *)ADDRESS(funp->ncc.body.obj);
            if (funp->ncc.env.obj != LTAG_NIL)
                SCPUSH(funp->ncc.env.obj);
        }
        break;

    case LTAG_CONTINUATION :
#ifdef CTHREAD
        argnum = (unsigned)active_reg - (unsigned)sp_reg;
        if (argnum != 8)
            if (argnum != 12){
                err_illegal_argument("The number of continuation's arguments must be one or two.");
                return;
            } else {
                optcnt.obj = mk_cons(*(active_reg - 3), optcnt.obj);
                funp = (ENTITYP)ADDRESS(*(active_reg - 1)); /* for GC */
            }
        else {
            funp = (ENTITYP)ADDRESS(fun.obj);
            optcnt.obj = funp->cnt.opt.obj;
        }
        reg_reg = *(active_reg - 2);
        more.obj = funp->cnt.body.obj;
        sp_reg = (unsigned *)stack_bottom;
        top();
        break;
#else
        if ( ((unsigned)active_reg - (unsigned)sp_reg) != 8 ){
            err_illegal_argument("The number of continuation's arguments must be one.");
            return;
        }
        reg_reg = *(active_reg - 2);
#ifdef CALLIOC
        if (((unsigned *)stack_bottom)[5]) {
            unsigned *new_stack, *__alloc_stack();

            new_stack = __alloc_stack();
            migrate_all_to_new_stack(new_stack);
            fun.obj = *(active_reg - 1);
        }
        funp = (ENTITYP)ADDRESS(fun.obj);
        ((unsigned *)stack_bottom)[5] = fix_to_int(funp->cnt.count.obj);
        more.obj = funp->cnt.body.obj;
#else
        funp = (ENTITYP)ADDRESS(fun.obj);
        more.obj = funp->cnt.body.obj;
#endif
        sp_reg = (unsigned *)stack_bottom;
        top();
        break;
#endif
#ifdef CALLIOC
    case LTAG_IOC :
    {
        SOBJ my_more;
        unsigned *active;
        unsigned *epframep;
        unsigned *fun_bottom;
        unsigned *__alloc_stack();

        funp = (ENTITYP)ADDRESS(fun.obj);
        reg_reg = *top_val();
#ifdef SPECIAL_CONS_TAG
        fun_bottom =(unsigned *)fix_to_int(funp->ioc.bottom.obj);
#else
        fun_bottom =(unsigned *)ADDRESS(fix_to_int(funp->ioc.bottom.obj));
#endif
        if (fun_bottom != (unsigned *)stack_bottom) {
            ((unsigned *)stack_bottom)[0] = more.obj;
            ((unsigned *)stack_bottom)[1] = (unsigned)active_reg;
            active = (unsigned *)fun_bottom[1];
        } else
            active = active_reg;

        while (active != 0)
            if (GETtag(*(active + 4)) != LTAG_IOC)
                active = (unsigned *)*(active + 1);
            else
                break;
        if (active != 0) {
            unsigned *act1_s;

            if (fun.obj == *(active + 4))
                active_reg = active;
            else {
                act1_s = active;
                while (active != 0)
                    if (*(active + 4) != fun.obj)
                        active = (unsigned *)*(active + 1);
                    else
                        break;
                if (active != 0) {
                    unsigned *act2_s, *new_stack;

                    act2_s = active;
                    s_push(LTAG_IOC | SSERDDA(funp));
                    new_stack = __alloc_stack();
                    funp = (ENTITYP)ADDRESS(s_pop());
                    migrate_to_new_stack_ss(act1_s, act2_s, fun_bottom, new_stack);
                    active_reg = act2_s;
                } else {
                    if (fun_bottom != (unsigned *)stack_bottom)
                        my_more.obj = fun_bottom[0];
                    else
                        my_more.obj = more.obj;
                    s_push(my_more.obj);
                    while (my_more.obj != LTAG_NIL) {
/*                        epframep = (unsigned *)ADDRESS(CARobj(CARobj(my_more.obj))); */
                        /* 961204 */
                        epframep = (unsigned *)ADDRESS(CDRobj(CARobj(my_more.obj)));
                        active = (unsigned *)*(epframep + ((IMM(*epframep)>>2) - 2));
                        
                        if ((unsigned int)active != fun.obj)
                            my_more.obj = CDRobj(my_more.obj);
                        else
                            break;
                    }
                    if (my_more.obj != LTAG_NIL) {
                        unsigned *new_stack;

                        more.obj = my_more.obj;
                        s_push(LTAG_IOC | SSERDDA(funp));
                        new_stack = __alloc_stack();
                        funp = (ENTITYP)ADDRESS(s_pop());
                        migrate_to_new_stack_sh(act1_s,
                                                (unsigned *)more.obj,
                                                fun_bottom,
                                                s_pop(),
                                                new_stack);
/*                        sp_reg = (unsigned *)stack_bottom; */
                        goto L1;
                    } else {
                        s_pop();
                        err_error("Invalid invocation of IOC.");
                    }
                }
            }
            if (fun_bottom != (unsigned *)stack_bottom)
                more.obj = fun_bottom[0];
        } else {
            if (fun_bottom != (unsigned *)stack_bottom)
                my_more.obj = fun_bottom[0];
            else
                my_more.obj = more.obj;
            while (my_more.obj != LTAG_NIL) {
/*                epframep = (unsigned *)ADDRESS(CARobj(CARobj(my_more.obj))); */
                /* 961204 */
                epframep = (unsigned *)ADDRESS(CDRobj(CARobj(my_more.obj)));
                active = (unsigned *)*(epframep + ((IMM(*epframep)>>2) - 2));
         
                if (GETtag(active) != LTAG_IOC)
                    my_more.obj = CDRobj(my_more.obj);
                else
                    break;
            }
            if (my_more.obj != LTAG_NIL) {
                unsigned *act1_h;

                act1_h = (unsigned *)my_more.obj;
                if (fun.obj == (unsigned int)active) {
                    more.obj = (unsigned)act1_h;
/*                    sp_reg = (unsigned *)stack_bottom; */
                    goto L1;
                } else {
                    while (my_more.obj != LTAG_NIL) {
/*                        epframep = (unsigned *)ADDRESS(CARobj(CARobj(my_more.obj))); */
                        /* 961204 */
                        epframep = (unsigned *)ADDRESS(CDRobj(CARobj(my_more.obj)));
                        active = (unsigned *)*(epframep + ((IMM(*epframep)>>2) - 2));
                        
                        if (active != (unsigned *)fun.obj)
                            my_more.obj = CDRobj(my_more.obj);
                        else
                            break;
                    }
                    if (my_more.obj != LTAG_NIL) {
                        unsigned *new_stack;

                        more.obj = my_more.obj;
                        s_push(LTAG_IOC | SSERDDA(funp));
                        new_stack = __alloc_stack();
                        funp = (ENTITYP)ADDRESS(s_pop());
                        migrate_to_new_stack_hh(act1_h, more.obj,
                                                fun_bottom, new_stack);
/*                        sp_reg = (unsigned *)stack_bottom; */
                        goto L1;
                    } else
                        err_error("Invalid invocation of IOC.");
                }
            } else
                err_error("Invalid invocation of IOC.");
        }
        if (fun_bottom != (unsigned *)stack_bottom) {
            if (((unsigned *)stack_bottom)[5] == 0)
                free_stack((unsigned *)stack_bottom);
            stack_bottom = (char *)fun_bottom;
            stack_top = stack_bottom - STACK_SIZE + 28 + GETA_OF_STACK;
        }
        xreturn();
        break;
    L1:
        if (fun_bottom != (unsigned *)stack_bottom) {
            if (((unsigned *)stack_bottom)[5] == 0)
                free_stack((unsigned *)stack_bottom);
            stack_bottom = (char *)fun_bottom;
            stack_top = stack_bottom - STACK_SIZE + 28 + GETA_OF_STACK;
        }
        sp_reg = (unsigned *)stack_bottom;
        top();
        break;
    }
#endif
    default :
        err_not_function_apply();  /*  By N.Watanabe Oct. 28, 1992  */
        return;
    }
}


void lcall()
{
    bcpc_reg = GETbcarg(1);
}


void trcall()
{
    unsigned *p;
    int i;

    p = base_reg;
    base_reg = sp_reg;
    sp_reg = p;
    for (i = GETbcarg(1) - 1; i >= 0; i--) {
        SCPUSH(FETCH0(i));
    }

    base_reg = p;
    SCPUSH(reg_reg);
    xcall();
}


void lreturn();

void trlcall()
{
    int level;  /* lexical level */
    int n;      /* the number of args */
    unsigned *p;
    level = GETbcarg(1);
    while ((level != 0) && (*(bcbody_reg + FETCH0(3)) == (unsigned)lreturn)) {
        base_reg = (unsigned *)(*base_reg);
        level--;
    }
    if (level != 0)
        STORE(fetch(0, level, 1), 1);

    n = GETbcarg(2);
    if (n == 0) {
        sp_reg = base_reg;
    } else {
        p = base_reg;
        base_reg = sp_reg;
        sp_reg = p;
        for (n = n - 2; n >= 0; n--) {
            SCPUSH(FETCH0(n));
        }
        base_reg = p;
        SCPUSH(reg_reg);
    }

    bcpc_reg = GETbcarg(3);
}

void xreturn();

void ltrxcall()
{
    int i;
    unsigned *p;
    while (*(bcbody_reg + FETCH0(3)) == (unsigned)lreturn) {
        base_reg = (unsigned *)(*base_reg);
    }
/* original
    if (base_reg != active_reg) {
        STORE(active_reg, 1);
        STORE(s_iretaddrx, 2);
    }
*/
    /* bug fix */
    if (*(bcbody_reg + FETCH0(3)) == (unsigned)xreturn) {
        base_reg = (unsigned *)(*base_reg);
    } else if (base_reg != active_reg) { /* unnecessary check? */
        STORE(active_reg, 1);
        STORE(s_iretaddrx, 2);
    }
    /* bug fix end */

    p = base_reg;
    base_reg = sp_reg;
    sp_reg = p;
    for (i = GETbcarg(1) - 1; i >= 0; i--) {
        SCPUSH(FETCH0(i));
    }
    base_reg = p;
    SCPUSH(reg_reg);
    xcall();
}


void strcall()
{
}


void xreturn()
{
    int (*retaddr)();
    sp_reg = active_reg;
    base_reg = (unsigned *)sp_reg[0];
    active_reg = (unsigned *)sp_reg[1];
    retaddr = (int (*)())sp_reg[2];
    sp_reg += 3;
    retaddr();
}


void lreturn()
{
    int (*retaddr)();
    sp_reg = base_reg;
    base_reg = (unsigned *)sp_reg[0];
/*    SCPOP(); */
    retaddr = (int (*)())sp_reg[2];
    sp_reg += 3;
    retaddr();
}


void jmp()
{
    bcpc_reg = GETbcarg(1);
}


#ifdef USE_FALSE
void jmp_if_true()
{
#ifdef IEEE_SPEC
    if (reg_reg != LTAG_FALSE)
        bcpc_reg = GETbcarg(1);
#else
    if (reg_reg != LTAG_FALSE && reg_reg != LTAG_NIL)
        bcpc_reg = GETbcarg(1);
#endif
    else
        PCUP(2);
}


#else
void jmp_if_true()
{
    if (reg_reg != LTAG_NIL)
        bcpc_reg = GETbcarg(1);
    else
        PCUP(2);
}


#endif


#ifdef USE_FALSE
void jmp_if_false()
{
#ifdef IEEE_SPEC
    if (reg_reg == LTAG_FALSE)
        bcpc_reg = GETbcarg(1);
#else
    if (reg_reg == LTAG_FALSE || reg_reg == LTAG_NIL)
        bcpc_reg = GETbcarg(1);
#endif
    else
        PCUP(2);
}


#else
void jmp_if_false()
{
    if (reg_reg == LTAG_NIL)
        bcpc_reg = GETbcarg(1);
    else
        PCUP(2);
}


#endif


void get_const()
{
    reg_reg = GETconst(1);
    PCUP(2);
}


void get_gval()
{
    SOBJ sym;
    SOBJ val;
    sym.obj = GETconst(1);
    val.obj = GET_SVAL(sym.obj);
    if (unbound_obj.obj == val.obj){
#ifdef PVM3    /* By A243 */
        if(pvm_parent() == -23){
#endif 
            err_unbound_symbol2(sym.obj);  /*  By N.Watanabe Oct. 05. 1992.  */
            return ;
#ifdef PVM3    /* By A243 */
        }
        else{
            val.obj = pvm_err_unbound_obj2(sym.obj);
            return;
        }
#endif 
    }
    reg_reg = val.obj;
    PCUP(2);
}


void get_lval()
{
    reg_reg = fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3));
    PCUP(4);
}

void get_lval0()
{
    reg_reg = FETCH0(GETbcarg(1));
    PCUP(2);
}


void get_hval()
{
    reg_reg = CARobj(fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3)));
    PCUP(4);
}


void get_hval0()
{
    reg_reg = CARobj(FETCH0(GETbcarg(1)));
    PCUP(2);
}


void get_cval()
{
    SOBJ  c_list;
    int i;
    c_list.obj = fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3));
    for (i = GETbcarg(4); i > 0; i--) {
        c_list.obj = CDRobj(c_list.obj);
    }
    reg_reg = CARobj(c_list.obj);
    PCUP(5);
}


void get_cval0()
{
    SOBJ  c_list;
    int i;
    c_list.obj = FETCH0(GETbcarg(1));
    for (i = GETbcarg(2); i > 0; i--)
        c_list.obj = CDRobj(c_list.obj);
    reg_reg = CARobj(c_list.obj);
    PCUP(3);
}


void push_reg()
{
    SCPUSH(reg_reg);
    PCUP(1);
}


void push_const()
{
    SCPUSH(GETconst(1));
    PCUP(2);
}

void push_gval()
{
    SOBJ sym;
    SOBJ val;

    sym.obj = GETconst(1);
    val.obj = GET_SVAL(sym.obj);
    if (unbound_obj.obj == val.obj){
#ifdef PVM3 /* By A243 */
        if(pvm_parent() == -23){
#endif
            err_unbound_symbol3(sym.obj);  /*  By N.Watanabe Oct. 05. 1992.  */
            return;
#ifdef PVM3 /* By A243 */
        }
        else{
            val.obj = pvm_err_unbound_obj3(sym.obj);
            return;
        }
#endif
    }
    SCPUSH(val.obj);
    PCUP(2);
}


void push_lval()
{
    SCPUSH(fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3)));
    PCUP(4);
}


void push_lval0()
{
    SCPUSH(FETCH0(GETbcarg(1)));
    PCUP(2);
}


void push_hval()
{
    SCPUSH(CARobj(fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3))));
    PCUP(4);
}


void push_hval0()
{
    SCPUSH(CARobj(FETCH0(GETbcarg(1))));
    PCUP(2);
}


void push_cval()
{
    SOBJ  c_list;
    int i;
    c_list.obj = fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3));
    for (i = GETbcarg(4); i > 0; i--)
        c_list.obj = CDRobj(c_list.obj);
    SCPUSH(CARobj(c_list.obj));
    PCUP(5);
}


void push_cval0()
{
    SOBJ  c_list;
    int i;
    c_list.obj = FETCH0(GETbcarg(1));
    for (i = GETbcarg(2); i > 0; i--)
        c_list.obj = CDRobj(c_list.obj);
    SCPUSH(CARobj(c_list.obj));
    PCUP(3);
}


void set_gval()     
{
    SOBJ sym;

    sym.obj = GETconst(1);

    if (GETtag(sym.obj) != LTAG_ORDINARY) {
        /* T.Komiya */
        print_con("Warning: ");
        printobj_con(sym.obj, PRINT, 5);
        print_con(" is being redefined.\n");
        ((ENTITYP)ADDRESS(sym.obj))->sym.sym_val.obj = reg_reg;
    } else if (((ENTITYP)ADDRESS(sym.obj))->sym.sym_val.obj
               == unbound_obj.obj ) {
#ifdef PVM3 /* By A243 */
        if(pvm_parent() == -23){
#endif
            err_unbound_symbol4(sym.obj);
//            return NULL;
            return;
#ifdef PVM3 /* By A243 */
        }       
        else {
/*          ((ENTITYP)ADDRESS(sym.obj))->sym.sym_val.obj 
                = pvm_err_unbound_obj4(sym.obj); */
            pvm_err_unbound_obj4(sym.obj);
            return;
        }
#endif
    } else
        ((ENTITYP)ADDRESS(sym.obj))->sym.sym_val.obj = reg_reg;

    if((GETtag(reg_reg) == LTAG_BCLOSURE)       /* By A243  Dec, 1992 */    
                  && (fix_to_int(GET_SVAL(recomlevel)) >= 1)){
        ((ENTITYP)ADDRESS(reg_reg))->bcc.fun_name.obj = sym_name(sym.obj);
        open_frame2(recover_set_gval, LTAG_STRING | SSERDDA(bcbody_reg), bcpc_reg);
        active_reg = base_reg;
        s_push(GET_SVAL(recomsym));
        s_push(reg_reg);
        xcall();
        open_frame2(iretaddrx, LTAG_STRING | SSERDDA(bcbody_reg), bcpc_reg);
        active_reg = base_reg;
        xreturn();
        return;
    }

    PCUP(2);
}

void recover_set_gval()            /* By A243  Dec,1992 */
{
    bcpc_reg = SCPOP();
    bcbody_reg = (unsigned *)ADDRESS(SCPOP());
    PCUP(2);
}

#ifdef CALLIOC
void recover_set_gval_ep()
{
    bcpc_reg = SCPOP();
    bcbody_reg = (unsigned *)ADDRESS(*((unsigned *)ADDRESS(SCPOP())+1));
    PCUP(2);
    ((unsigned *)stack_bottom)[5]--;
}
#endif

void set_lval0()
{
    STORE(reg_reg, GETbcarg(1));
    PCUP(2);
}


void set_hval()
{
#ifdef SPECIAL_CONS_TAG
    *((unsigned *)fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3))) = reg_reg;
#else
    *((unsigned *)ADDRESS(fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3)))) = reg_reg;
#endif
    PCUP(4);
}


void set_hval0()
{
#ifdef SPECIAL_CONS_TAG
    *((unsigned *)FETCH0(GETbcarg(1))) = reg_reg;
#else
    *((unsigned *)ADDRESS(FETCH0(GETbcarg(1)))) = reg_reg;
#endif
    PCUP(2);
}


void set_cval()
{
    SOBJ  c_list;
    int i;
    c_list.obj = fetch(GETbcarg(1), GETbcarg(2), GETbcarg(3));
    for (i = GETbcarg(4); i > 0; i--)
        c_list.obj = CDRobj(c_list.obj);
#ifdef SPECIAL_CONS_TAG
    *((unsigned *)c_list.obj) = reg_reg;
#else
    *((unsigned *)ADDRESS(c_list.obj)) = reg_reg;
#endif
    PCUP(5);
}


void set_cval0()
{
    SOBJ  c_list;
    int i;
    c_list.obj = FETCH0(GETbcarg(1));
    for (i = GETbcarg(2); i > 0; i--)
        c_list.obj = CDRobj(c_list.obj);
#ifdef SPECIAL_CONS_TAG
    *((unsigned *)c_list.obj) = reg_reg;
#else
    *((unsigned *)ADDRESS(c_list.obj)) = reg_reg;
#endif
    PCUP(3);
}


void lift()
{
    int offset0;
    offset0 = GETbcarg(1);
    STORE(mk_cons(FETCH0(offset0), fetch(GETbcarg(2), GETbcarg(3), GETbcarg(4))),
        offset0);
    PCUP(5);
}


void lift0()
{
    int offset0;
    offset0 = GETbcarg(1);
    STORE(mk_cons(FETCH0(offset0), FETCH0(GETbcarg(2))), offset0);
    PCUP(3);
}


void lift_nil()
{
    int offset0;
    offset0 = GETbcarg(1);
    STORE(mk_cons(FETCH0(offset0), LTAG_NIL), offset0);
    PCUP(2);
}


void make_rest()
{
    SOBJ  rest_list;
    unsigned *p;
    reg1 = LTAG_NIL;
    for (p = base_reg + (int)GETbcarg(1); sp_reg <= p; ) {
        reg1 = mk_cons(SCPOP(), reg1);
    }
    SCPUSH(reg1);
    reg1 = LTAG_NIL;
    PCUP(2);
}


void make_closure()
{
    BCCLOS  * parent;
    parent = (BCCLOS * )ADDRESS(*(active_reg - 1));
    reg_reg = mk_bcc(LTAG_BCLOSURE,  parent->const_vec.obj, LTAG_NIL,
        GETconst(3), reg_reg, parent->body.obj,
        trans_arginfo(GETbcarg(1)), GETbcarg(2));
    PCUP(4);
}


void make_promise()
{
    reg_reg = mk_promise(getarg(0), reg_reg, GETbcarg(1));
    PCUP(2);
}

#ifdef CTHREAD

# ifndef SHE
void make_future()
{
    reg_reg = mk_future(getarg(0), reg_reg, GETbcarg(1));
    PCUP(2);
}
#endif

#ifdef SHE
extern thread_info_t pri_act, sub_act;
#endif

void touch()
{
    SOBJ val;
/*
    printf("bcbody = %x, bcpc = %x\n", bcbody_reg, bcpc_reg);
 */
    val.obj = FETCH0(GETbcarg(1));
#if 0
    for (;;) {
        if (GETtag((val.obj)) == LTAG_FUTURE) {
            ENTITYP         fut = (ENTITYP) ADDRESS((val.obj));
            sst_frame_t    *sstfr = fut->future.sstfr;
            if (sstfr == NULL) {
                (val.obj) = fut->future.value.obj;
                ((STORE(val.obj, GETbcarg(1))));
                continue;
            }
            mutex_lock(sstfr->frame_lock);
            if (fut->future.sstfr == NULL) {
                mutex_unlock(sstfr->frame_lock);
                (val.obj) = fut->future.value.obj;
                ((STORE(val.obj, GETbcarg(1))));
                continue;
            } else if (sstfr_status(sstfr) == EXECUTING) {
                if (thr_data[THR_STATUS] == SUBSIDIARY_S) {
                    if (thr_data[THR_STATE] == RUN) {
                        mutex_lock(pri_act.lock);
                        sub_act.active--;
                        mutex_unlock(pri_act.lock);
                    }
                } else {
                    mutex_lock(pri_act.lock);
                    pri_act.active--;
                    mutex_unlock(pri_act.lock);
                }
                while (((fut->future.sstfr != NULL)))
                    condition_wait(sstfr->cond_obj, sstfr->frame_lock);
                mutex_unlock(sstfr->frame_lock);
                if (thr_data[THR_STATUS] == SUBSIDIARY_S) {
                    if (thr_data[THR_STATE] == RUN) {
                        mutex_lock(pri_act.lock);
                        sub_act.active++;
                        mutex_unlock(pri_act.lock);
                    }
                } else {
                    mutex_lock(pri_act.lock);
                    pri_act.active++;
                    mutex_unlock(pri_act.lock);
                } (((val.obj) = fut->future.value.obj));
                (((STORE(val.obj, GETbcarg(1)))));;
                continue;
            } else {
                mutex_unlock(sstfr->frame_lock);
                {
                    sst_frame_t    *poped_sstfr;
                    poped_sstfr = get_sst_frame_aux(sstfr);
                    if (poped_sstfr) {
                        if (par_constr(poped_sstfr) == FUTURE_CONST) {
                            open_frame2(iretaddrx_she_seq_fut, LTAG_STRING | SSERDDA(bcbody_reg), bcpc_reg);
                            s_push(poped_sstfr->func_obj.obj);
                            s_push(poped_sstfr->result.obj);
                            s_push(reg_reg);
                            copy_env_to_cstack(poped_sstfr);
                            if (poped_sstfr->env_list.obj != LTAG_NIL)
                                s_push(poped_sstfr->env_list.obj);
                            active_reg = base_reg;
                            bcpc_reg = poped_sstfr->pc1;
                            bcbody_reg = (unsigned *) ADDRESS(((ENTITYP) ADDRESS(poped_sstfr->func_obj.obj))->bcc.body.obj);
                        } else {
                            open_frame2(iretaddrx_she_seq_st, LTAG_STRING | SSERDDA(bcbody_reg), bcpc_reg);
                            s_push(poped_sstfr->func_obj.obj);
                            s_push(LTAG_STEALABLE | SSERDDA(poped_sstfr));
                            s_push(reg_reg);
                            copy_env_to_cstack(poped_sstfr);
                            if (poped_sstfr->env_list.obj != LTAG_NIL)
                                s_push(poped_sstfr->env_list.obj);
                            active_reg = base_reg;
                            bcpc_reg = poped_sstfr->pc1;
                            bcbody_reg = (unsigned *) ADDRESS(((ENTITYP) ADDRESS(poped_sstfr->func_obj.obj))->bcc.body.obj);
                        };
                        return;
                    } else {
                        mutex_lock(sstfr->frame_lock);
                        if (thr_data[THR_STATUS] == SUBSIDIARY_S) {
                            if (thr_data[THR_STATE] == RUN) {
                                mutex_lock(pri_act.lock);
                                sub_act.active--;
                                mutex_unlock(pri_act.lock);
                            }
                        } else {
                            mutex_lock(pri_act.lock);
                            pri_act.active--;
                            mutex_unlock(pri_act.lock);
                        }
                        while ((((fut->future.sstfr != NULL))))
                            condition_wait(sstfr->cond_obj, sstfr->frame_lock);
                        mutex_unlock(sstfr->frame_lock);
                        if (thr_data[THR_STATUS] == SUBSIDIARY_S) {
                            if (thr_data[THR_STATE] == RUN) {
                                mutex_lock(pri_act.lock);
                                sub_act.active++;
                                mutex_unlock(pri_act.lock);
                            }
                        } else {
                            mutex_lock(pri_act.lock);
                            pri_act.active++;
                            mutex_unlock(pri_act.lock);
                        } ((((val.obj) = fut->future.value.obj)));
                        ((((STORE(val.obj, GETbcarg(1))))));;
                    }
                };
                continue;
            }
        } else if (GETtag((val.obj)) == LTAG_STEALABLE) {
            sst_frame_t    *sstfr = (sst_frame_t *) ADDRESS((val.obj));
            if (sstfr_status(sstfr) == DETERMINED) {
                (val.obj) = sstfr->result.obj;
                free_sst_frame(sstfr);
                ((STORE(val.obj, GETbcarg(1))));
                continue;
            }
            mutex_lock(sstfr->frame_lock);
            if (sstfr_status(sstfr) == DETERMINED) {
                mutex_unlock(sstfr->frame_lock);
                (val.obj) = sstfr->result.obj;
                free_sst_frame(sstfr);
                ((STORE(val.obj, GETbcarg(1))));
                continue;
            } else if (sstfr_status(sstfr) == EXECUTING) {
                if (thr_data[THR_STATUS] == SUBSIDIARY_S) {
                    if (thr_data[THR_STATE] == RUN) {
                        mutex_lock(pri_act.lock);
                        sub_act.active--;
                        mutex_unlock(pri_act.lock);
                    }
                } else {
                    mutex_lock(pri_act.lock);
                    pri_act.active--;
                    mutex_unlock(pri_act.lock);
                }
                while (((sstfr_status(sstfr) != DETERMINED)))
                    condition_wait(sstfr->cond_obj, sstfr->frame_lock);
                mutex_unlock(sstfr->frame_lock);
                if (thr_data[THR_STATUS] == SUBSIDIARY_S) {
                    if (thr_data[THR_STATE] == RUN) {
                        mutex_lock(pri_act.lock);
                        sub_act.active++;
                        mutex_unlock(pri_act.lock);
                    }
                } else {
                    mutex_lock(pri_act.lock);
                    pri_act.active++;
                    mutex_unlock(pri_act.lock);
                } ({
                    (val.obj) = sstfr->result.obj;
                    free_sst_frame(sstfr);
                });
                (((STORE(val.obj, GETbcarg(1)))));;
                continue;
            } else {
                mutex_unlock(sstfr->frame_lock);
                {
                    sst_frame_t    *poped_sstfr;
                    poped_sstfr = get_sst_frame_aux(sstfr);
                    if (poped_sstfr) {
                        if (par_constr(poped_sstfr) == FUTURE_CONST) {
                            open_frame2(iretaddrx_she_seq_fut, LTAG_STRING | SSERDDA(bcbody_reg), bcpc_reg);
                            s_push(poped_sstfr->func_obj.obj);
                            s_push(poped_sstfr->result.obj);
                            s_push(reg_reg);
                            copy_env_to_cstack(poped_sstfr);
                            if (poped_sstfr->env_list.obj != LTAG_NIL)
                                s_push(poped_sstfr->env_list.obj);
                            active_reg = base_reg;
                            bcpc_reg = poped_sstfr->pc1;
                            bcbody_reg = (unsigned *) ADDRESS(((ENTITYP) ADDRESS(poped_sstfr->func_obj.obj))->bcc.body.obj);
                        } else {
                            open_frame2(iretaddrx_she_seq_st, LTAG_STRING | SSERDDA(bcbody_reg), bcpc_reg);
                            s_push(poped_sstfr->func_obj.obj);
                            s_push(LTAG_STEALABLE | SSERDDA(poped_sstfr));
                            s_push(reg_reg);
                            copy_env_to_cstack(poped_sstfr);
                            if (poped_sstfr->env_list.obj != LTAG_NIL)
                                s_push(poped_sstfr->env_list.obj);
                            active_reg = base_reg;
                            bcpc_reg = poped_sstfr->pc1;
                            bcbody_reg = (unsigned *) ADDRESS(((ENTITYP) ADDRESS(poped_sstfr->func_obj.obj))->bcc.body.obj);
                        };
                        return;
                    } else {
                        mutex_lock(sstfr->frame_lock);
                        if (thr_data[THR_STATUS] == SUBSIDIARY_S) {
                            if (thr_data[THR_STATE] == RUN) {
                                mutex_lock(pri_act.lock);
                                sub_act.active--;
                                mutex_unlock(pri_act.lock);
                            }
                        } else {
                            mutex_lock(pri_act.lock);
                            pri_act.active--;
                            mutex_unlock(pri_act.lock);
                        }
                        while ((((sstfr_status(sstfr) != DETERMINED))))
                            condition_wait(sstfr->cond_obj, sstfr->frame_lock);
                        mutex_unlock(sstfr->frame_lock);
                        if (thr_data[THR_STATUS] == SUBSIDIARY_S) {
                            if (thr_data[THR_STATE] == RUN) {
                                mutex_lock(pri_act.lock);
                                sub_act.active++;
                                mutex_unlock(pri_act.lock);
                            }
                        } else {
                            mutex_lock(pri_act.lock);
                            pri_act.active++;
                            mutex_unlock(pri_act.lock);
                        } (({
                            (val.obj) = sstfr->result.obj;
                            free_sst_frame(sstfr);
                        }));
                        ((((STORE(val.obj, GETbcarg(1))))));;
                    }
                };
                continue;
            }
        } else
            break;
    }
#else
    TOUCH(val.obj, (STORE(val.obj, GETbcarg(1))));
#endif
    PCUP(2);
}

void touch_top()
{
    SOBJ val;

    val.obj = *top_val();
    TOUCH(val.obj, (*top_val() = val.obj));
    PCUP(1);
}

void touch_bx()
{
    SOBJ val;

    val.obj = reg_reg;
    TOUCH(val.obj, (reg_reg = val.obj));
    PCUP(1);
}

void make_plchld()
{
    FUTURE *mk_plchld();

    reg_reg = LTAG_FUTURE | SSERDDA(mk_plchld(LTAG_NIL, LTAG_NIL, 0));
    PCUP(1);
}

void context_cp(dest)
REGS *dest;
{
    dest->_active_reg = active_reg;
    dest->_base_reg = base_reg;
    dest->_sp_reg = sp_reg;
    dest->_reg_reg = reg_reg;
    dest->_bcpc_reg = bcpc_reg;
    dest->_bcbody_reg = bcbody_reg;
    dest->_stack_top = stack_top;
    dest->_stack_bottom = stack_bottom;
    dest->_stack_area_bottom = stack_area_bottom;
    dest->_old_top = old_top;
    dest->_old_top1 = old_top1;
    dest->_reg1 = reg1;
    dest->_reg2 = reg2;
    dest->_reg3 = reg3;
    dest->_call_cc_arg = call_cc_arg;
    dest->_top_of_aflk = top_of_aflk;
    dest->_end_of_aflk = end_of_aflk;
    dest->_af_tmp = af_tmp;
    dest->_more = more;
    dest->_optcnt = optcnt;
}


void setup_child(context)
REGS *context;
{
    cthread_set_data(cthread_self(), (any_t)context);
#if defined(USE_REG)
# ifdef LUNA
/*    active_reg = context->_active_reg; */
    base_reg = context->_base_reg;
    sp_reg = context->_sp_reg;
    reg_reg = context->_reg_reg;
    bcpc_reg = context->_bcpc_reg;
    bcbody_reg = context->_bcbody_reg;
# elif defined(SUN)
    *main_thread_regs = *context;
# endif
#endif
    reg_reg = LTAG_FALSE;
    PCUP(2);
    while (1)
        ((int (*)())(*(bcbody_reg + bcpc_reg)))();
}

#ifndef SHE

void xspawn_thread()
{
    int bc_intrp_future2();
    SOBJ obj;
    FUTURE *fut, *mk_plchld();
    cthread_t child;
    REGS *regs;

    obj.obj = *sp_reg;
    if (GETtag(obj.obj) != LTAG_FUTURE){
        err_illegal_argument_type(obj.obj, "future object");
        return;
    }
    fut = (FUTURE *)(ADDRESS(obj.obj));
    regs = (REGS *)malloc(sizeof(REGS));
    context_cp(regs);
    mutex_lock(fut->lock);
    fut->body.obj = getarg(0);
    fut->env.obj = reg_reg;
    fut->start_addr.obj = mk_fix(GETbcarg(1));
#ifdef PTHREAD
    child = cthread_fork(setup_child, regs);
    cthread_detach(child);
#else
    cthread_detach(child = cthread_fork(setup_child, regs));
#endif
    mutex_lock(cthread_fork_lock);
    current_thread_no++;
    mutex_unlock(cthread_fork_lock);
    fut->proc_id = cthread_self();
#ifndef PTHREAD
    fut->thread_id = thread_self();
#endif
    fut->status = UNDETERMINED;
/*
    stack_top = (stack_bottom - CHILD_STACK_SIZE + 28) + CHILD_GETA_OF_STACK;
*/
    stack_top = (char *)malloc(CHILD_STACK_SIZE);
    stack_bottom = stack_top + CHILD_STACK_SIZE - 28;
    stack_top = stack_top + CHILD_GETA_OF_STACK;

    mutex_unlock(fut->lock);

    *((unsigned int *)stack_bottom + 1) = (unsigned int)LTAG_NIL;
    *((unsigned int *)stack_bottom + 2) = 0;
    *((unsigned int *)stack_bottom + 3) = 0;
    stack_area_bottom = stack_bottom;
    set_top(stack_bottom);
    s_push(top);
    s_push(0);
    s_push(0);
    set_base(top_val());
    set_active(top_val());
    set_a2(LTAG_NIL);
    set_reg(LTAG_NIL);

    old_top = (unsigned *)NULL;
    old_top1 = (unsigned *)NULL;
    reg1 = LTAG_NIL;
    reg2 = LTAG_NIL;
    reg3 = LTAG_NIL;
    call_cc_arg.obj = LTAG_NIL;
    top_of_aflk.obj = LTAG_NIL;
    end_of_aflk.obj = LTAG_NIL;
    af_tmp.obj = LTAG_NIL;
    more.obj = LTAG_NIL;
    optcnt.obj = LTAG_NIL;

    open_frame2(bc_intrp_future2, LTAG_FUTURE | SSERDDA(fut), 0);
    s_push(fut->body.obj);
    if (fut->env.obj != LTAG_NIL)
        s_push(fut->env.obj);
    active_reg = base_reg;
    bcpc_reg = fix_to_int(fut->start_addr.obj);
    bcbody_reg = (unsigned *)ADDRESS(((BCCLOS *)ADDRESS(fut->body.obj))->body.obj);
    return;
}

#endif

#ifndef SHE
void spawn_thread()
{
    reg_reg = mk_spawn(*sp_reg, getarg(0), reg_reg, GETbcarg(1));
    PCUP(2);
}
#endif

#endif

void reset_sp()
{
    sp_reg = sp_reg + (int)GETbcarg(1);
    PCUP(2);
}


void exec_sysfun()
{
    ((void (*)())GETbcarg(1))();
    xreturn();
}

void error_cxxdr(char *);

void get_car()
{
#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    switch (GETtag(reg_reg)) {
    case LTAG_CONS :
        set_reg(CARobj(reg_reg));
    case LTAG_NIL  :
        break;
    default :
        error_cxxdr("car");
    }
    PCUP(1);
}


void get_cdr()
{
#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    switch (GETtag(reg_reg)) {
    case LTAG_CONS :
        set_reg(CDRobj(reg_reg));
    case LTAG_NIL  :
        break;
    default :
        error_cxxdr("cdr");
    }
    PCUP(1);
}


void get_cadr()
{
    PCUP(1);
#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    switch (GETtag(reg_reg)) {
    case LTAG_CONS :
        set_reg(CDRobj(reg_reg));
        break;
    case LTAG_NIL  :
        return;
    default :
        error_cxxdr("cdr");
    }
#ifdef CTHREAD
    reg_reg = touch_obj(reg_reg);
#endif
    switch (GETtag(reg_reg)) {
    case LTAG_CONS :
        set_reg(CARobj(reg_reg));
    case LTAG_NIL  :
        break;
    default :
        error_cxxdr("car");
    }
}


void get_caar()
{
    PCUP(1);
#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    switch (GETtag(reg_reg)) {
    case LTAG_CONS :
        set_reg(CARobj(reg_reg));
        break;
    case LTAG_NIL  :
        return;
    default :
        error_cxxdr("car");
    }
#ifdef CTHREAD
    reg_reg = touch_obj(reg_reg);
#endif
    switch (GETtag(reg_reg)) {
    case LTAG_CONS :
        set_reg(CARobj(reg_reg));
    case LTAG_NIL  :
        break;
    default :
        error_cxxdr("car");
    }
}


void get_cdar()
{
    PCUP(1);
#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    switch (GETtag(reg_reg)) {
    case LTAG_CONS :
        set_reg(CARobj(reg_reg));
        break;
    case LTAG_NIL  :
        return;
    default :
        error_cxxdr("car");
    }
#ifdef CTHREAD
    reg_reg = touch_obj(reg_reg);
#endif
    switch (GETtag(reg_reg)) {
    case LTAG_CONS :
        set_reg(CDRobj(reg_reg));
    case LTAG_NIL  :
        break;
    default :
        error_cxxdr("cdr");
    }
}


void get_cddr()
{
    PCUP(1);
#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    switch (GETtag(reg_reg)) {
    case LTAG_CONS :
        set_reg(CDRobj(reg_reg));
        break;
    case LTAG_NIL  :
        return;
    default :
        error_cxxdr("cdr");
    }
#ifdef CTHREAD
    reg_reg = touch_obj(reg_reg);
#endif
    switch (GETtag(reg_reg)) {
    case LTAG_CONS :
        set_reg(CDRobj(reg_reg));
    case LTAG_NIL  :
        break;
    default :
        error_cxxdr("cdr");
    }
}


void error_cxxdr(name)
char *name;
{
    sbokkan1(name, reg_reg, "Argument of 'car' or 'cdr' must be a pair or ().");
}


void make_cons()
{
    reg_reg = mk_cons(SCPOP(), reg_reg);
    PCUP(1);
}


void make_cons1()
{
    reg_reg = mk_cons(GETconst(1), reg_reg);
    PCUP(2);
}


void make_cons2()
{
    reg_reg = mk_cons(reg_reg, GETconst(1));
    PCUP(2);
}


void ncons()
{
    reg_reg = mk_cons(reg_reg, LTAG_NIL);
    PCUP(1);
}


void pairp()
{
#ifdef TOUCH_ARG
    switch (GETtag(touch_obj(reg_reg))) {
#else
    switch (GETtag(reg_reg)) {
#endif
    case LTAG_CONS :
        reg_reg = LTAG_T;
        break;
    default :
        reg_reg = LTAG_FALSE;
    }
    PCUP(1);
}


void symbolp()
{
#ifdef TOUCH_ARG
    if (GETsym_tag(touch_obj(reg_reg)) == LTAG_SYMBOL)
#else
    if (GETsym_tag(reg_reg) == LTAG_SYMBOL)
#endif
        reg_reg = LTAG_T;
    else
        reg_reg = LTAG_FALSE;
    PCUP(1);
}


void nullp()
{
#ifdef TOUCH_ARG
    if (touch_obj(reg_reg) == LTAG_NIL)
#else
    if (reg_reg == LTAG_NIL)
#endif
        reg_reg = LTAG_T;
    else
        reg_reg = LTAG_FALSE;
    PCUP(1);
}


void notp()
{
#ifdef USE_FALSE

# ifdef IEEE_SPEC

#  ifdef TOUCH_ARG
    if (touch_obj(reg_reg) == LTAG_FALSE)
#  else
    if (reg_reg == LTAG_FALSE)
#  endif /* TOUCH_ARG */

# else  /* not IEEE_SPEC */
#  ifdef TOUCH_ARG
    if (touch_obj(reg_reg) == LTAG_FALSE || touch_obj(reg_reg) == LTAG_NIL)
#  else
    if (reg_reg == LTAG_FALSE || reg_reg == LTAG_NIL)
#  endif /* TOUCH_ARG */
# endif /* IEEE_SPEC */
#else  /* not USE_FALSE */
# ifdef TOUCH_ARG
    if (touch_obj(reg_reg) == LTAG_FALSE)
# else
    if (reg_reg == LTAG_FALSE)
# endif /* TOUCH_ARG */
#endif
        reg_reg = LTAG_T;
    else
        reg_reg = LTAG_FALSE;
    PCUP(1);
}


void zerop()
{
    unsigned tag;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(reg_reg) != LTAG_FLOAT)
        error_invalid_arg1("zero?");

    if (reg_reg == SCOBJ_0 || reg_reg == SCOBJ_SF0)
        reg_reg = LTAG_T;
    else
        reg_reg = LTAG_FALSE;
    PCUP(1);
}


void eqp()
{
#ifdef TOUCH_ARG
    if (touch_obj(reg_reg) == touch_obj(SCPOP()))
#else
    if (reg_reg == SCPOP())
#endif
        reg_reg = LTAG_T;
    else
        reg_reg = LTAG_FALSE;
    PCUP(1);
}


eqvp()
{
#ifdef TOUCH_ARG
    if (eqv(touch_obj(reg_reg), touch_obj(SCPOP())))
#else
    if (eqv(reg_reg, SCPOP()))
#endif
        reg_reg = LTAG_T;
    else
        reg_reg = LTAG_FALSE;
    PCUP(1);
}


void equalp()
{
    if (equal(reg_reg, SCPOP()))
        reg_reg = LTAG_T;
    else
        reg_reg = LTAG_FALSE;
    PCUP(1);
}


void eq_constp()
{
#ifdef TOUCH_ARG
    if (touch_obj(reg_reg) ==  GETconst(1))
#else
    if (reg_reg ==  GETconst(1))
#endif
        reg_reg = LTAG_T;
    else
        reg_reg = LTAG_FALSE;
    PCUP(2);
}


void eqv_constp()
{
#ifdef TOUCH_ARG
    if (eqv(touch_obj(reg_reg), GETconst(1)))
#else
    if (eqv(reg_reg, GETconst(1)))
#endif
        reg_reg = LTAG_T;
    else
        reg_reg = LTAG_FALSE;
    PCUP(2);
}


void equal_constp()
{
    if (equal(reg_reg, GETconst(1)))
        reg_reg = LTAG_T;
    else
        reg_reg = LTAG_FALSE;
    PCUP(2);
}

void error_invalid_arg1(name)
char *name;
{
    sbokkan1(name, reg_reg, "Illegal argument.");
}

void error_invalid_arg2(name, arg)
char *name;
unsigned arg;
{
    sbokkan2(name, arg, reg_reg, "Illegal arguments.");
}

void error_invalid_arg3(name, arg1, arg2, arg3)
char *name;
unsigned arg1, arg2, arg3;
{
    sbokkan3(name, arg1, arg2, arg3, "Illegal arguments.");
}

void add()                           /* T.Komiya */
{ 
    unsigned tag;
    SOBJ obj;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(reg_reg) != LTAG_FLOAT)
        error_invalid_arg2("+", *top_val());

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    tag = GETnum_tag(obj.obj);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(obj.obj) != LTAG_FLOAT)
        error_invalid_arg2("+", obj.obj);

    set_reg(add_obj(reg_reg, obj.obj));
    PCUP(1);
}

void mul()                           /* T.Komiya */
{ 
    unsigned tag;
    SOBJ obj;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(reg_reg) != LTAG_FLOAT)
        error_invalid_arg2("*", *top_val());

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    tag = GETnum_tag(obj.obj);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(obj.obj) != LTAG_FLOAT)
        error_invalid_arg2("*", obj.obj);

    set_reg(mul_obj(reg_reg, obj.obj));
    PCUP(1);
}

void sub()                           /* T.Komiya */
{ 
    unsigned tag;
    SOBJ obj;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(reg_reg) != LTAG_FLOAT)
        error_invalid_arg2("-", *top_val());

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    tag = GETnum_tag(obj.obj);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(obj.obj) != LTAG_FLOAT)
        error_invalid_arg2("-", obj.obj);

    set_reg(sub_obj(obj.obj, reg_reg));
    PCUP(1);
}

void minus()                         /* T.Komiya */
{
    void error_invalid_arg1(char *);
    unsigned tag;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(reg_reg) != LTAG_FLOAT)
        error_invalid_arg1("-");

    set_reg(minus_obj(reg_reg));
    PCUP(1);
}

void tus_div()                           /* T.Komiya */
{
    unsigned tag;
    SOBJ obj;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(reg_reg) != LTAG_FLOAT)
        error_invalid_arg2("/", *top_val());

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    tag = GETnum_tag(obj.obj);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(obj.obj) != LTAG_FLOAT)
        error_invalid_arg2("/", obj.obj);

    if (reg_reg == SCOBJ_0 || reg_reg == SCOBJ_SF0)
        error_invalid_arg2("/", obj.obj);
    set_reg(div_obj(obj.obj, reg_reg));
    PCUP(1);
}

void inverse()                       /* T.Komiya */
{
    unsigned tag;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(reg_reg) != LTAG_FLOAT)
        error_invalid_arg1("/");

    set_reg(div_obj(SCOBJ_1, reg_reg));
    PCUP(1);
}

void add1()                          /* T.Komiya */
{
    unsigned tag;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag == LTAG_NUMBER1 || tag == LTAG_NUMBER3)
        set_reg(add_obj(reg_reg, SCOBJ_1));
    else if (tag == LTAG_NUMBER2 || GETtag(reg_reg) == LTAG_FLOAT)
        set_reg(add_obj(reg_reg, mk_flt(1.0)));
    else
        error_invalid_arg1("1+");
    PCUP(1);
}

void sub1()                          /* T.Komiya */
{
    unsigned tag;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag == LTAG_NUMBER1 || tag == LTAG_NUMBER3)
        set_reg(sub_obj(reg_reg, SCOBJ_1));
    else if (tag == LTAG_NUMBER2 || GETtag(reg_reg) == LTAG_FLOAT)
        set_reg(sub_obj(reg_reg, mk_flt(1.0)));
    else
        error_invalid_arg1("1-");
    PCUP(1);
}

void lessthan2()                     /* T.Komiya */
{
    unsigned tag;
    SOBJ obj;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && GETtag(reg_reg) != LTAG_FLOAT
        && (tag != LTAG_NUMBER3 || GETtag(reg_reg) == LTAG_COMPLEX))
        error_invalid_arg2("<", *top_val());

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    tag = GETnum_tag(obj.obj);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && GETtag(obj.obj) != LTAG_FLOAT
        && (tag != LTAG_NUMBER3 || GETtag(obj.obj) == LTAG_COMPLEX))
        error_invalid_arg2("<", obj.obj);

    if (comp_obj(obj.obj, reg_reg, 1) & GREATERE)
        set_reg(LTAG_FALSE);
    else
        set_reg(LTAG_T);
    PCUP(1);
}

void greaterthan2()                  /* T.Komiya */
{
    unsigned tag;
    SOBJ obj;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && GETtag(reg_reg) != LTAG_FLOAT
        && (tag != LTAG_NUMBER3 || GETtag(reg_reg) == LTAG_COMPLEX))
        error_invalid_arg2(">", *top_val());

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    tag = GETnum_tag(obj.obj);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && GETtag(obj.obj) != LTAG_FLOAT
        && (tag != LTAG_NUMBER3 || GETtag(obj.obj) == LTAG_COMPLEX))
        error_invalid_arg2(">", obj.obj);

    if (comp_obj(obj.obj, reg_reg, 1) & LESSE)
        set_reg(LTAG_FALSE);
    else
        set_reg(LTAG_T);
    PCUP(1);
}

void lep2()                          /* T.Komiya */
{
    unsigned tag;
    SOBJ obj;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && GETtag(reg_reg) != LTAG_FLOAT
        && (tag != LTAG_NUMBER3 || GETtag(reg_reg) == LTAG_COMPLEX))
        error_invalid_arg2("<=", *top_val());

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    tag = GETnum_tag(obj.obj);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && GETtag(obj.obj) != LTAG_FLOAT
        && (tag != LTAG_NUMBER3 || GETtag(obj.obj) == LTAG_COMPLEX))
        error_invalid_arg2("<=", obj.obj);

    if (comp_obj(obj.obj, reg_reg, 1) & GREATERT)
        set_reg(LTAG_FALSE);
    else
        set_reg(LTAG_T);
    PCUP(1);
}

void gep2()                          /* T.Komiya */
{
    unsigned tag;
    SOBJ obj;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && GETtag(reg_reg) != LTAG_FLOAT
        && (tag != LTAG_NUMBER3 || GETtag(reg_reg) == LTAG_COMPLEX))
        error_invalid_arg2(">=", *top_val());

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    tag = GETnum_tag(obj.obj);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && GETtag(obj.obj) != LTAG_FLOAT
        && (tag != LTAG_NUMBER3 || GETtag(obj.obj) == LTAG_COMPLEX))
        error_invalid_arg2(">=", obj.obj);

    if (comp_obj(obj.obj, reg_reg, 1) & LESST)
        set_reg(LTAG_FALSE);
    else
        set_reg(LTAG_T);
    PCUP(1);
}

void numbereq2()                     /* T.Komiya */
{
    unsigned tag;
    SOBJ obj;
    int cplx = 0;

#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    tag = GETnum_tag(reg_reg);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(reg_reg) != LTAG_FLOAT)
        error_invalid_arg2("=", *top_val());

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    tag = GETnum_tag(obj.obj);
    if (tag != LTAG_NUMBER1 && tag != LTAG_NUMBER2
        && tag != LTAG_NUMBER3 && GETtag(obj.obj) != LTAG_FLOAT)
        error_invalid_arg2("=", obj.obj);

    if (GETtag(reg_reg) == LTAG_COMPLEX) ++cplx;
    if (GETtag(obj.obj) == LTAG_COMPLEX) ++cplx;

    if (cplx == 0)
        if (comp_obj(reg_reg, obj.obj, 1) == EQUAL)
            set_reg(LTAG_T);
        else
            set_reg(LTAG_FALSE);
    else if (cplx == 2)
        if (comp_cplx(reg_reg, obj.obj) == EQUAL)
            set_reg(LTAG_T);
        else
            set_reg(LTAG_FALSE);
    else
        set_reg(LTAG_FALSE);
    PCUP(1);
}

void setcar()                        /* T.Komiya */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(SCPOP());
#else
    pair.obj = SCPOP();
#endif
    if (GETtag(pair.obj) != LTAG_CONS)
        error_invalid_arg2("set-car!", pair.obj);
    set_car(pair.obj, reg_reg);
    set_reg(pair.obj);
    PCUP(1);
}

void setcdr()                        /* T.Komiya */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(SCPOP());
#else
    pair.obj = SCPOP();
#endif
    if (GETtag(pair.obj) != LTAG_CONS)
        error_invalid_arg2("set-cdr!", pair.obj);
    set_cdr(pair.obj, reg_reg);
    set_reg(pair.obj);
    PCUP(1);
}

void vectorref()                     /* T.Komiya */
{ 
    SOBJ obj;
    int index, tag;

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    if (GETtag(obj.obj) != LTAG_VECTOR)
        error_invalid_arg2("vector-ref", obj.obj);
#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    if ((tag = GETtag(reg_reg)) != LTAG_FIXNUM && tag != LTAG_FIXNUM1)
        error_invalid_arg2("vector-ref", obj.obj);
    index = fix_to_int(reg_reg);
    if (nonnil_vector_length(obj.obj) <= index)
        error_invalid_arg2("vector-ref", obj.obj);

    set_reg(*((unsigned *)(ADDRESS(obj.obj) + (index + 1) * 4)));
    PCUP(1);
}

void vectorset()                     /* T.Komiya */
{ 
    SOBJ obj;
    int index, tag;

#ifdef TOUCH_ARG
    obj.obj = touch_obj(SCPOP());
#else
    obj.obj = SCPOP();
#endif
    if (GETtag(obj.obj) != LTAG_VECTOR)
        error_invalid_arg3("vector-set!", obj.obj, reg_reg, *top_val());
#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    if ((tag = GETtag(reg_reg)) != LTAG_FIXNUM && tag != LTAG_FIXNUM1)
        error_invalid_arg3("vector-set!", obj.obj, reg_reg, *top_val());
    index = fix_to_int(reg_reg);
    if (nonnil_vector_length(obj.obj) <= index)
        error_invalid_arg3("vector-set!", obj.obj, reg_reg, *top_val());

    *((unsigned *)(ADDRESS(obj.obj) + (index + 1) * 4)) = SCPOP();
    set_reg(obj.obj);
    PCUP(1);
}

void vectorp()                       /* T.Komiya */
{
#ifdef TOUCH_ARG
    if (GETtag(touch_obj(reg_reg)) == LTAG_VECTOR
        || GETtag(touch_obj(reg_reg)) == LTAG_NILVECTOR)
#else
    if (GETtag(reg_reg) == LTAG_VECTOR || GETtag(reg_reg) == LTAG_NILVECTOR)
#endif
        set_reg(LTAG_T);
    else
        set_reg(LTAG_FALSE);
    PCUP(1);
}

void vectorlen()                     /* T.Komiya */
{ 
#ifdef TOUCH_ARG
    reg_reg = touch_obj(reg_reg);
#endif
    if (GETtag(reg_reg) != LTAG_VECTOR)
        if (GETtag(reg_reg) != LTAG_NILVECTOR)
            error_invalid_arg1("vector-length");
        else
            set_reg(SCOBJ_0);
    else
        set_reg(mk_fix(nonnil_vector_length(reg_reg)));
    PCUP(1);
}

/*
 *   ----
 *   misc
 *   ----
 */

void iretaddrx()
{
    bcpc_reg = SCPOP();
    bcbody_reg = (unsigned *)ADDRESS(SCPOP());
}

#ifdef CALLIOC
void iretaddrep()
{
    bcpc_reg = SCPOP();
    bcbody_reg = (unsigned *)ADDRESS(*((unsigned *)ADDRESS(SCPOP())+1));
    ((unsigned *)stack_bottom)[5]--;
}

void iretaddrx_tr()
{
    ((unsigned *)stack_bottom)[5]--;
    xreturn();
}
#endif

void iretaddrl()
{
    bcpc_reg = SCPOP();
}


void s_iretaddrx()
{
    bcbody_reg = (unsigned *)ADDRESS( *(((unsigned *)ADDRESS(*(active_reg - 1))) + 
        5) );
    bcpc_reg = SCPOP();
}


void top()
{
    SOBJ actframe;
    unsigned *stackobj;
    int fsize;
    int active;

    if (more.obj != LTAG_NIL) {
#ifdef CALLIOC
        actframe.obj = CARobj(CARobj(more.obj));
#else
        actframe.obj = CARobj(more.obj);
#endif
        more.obj = CDRobj(more.obj);
        sp_reg = sp_reg - 3;
        active_reg = sp_reg;
        base_reg = sp_reg;
        while (1) {
            stackobj = (unsigned *)ADDRESS(CARobj(actframe.obj));
            fsize = (0x0ffffff & (*stackobj >> 2)) - 2;
            stackobj++;
            for (; fsize > 0; fsize--)
                s_push(*(stackobj++));
            s_push((int)(fix_to_int(*(stackobj++)))); /* pc */
#if defined(__ELF__) && defined(i386)
            s_push(ADDRESS(fix_to_int(*(stackobj++)))); /* ret */
#else
            s_push(fix_to_int(*(stackobj++))); /* ret */
#endif
            if ((actframe.obj = CDRobj(actframe.obj)) == LTAG_NIL)
                break;
            active = (int)IMM(CARobj(actframe.obj)) + (unsigned)sp_reg;
            s_push(active);
            s_push(base_reg);
            base_reg = sp_reg;
            actframe.obj = CDRobj(actframe.obj);
        }
        ((int (*)())s_pop())();
    } else {
#ifdef CALLIOC
        extern system_function_loading_flag;
        if (system_function_loading_flag == OFF)
# if defined(__CYGWIN32__) || defined(linux)
            siglongjmp(exit_bc_interpreter, 3);
# else
            longjmp(exit_bc_interpreter, 3);
# endif
        else
#endif
#if defined(__CYGWIN32__) || defined(linux)
        siglongjmp(exit_bc_interpreter, 1);
#else
        longjmp(exit_bc_interpreter, 1);
#endif
    }
}
