/*
 * bignum.c - 
 */

/*
 * $Id: bignum.c,v 1.3 2014/10/28 02:22:55 komiya Exp $
 */

/************************************/
/*                  */
/*  File Name : bignum.c            */
/*  Update    : Yass.  Sep. 11 '90  */
/*              Yass.  Sep.  4 '90  */
/*              Yass.  Sep.  3 '90  */
/*              Yass.  Aug. 31 '90  */
/*              Yass.  Aug. 30 '90  */
/*              Yass.  Aug. 29 '90  */
/*              Yass.  Aug. 28 '90  */
/*              Yass.  Aug. 27 '90  */
/*              Yass.  Aug. 24 '90  */
/*              Yass.  Aug. 23 '90  */
/*                                  */
/*    Bignum operation routines     */
/*                      */
/************************************/

static char *RCSid="$Id: bignum.c,v 1.3 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"


/*  Written by Yass. Sep. 4 '90  */
int stretch_big(obj, i)
int obj,  /* any bignum or nil -> update */
i;    /* any int */
{ 
    SOBJ  new_obj;

    s_push(obj);
    new_obj.obj = mk_bign(i);
    obj = s_pop();
    if (obj != LTAG_NIL)
        ((BIGNUM * )ADDRESS(obj))->next.obj = new_obj.obj;
    return(new_obj.obj);
}


/*  Written by Yass. Aug. 31 '90  */
void normalize_big(obj)
int obj;  /*  any bignum -> normalized  */
{
    if (((BIGNUM * )ADDRESS(obj))->next.obj == LTAG_NIL)
        return;
    normalize_big(((BIGNUM * )ADDRESS(obj))->next.obj);
    if (((BIGNUM * )ADDRESS(((BIGNUM * )ADDRESS(obj))->next.obj))->next.obj == LTAG_NIL) { 
        if (((BIGNUM * )ADDRESS(((BIGNUM * )ADDRESS(obj))->next.obj))->val == 0)
            ((BIGNUM * )ADDRESS(obj))->next.obj = LTAG_NIL;
        else
         { 
            if (((BIGNUM * )ADDRESS(((BIGNUM * )ADDRESS(obj))->next.obj))->val == -1) { 
                ((BIGNUM * )ADDRESS(obj))->next.obj = LTAG_NIL;
                ((BIGNUM * )ADDRESS(obj))->val |= ~INT_MASK;
            }
        }
    }
}


/*  Written by Yass. Aug. 31 '90  */
int big_length(obj)
int obj;  /* any bignum  */
{ 
    int i;

    for (i = 1 ; ((BIGNUM * )ADDRESS(obj))->next.obj != LTAG_NIL ; i++)
        obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
    return(i);
}


/*  Written by Yass. Aug. 27 '90  */
int copy_big(obj)
int obj;  /* any bignum */
{ 
    SOBJ new_obj;

    s_push(obj);
    new_obj.obj = mk_bign(((BIGNUM * )ADDRESS(obj))->val);
    obj = ((BIGNUM * )ADDRESS(s_pop()))->next.obj;
    s_push(new_obj.obj);
    while (obj != LTAG_NIL) { 
        s_push(((BIGNUM * )ADDRESS(obj))->next.obj);
        new_obj.obj = stretch_big(new_obj.obj, ((BIGNUM * )ADDRESS(obj))->val);
        obj = s_pop();
    }
    return(s_pop());
}


/*  Written by Yass. Aug. 27 '90  */
int big_zerop(obj)
int obj;  /* any bignum */
{
    for (; ; ) { 
        if (((BIGNUM * )ADDRESS(obj))->val != 0)
            return(FALSE);
        else
         { 
            obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
            if (GETtag(obj) == LTAG_NIL)
                return(TRUE);
        }
    }
}


/*  Written by Yass. Aug. 27 '90  */
int big_sign(obj)
int obj;  /* any bignum */
{
    while (((BIGNUM * )ADDRESS(obj))->next.obj != LTAG_NIL)
        obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
    if (((BIGNUM * )ADDRESS(obj))->val < 0)
        return(-1);
    else
        return(1);
}


/*  Written by Yass. Sep. 4 '90  */
int big_compare(obj1, obj2)
int obj1,  /* any bignum */
obj2;  /* any bignum */
{ 
    int result = EQUAL,
    val1, val2;

    while (1) { 
        if (((BIGNUM * )ADDRESS(obj1))->next.obj == LTAG_NIL) { 
            val1 = ((BIGNUM * )ADDRESS(obj1))->val;
            val2 = ((BIGNUM * )ADDRESS(obj2))->val;
            if (((BIGNUM * )ADDRESS(obj2))->next.obj == LTAG_NIL) { 
                if (val1 < val2)
                    return(LESST);
                else if (val1 == val2)
                    return(result);
                else
                    return(GREATERT);
            } else if (big_zerop(((BIGNUM * )ADDRESS(obj2))->next.obj)) { 
                if (val1 < val2)
                    return(LESST);
                else if (val1 == val2)
                    return(result);
                else
                    return(GREATERT);
            } else if (big_sign(((BIGNUM * )ADDRESS(obj2))->next.obj) > 0)
                return(LESST);
            else
                return(GREATERT);
        } else if (((BIGNUM * )ADDRESS(obj2))->next.obj == LTAG_NIL) { 
            if (big_zerop(((BIGNUM * )ADDRESS(obj1))->next.obj)) { 
                val1 = ((BIGNUM * )ADDRESS(obj1))->val;
                val2 = ((BIGNUM * )ADDRESS(obj2))->val;
                if (val1 < val2)
                    return(LESST);
                else if (val1 == val2)
                    return(result);
                else
                    return(GREATERT);
            } else if (big_sign(((BIGNUM * )ADDRESS(obj1))->next.obj) < 0)
                return(LESST);
            else
                return(GREATERT);
        } else
         { 
            val1 = ((BIGNUM * )ADDRESS(obj1))->val;
            val2 = ((BIGNUM * )ADDRESS(obj2))->val;
            if (val1 < val2)
                result = LESST;
            else if (val1 > val2)
                result = GREATERT;
            obj1 = ((BIGNUM * )ADDRESS(obj1))->next.obj;
            obj2 = ((BIGNUM * )ADDRESS(obj2))->next.obj;
        }
    }
}


/*  Written by Yass. Aug. 27 '90  */
int complement_big(obj)
int obj;  /* any bignum -> updated,returned */
{
    s_push(obj);
    while (((BIGNUM * )ADDRESS(obj))->next.obj != LTAG_NIL) { 
        if (((BIGNUM * )ADDRESS(obj))->val != 0) { 
            ((BIGNUM * )ADDRESS(obj))->val = (-(((BIGNUM * )ADDRESS(obj))->val)) & INT_MASK;
            goto CON;
        } else
            obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
    }
    if (((BIGNUM * )ADDRESS(obj))->val == ~INT_MASK) { 
        ((BIGNUM * )ADDRESS(obj))->val = 0;
        obj = stretch_big(obj, 1);
    } else
        ((BIGNUM * )ADDRESS(obj))->val = -(((BIGNUM * )ADDRESS(obj))->val);
    return(s_pop());

CON:

    for (; ; ) { 
        obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
        if (((BIGNUM * )ADDRESS(obj))->next.obj == LTAG_NIL)
            break;
        ((BIGNUM * )ADDRESS(obj))->val = (~(((BIGNUM * )ADDRESS(obj))->val)) & INT_MASK;
    }
    ((BIGNUM * )ADDRESS(obj))->val = ~(((BIGNUM * )ADDRESS(obj))->val);
    return(s_pop());
}


/*  Written by Yass. Aug. 27 '90  */
int minus_big(obj)
int obj;  /* any bignum */
{ 
    SOBJ new_obj;

    s_push(obj);
    new_obj.obj = mk_bign(0);
    obj = s_pop();
    s_push(new_obj.obj);
    while (((BIGNUM * )ADDRESS(obj))->next.obj != LTAG_NIL) { 
        if (((BIGNUM * )ADDRESS(obj))->val != 0) { 
            ((BIGNUM * )ADDRESS(new_obj.obj))->val = (-(((BIGNUM * )ADDRESS(obj))->val)) & INT_MASK;
            goto CON;
        }
        s_push(((BIGNUM * )ADDRESS(obj))->next.obj);
        new_obj.obj = stretch_big(new_obj.obj, 0);
        obj = s_pop();
    }
    if (((BIGNUM * )ADDRESS(obj))->val == ~INT_MASK) { 
        ((BIGNUM * )ADDRESS(new_obj.obj))->val = 0;
        stretch_big(new_obj.obj, 1);
    } else
        ((BIGNUM * )ADDRESS(new_obj.obj))->val = -(((BIGNUM * )ADDRESS(obj))->val);
    return(s_pop());

CON:

    for (; ; ) { 
        s_push(((BIGNUM * )ADDRESS(obj))->next.obj);
        new_obj.obj = stretch_big(new_obj.obj, 0);
        obj = s_pop();
        if (((BIGNUM * )ADDRESS(obj))->next.obj == LTAG_NIL)
            break;
        ((BIGNUM * )ADDRESS(new_obj.obj))->val = (~(((BIGNUM * )ADDRESS(obj))->val)) & INT_MASK;
    }
    ((BIGNUM * )ADDRESS(new_obj.obj))->val = ~(((BIGNUM * )ADDRESS(obj))->val);
    return(s_pop());
}


/*  Written by Yass. Sep. 3 '90  */
int add_int_big(obj, i)
int obj,  /* any bignum -> updated,returned */
i;    /* any int  */
{
    if (i < 0)
        return(sub_int_big(obj, -i));

    s_push(obj);
    while (((BIGNUM * )ADDRESS(obj))->next.obj != LTAG_NIL) { 
        ((BIGNUM * )ADDRESS(obj))->val += i;
        if (((BIGNUM * )ADDRESS(obj))->val < 0)  /* carry */ { 
            i = 1;
            ((BIGNUM * )ADDRESS(obj))->val &= INT_MASK;
            obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
        } else
         { 
            obj = s_pop();
            normalize_big(obj);
            return(obj);
        }
    }
    if (((BIGNUM * )ADDRESS(obj))->val >= 0) { 
        ((BIGNUM * )ADDRESS(obj))->val += i;
        if (((BIGNUM * )ADDRESS(obj))->val < 0)  /* overflow */ { 
            ((BIGNUM * )ADDRESS(obj))->val &= INT_MASK;
            stretch_big(obj, 1);
        }
    } else
        ((BIGNUM * )ADDRESS(obj))->val += i;
    obj = s_pop();
    normalize_big(obj);
    return(obj);
}


/*  Written by Yass. Sep. 3 '90  */
int sub_int_big(obj, i)
int obj,  /* any bignum -> updated,returned */
i;    /* any int  */
{
    if (i < 0)
        return(add_int_big(obj, -i));

    s_push(obj);
    while (((BIGNUM * )ADDRESS(obj))->next.obj != LTAG_NIL) { 
        ((BIGNUM * )ADDRESS(obj))->val -= i;
        if (((BIGNUM * )ADDRESS(obj))->val < 0)  /* borrow */ { 
            i = 1;
            ((BIGNUM * )ADDRESS(obj))->val &= INT_MASK;
            obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
        } else
         { 
            obj = s_pop();
            normalize_big(obj);
            return(obj);
        }

    }
    if (((BIGNUM * )ADDRESS(obj))->val < 0) { 
        ((BIGNUM * )ADDRESS(obj))->val -= i;
        if (((BIGNUM * )ADDRESS(obj))->val >= 0) { 
            ((BIGNUM * )ADDRESS(obj))->val &= INT_MASK;
            stretch_big(obj, -2);
        }
    } else
        ((BIGNUM * )ADDRESS(obj))->val -= i;
    obj = s_pop();
    normalize_big(obj);
    return(obj);
}


/*  Written by Yass. Aug. 30 '90  */
int mul_int_big(obj, i)
int obj,  /* any bignum -> updated,returned */
i;    /* any int  */
{ 
    int h = 0,
    sign = 1;

    if (i < 0) { 
        i = -i;
        if (big_sign(obj) < 0)
            obj = complement_big(obj);
        else
            sign = -1;
    } else
     { 
        if (big_sign(obj) < 0) { 
            obj = complement_big(obj);
            sign = -1;
        }
    }

    s_push(obj);
    for (; ; ) { 
        extended_mul(i, ((BIGNUM * )ADDRESS(obj))->val, h, &h, &(((BIGNUM * )ADDRESS(obj))->val));
        if (((BIGNUM * )ADDRESS(obj))->next.obj == LTAG_NIL)
            break;
        obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
    }
    if (h > 0)
        stretch_big(obj, h);

    if (sign < 0)
        return(complement_big(s_pop()));
    else
        return(s_pop());
}


/*  Written by Yass. Sep. 3 '90  */
int div_int_big(obj, i, r)
int obj,  /* any bignum -> updated,returned */
i,    /* any int  */
*r;   /* returned any remainder (int) */
{ 
/*********************
    int sign1 = 1,
    sign2 = 1;
**********************/

    if (i == 0){
	err_division_by_zero();  /*  By N.Watanabe Nov. 04, 1992  */
            /*  bokkan("0 divide.");  */
    }
/*********************
    if (i < 0) { 
        i = -i;
        if (big_sign(obj) < 0) { 
            obj = complement_big(obj);
            sign2 = -1;
        } else
            sign1 = -1;
    } else
     { 
        if (big_sign(obj) < 0) { 
            obj = complement_big(obj);
            sign1 = sign2 = -1;
        }
    }
**********************/
    if (((BIGNUM * )ADDRESS(obj))->next.obj == LTAG_NIL) { 
        *r = ((BIGNUM * )ADDRESS(obj))->val % i;
        ((BIGNUM * )ADDRESS(obj))->val /= i;
    } else
     { 
        div_int_big(((BIGNUM * )ADDRESS(obj))->next.obj, i, r);
        extended_div(i, *r, ((BIGNUM * )ADDRESS(obj))->val, &(((BIGNUM * )ADDRESS(obj))->val), r);
    }
/*********************
    if (sign2 < 0)
        *r = -(*r);
**********************/
    normalize_big(obj);
/*********************
    if (sign1 < 0)
        return(complement_big(obj));
    else  ************/
        return(obj);
}


/*  Written by Yass. Aug. 23 '90  */
extended_mul(d, q, r, hp, lp)  /* <copy from kcl> */
int d, q, r, *hp, *lp;
{ 
    int ld, hd, lq, hq, m, lm, hm, lr, hz, lz;

    ld = (d & 0x0000ffff);
    hd = (d >> 16);
    hd = (hd & 0x00007fff);
    lq = (q & 0x0000ffff);
    hq = (q >> 16);
    hq = (hq & 0x00007fff);
    lr = r;

    m = ld * lq;
    hm = (m & 0x80000000);
    hm = (hm >> 16);
    hm = (hm & 0x00008000);
    m = (m & 0x7fffffff);

    m = m + lr;

    lz = (m & 0x0000ffff);

    m = (m >> 16);
    m = (m & 0x0000ffff);
    m = m + hm;
    m = m + (lq * hd);
    hm = (m & 0x80000000);
    hm = (hm >> 15);
    hm = (hm & 0x0001ffff);
    m = (m & 0x7fffffff);

    m = m + (hq * ld);

    lm = (m & 0x00007fff);

    lz = lz + (lm << 16);

    m = (m >> 15);
    m = (m & 0x0001ffff);
    m = m + hm;

    hz = m + (hq * hd * 2);

    *hp = hz;
    *lp = lz;
}


/*  Written by Yass. Aug. 23 '90  */
extended_div(d, h, l, qp, rp)/* <copy from kcl> */
int d, h, l, *qp, *rp;
{ 
    int lh, ld, ll, i, ans;

    ld = d;
    lh = h;
    ll = l;

    ans = 0;
    for (i = 31; i > 0; i--) { 
        ans = (ans << 1);
        lh = (lh << 1);
        ll = (ll << 1);
        if (ll < 0)
            lh += 1;
        if (lh >= d) { 
            lh -= ld;
            ans += 1;
        }
        if (lh < 0) { 
            lh = (0x7fffffff - ld) + (lh & 0x7fffffff) + 1;
            ans += 1;
        }
    }

    *qp = ans;
    *rp = lh;
}


/*  Written by Yass. Sep. 3 '90  */
int add_big_big(x, y)
int x,  /* any bignum */
y;  /* any bignum */
{ 
    SOBJ new_obj;
    int c = 0;

    s_push(x);  
    s_push(y);
    new_obj.obj = mk_bign(0);
    y = s_pop();  
    x = s_pop();
    s_push(new_obj.obj);
    s_push(x);  
    s_push(y);
    for (; ; ) {   
        y = s_pop();  
        x = s_pop();
        ((BIGNUM * )ADDRESS(new_obj.obj))->val = 
            ((BIGNUM * )ADDRESS(x))->val + ((BIGNUM * )ADDRESS(y))->val + c;
        if (((BIGNUM * )ADDRESS(x))->next.obj == LTAG_NIL) { 
            if (((BIGNUM * )ADDRESS(y))->next.obj == LTAG_NIL)
                goto BOTH_END;
            else
                goto X_END;
        } else if (((BIGNUM * )ADDRESS(y))->next.obj == LTAG_NIL)
            goto Y_END;
        if (((BIGNUM * )ADDRESS(new_obj.obj))->val < 0) { 
            c = 1;
            ((BIGNUM * )ADDRESS(new_obj.obj))->val &= INT_MASK;
        } else
            c = 0;
        s_push(((BIGNUM * )ADDRESS(x))->next.obj);
        s_push(((BIGNUM * )ADDRESS(y))->next.obj);
        new_obj.obj = stretch_big(new_obj.obj, 0);
    }

BOTH_END:

    if (((BIGNUM * )ADDRESS(x))->val >= 0 && ((BIGNUM * )ADDRESS(y))->val >= 0 && 
        ((BIGNUM * )ADDRESS(new_obj.obj))->val < 0                              ) { 
        ((BIGNUM * )ADDRESS(new_obj.obj))->val &= INT_MASK;
        stretch_big(new_obj.obj, 1);
    } else if (((BIGNUM * )ADDRESS(x))->val < 0 && ((BIGNUM * )ADDRESS(y))->val < 0 && 
        ((BIGNUM * )ADDRESS(new_obj.obj))->val >= 0                           )
        stretch_big(new_obj.obj, -2);
    goto FIN;

X_END:

    if (((BIGNUM * )ADDRESS(x))->val >= 0)
        c = 1;
    else
        c = -1;
    for (; ; ) { 
        if (((BIGNUM * )ADDRESS(new_obj.obj))->val < 0)
            ((BIGNUM * )ADDRESS(new_obj.obj))->val &= INT_MASK;
        else
         { 
            y = ((BIGNUM * )ADDRESS(y))->next.obj;
            while (y != LTAG_NIL) { 
                s_push(((BIGNUM * )ADDRESS(y))->next.obj);
                new_obj.obj = stretch_big(new_obj.obj, ((BIGNUM * )ADDRESS(y))->val);
                y = s_pop();
            }
            break;
        }
        s_push(((BIGNUM * )ADDRESS(y))->next.obj);
        new_obj.obj = stretch_big(new_obj.obj, 0);
        y = s_pop();
        ((BIGNUM * )ADDRESS(new_obj.obj))->val = ((BIGNUM * )ADDRESS(y))->val + c;
        if (((BIGNUM * )ADDRESS(y))->next.obj == LTAG_NIL) { 
            if (c >= 0 && ((BIGNUM * )ADDRESS(y))->val >= 0 && 
                ((BIGNUM * )ADDRESS(new_obj.obj))->val < 0    ) { 
                ((BIGNUM * )ADDRESS(new_obj.obj))->val &= INT_MASK;
                stretch_big(new_obj.obj, 1);
            } else if (c < 0 && ((BIGNUM * )ADDRESS(y))->val < 0 && 
                ((BIGNUM * )ADDRESS(new_obj.obj))->val >= 0 )
                stretch_big(new_obj.obj, -2);
            break;
        }
    }
    goto FIN;

Y_END:

    if (((BIGNUM * )ADDRESS(y))->val >= 0)
        c = 1;
    else
        c = -1;
    for (; ; ) { 
        if (((BIGNUM * )ADDRESS(new_obj.obj))->val < 0)
            ((BIGNUM * )ADDRESS(new_obj.obj))->val &= INT_MASK;
        else
         { 
            x = ((BIGNUM * )ADDRESS(x))->next.obj;
            while (x != LTAG_NIL) { 
                s_push(((BIGNUM * )ADDRESS(x))->next.obj);
                new_obj.obj = stretch_big(new_obj.obj, ((BIGNUM * )ADDRESS(x))->val);
                x = s_pop();
            }
            break;
        }
        s_push(((BIGNUM * )ADDRESS(x))->next.obj);
        new_obj.obj = stretch_big(new_obj.obj, 0);
        x = s_pop();
        ((BIGNUM * )ADDRESS(new_obj.obj))->val = ((BIGNUM * )ADDRESS(x))->val + c;
        if (((BIGNUM * )ADDRESS(x))->next.obj == LTAG_NIL) { 
            if (c >= 0 && ((BIGNUM * )ADDRESS(x))->val >= 0 && 
                ((BIGNUM * )ADDRESS(new_obj.obj))->val < 0    ) { 
                ((BIGNUM * )ADDRESS(new_obj.obj))->val &= INT_MASK;
                stretch_big(new_obj.obj, 1);
            } else if (c < 0 && ((BIGNUM * )ADDRESS(x))->val < 0 && 
                ((BIGNUM * )ADDRESS(new_obj.obj))->val >= 0 )
                stretch_big(new_obj.obj, -2);
            break;
        }
    }

FIN:

    new_obj.obj = s_pop();
    normalize_big(new_obj.obj);
    return(new_obj.obj);
}


/*  Written by Yass. Aug. 28 '90  */
int sub_big_big(x, y)
int x,  /* any bignum */
y;  /* any bignum */
{
    s_push(x);
    y = minus_big(y);
    return(add_big_big(s_pop(), y));
}


/*  Written by Yass. Sep. 3 '90  */
int mul_big_big(x, y)
int x,  /* any bignum */
y;  /* any bignum */
{ 
    SOBJ new_obj, tmp_obj;
    int sign, i, h, l, tmp;

    sign = big_sign(x) * big_sign(y);
    s_push(y);
    if (big_sign(x) < 0)  /*  x < 0  */
        x = minus_big(x);
    y = s_pop();
    s_push(x);
    if (big_sign(y) < 0)  /*  y < 0  */
        y = minus_big(y);
    s_push(y);

    tmp_obj.obj = new_obj.obj = mk_bign(0);
    y = s_pop();  
    x = s_pop();
    s_push(new_obj.obj);
    s_push(x);

    while (1) { 
        i = ((BIGNUM * )ADDRESS(y))->val;
        new_obj.obj = tmp_obj.obj;
        tmp = getsvt(0);
        s_push(tmp);
        h = 0;
        while (1) { 
            x = s_pop();
            extended_mul(i, ((BIGNUM * )ADDRESS(x))->val, h, &h, &l);
            ((BIGNUM * )ADDRESS(new_obj.obj))->val += l;
            if (((BIGNUM * )ADDRESS(new_obj.obj))->val < 0) { 
                ((BIGNUM * )ADDRESS(new_obj.obj))->val &= INT_MASK;
                h++;
            }
            if (((BIGNUM * )ADDRESS(x))->next.obj == LTAG_NIL) { 
                while (h > 0) { 
                    if (((BIGNUM * )ADDRESS(new_obj.obj))->next.obj == LTAG_NIL) { 
                        s_push(y);  
                        s_push(tmp_obj.obj);
                        new_obj.obj = stretch_big(new_obj.obj, 0);
                        tmp_obj.obj = s_pop();
                        y = s_pop();
                    } else
                        new_obj.obj = ((BIGNUM * )ADDRESS(new_obj.obj))->next.obj;
                    ((BIGNUM * )ADDRESS(new_obj.obj))->val += h;
                    if (((BIGNUM * )ADDRESS(new_obj.obj))->val < 0) { 
                        ((BIGNUM * )ADDRESS(new_obj.obj))->val &= INT_MASK;
                        h = 1;
                    } else
                        break;
                }
                if (((BIGNUM * )ADDRESS(y))->next.obj == LTAG_NIL)
                    goto CON;
                y = ((BIGNUM * )ADDRESS(y))->next.obj;
                if (((BIGNUM * )ADDRESS(tmp_obj.obj))->next.obj == LTAG_NIL) { 
                    s_push(y);
                    tmp_obj.obj = stretch_big(tmp_obj.obj, 0);
                    y = s_pop();
                } else
                    tmp_obj.obj = ((BIGNUM * )ADDRESS(tmp_obj.obj))->next.obj;
                break;
            }
            s_push(((BIGNUM * )ADDRESS(x))->next.obj);
            if (((BIGNUM * )ADDRESS(new_obj.obj))->next.obj == LTAG_NIL) { 
                s_push(y);  
                s_push(tmp_obj.obj);
                new_obj.obj = stretch_big(new_obj.obj, 0);
                tmp_obj.obj = s_pop();
                y = s_pop();
            } else
                new_obj.obj = ((BIGNUM * )ADDRESS(new_obj.obj))->next.obj;
        }
    }

CON:

    s_pop();  /*  pop x value  */
    new_obj.obj = s_pop();
    normalize_big(new_obj.obj);
    if (sign < 0)
        return(complement_big(new_obj.obj));
    else
        return(new_obj.obj);
}


/*  Written by Yass. Sep. 3 '90  */
int big_q_r(x, y, r)
int x,   /*  positive bignum, greater than y  */
y,   /*  positive bignum, less than x  */
*r;  /*  returned value  */
{ 
    SOBJ tmp;
    int i, lx, ly, dummy;

    s_push(y);
    x = copy_big(x);
    y = s_pop();

    i = standardizing_factor_and_normalize(y);
    s_push(y);
    x = mul_int_big(x, i);
    y = s_pop();
    s_push(x);
    y = mul_int_big(y, i);
    x = s_pop();
    lx = big_length(x);
    ly = big_length(y);
    s_push(x);			/* T.Komiya */
    s_push(y);			/*     "    */
    tmp.obj = big_q_r_aux(x, y, lx - ly);
    y = s_pop();		/*     "    */
    x = s_pop();		/*     "    */
    if (tmp.obj == LTAG_NIL) { 
        s_push(x);
        s_push(y);
        tmp.obj = mk_bign(0);
        y = s_pop();
        x = s_pop();
    }
    s_push(tmp.obj);
    s_push(x);
    y = div_int_big(y, i, &dummy);
    x = s_pop();
    *r = div_int_big(x, i, &dummy);

    return(s_pop());
}


/*  Written by Yass. Aug. 31 '90  */
int standardizing_factor_and_normalize(obj)
int obj;  /*  positive bignum  */
{ 
    int i, j;

    if (((BIGNUM * )ADDRESS(obj))->next.obj == LTAG_NIL) { 
        if (((BIGNUM * )ADDRESS(obj))->val == 0)
            return(-1);
        for (i = 1, j = ((BIGNUM * )ADDRESS(obj))->val ; (j *= 2) >= 0 ; i *= 2)
            ;
        return(i);
    }
    i = standardizing_factor_and_normalize(((BIGNUM * )ADDRESS(obj))->next.obj);
    if (i < 0) { 
        ((BIGNUM * )ADDRESS(obj))->next.obj = LTAG_NIL;
        if (((BIGNUM * )ADDRESS(obj))->val == 0)
            return(-1);
        for (i = 1, j = ((BIGNUM * )ADDRESS(obj))->val ; (j *= 2) >= 0 ; i *= 2)
            ;
        return(i);
    }
    return(i);
}


/*  Written by Yass. Sep. 3 '90  */
int big_q_r_aux(x, y, i)
int x, y, 
i;     /*  non-negative int  */
{ 
    SOBJ q, qq;

    if (i < 0) {
/*        bokkan("INTERNAL ERROR(big_q_r_aux())."); */
		return(LTAG_NIL);
	}
    if (i == 0) { 
        i = div_big_big(y, x);
        if (i == 0)
            return(LTAG_NIL);
        qq.obj = mk_bign(i);
        return(qq.obj);
    }
    s_push(x);			/* T.Komiya */
    s_push(y);			/*     "    */
    q.obj = big_q_r_aux(((BIGNUM * )ADDRESS(x))->next.obj, y, i - 1);
    y = s_pop();		/*     "    */
    x = s_pop();		/*     "    */
    i = div_big_big(y, x);
    s_push(q.obj);		/*     "    */
    qq.obj = mk_bign(i);
    q.obj = s_pop();		/*     "    */
    ((BIGNUM * )ADDRESS(qq.obj))->next.obj = q.obj;
    return(qq.obj);
}


/*  Written by Yass. Sep. 3 '90  */
int div_big_big(x, y)
int x,  /*  positive bignum  */
y;  /*  positive bignum -> updated  */
{ 
    SOBJ xx, yy, debug;
    int q, r;

    xx.obj = x;
    yy.obj = y;

    while (((BIGNUM * )ADDRESS(xx.obj))->next.obj != LTAG_NIL) { 
        xx.obj = ((BIGNUM * )ADDRESS(xx.obj))->next.obj;
        yy.obj = ((BIGNUM * )ADDRESS(yy.obj))->next.obj;
    }
    if (((BIGNUM * )ADDRESS(yy.obj))->next.obj != LTAG_NIL) { 
        if (((BIGNUM * )ADDRESS(((BIGNUM * )ADDRESS(yy.obj))->next.obj))->val >= 
            ((BIGNUM * )ADDRESS(xx.obj))->val                                  )
            q = INT_MASK - 1;
        else
         { 
            extended_div(((BIGNUM * )ADDRESS(xx.obj))->val,
                ((BIGNUM * )ADDRESS(((BIGNUM * )ADDRESS(yy.obj))->next.obj))->val,
                ((BIGNUM * )ADDRESS(yy.obj))->val,
                &q, &r);
            q -= 2;
        }
    } else
        q = ((BIGNUM * )ADDRESS(yy.obj))->val / ((BIGNUM * )ADDRESS(xx.obj))->val - 2;
    if (q <= 0)
        q = 0;
    else
        sub_int_big_big(q, x, y);
    while (big_compare(x, y) & LESSE) { 
        q++;
        sub_int_big_big(1, x, y);
    }
    return(q);
}


/*  Written by Yass. Sep. 4 '90  */
int sub_int_big_big(i, x, y)
int i,  /*  non-negative int  */
x,  /*  non-negative bignum  */
y;  /*  non-negative bignum -> updated  */
{ 
    int h = 0,
    l;

    for (; ; ) { 
        extended_mul(i, ((BIGNUM * )ADDRESS(x))->val, h, &h, &l);
        ((BIGNUM * )ADDRESS(y))->val -= l;
        if (((BIGNUM * )ADDRESS(y))->val < 0) { 
            ((BIGNUM * )ADDRESS(y))->val &= INT_MASK;
            h++;
        }
        if (((BIGNUM * )ADDRESS(x))->next.obj == LTAG_NIL) { 
            while (h > 0) { 
                y = ((BIGNUM * )ADDRESS(y))->next.obj;
                ((BIGNUM * )ADDRESS(y))->val -= h;
                if (((BIGNUM * )ADDRESS(y))->val < 0) { 
                    ((BIGNUM * )ADDRESS(y))->val &= INT_MASK;
                    h = 1;
                } else
                    break;
            }
            break;
        }
        x = ((BIGNUM * )ADDRESS(x))->next.obj;
        y = ((BIGNUM * )ADDRESS(y))->next.obj;
    }
}



/*  Written by Yass. Sep. 7 '90  */
int complement_big2(obj, extnded)
int obj, extnded;
{
    while (((BIGNUM * )ADDRESS(obj))->next.obj != LTAG_NIL) { 
        if (((BIGNUM * )ADDRESS(obj))->val != 0) { 
            ((BIGNUM * )ADDRESS(obj))->val = (-(((BIGNUM * )ADDRESS(obj))->val)) & INT_MASK;
            goto CON;
        } else
            obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
    }
    if (((BIGNUM * )ADDRESS(obj))->val == ~INT_MASK) { 
        ((BIGNUM * )ADDRESS(obj))->val = 0;
        return(1);
    } else
        ((BIGNUM * )ADDRESS(obj))->val = -(((BIGNUM * )ADDRESS(obj))->val);
    if (extnded == 1)
        ((BIGNUM * )ADDRESS(obj))->val |= ~INT_MASK;
    return(0);

CON:

    for (; ; ) { 
        obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
        if (((BIGNUM * )ADDRESS(obj))->next.obj == LTAG_NIL)
            break;
        ((BIGNUM * )ADDRESS(obj))->val = (~(((BIGNUM * )ADDRESS(obj))->val)) & INT_MASK;
    }
    ((BIGNUM * )ADDRESS(obj))->val = ~(((BIGNUM * )ADDRESS(obj))->val);
    if (extnded == 1)
        ((BIGNUM * )ADDRESS(obj))->val |= ~INT_MASK;
    return(0);
}


/*  Written by Yass. Sep. 7 '90  */
void add_int_big2(obj, i)
int obj,  /* any bignum -> updated,returned */
i;    /* any int  */
{
    while (((BIGNUM * )ADDRESS(obj))->next.obj != LTAG_NIL) { 
        ((BIGNUM * )ADDRESS(obj))->val += i;
        if (((BIGNUM * )ADDRESS(obj))->val < 0)  /* carry */ { 
            i = 1;
            ((BIGNUM * )ADDRESS(obj))->val &= INT_MASK;
            obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
        } else
            return;
    }
    if (((BIGNUM * )ADDRESS(obj))->val >= 0) { 
        ((BIGNUM * )ADDRESS(obj))->val += i;
        if (((BIGNUM * )ADDRESS(obj))->val < 0)  /* overflow */
            ((BIGNUM * )ADDRESS(obj))->val &= INT_MASK;
    } else
        ((BIGNUM * )ADDRESS(obj))->val += i;
    return;
}


/*  Written by Yass. Sep. 7 '90  */
mul_int_big2(obj, i)
int obj, i;
{ 
    int h = 0,
    sign = 1;

    for (; ; ) { 
        extended_mul(i, ((BIGNUM * )ADDRESS(obj))->val, h, &h, &(((BIGNUM * )ADDRESS(obj))->val));
        if (((BIGNUM * )ADDRESS(obj))->next.obj == LTAG_NIL)
            break;
        obj = ((BIGNUM * )ADDRESS(obj))->next.obj;
    }
}


/*  Written by Yass. Sep. 7 '90  */
div_int_big2(obj, i, c, r)
int obj, i, *c, *r;
{
    if (((BIGNUM * )ADDRESS(obj))->next.obj == LTAG_NIL) { 
        if (*c == 1) { 
            *r = *c % i;
            *c /= i;
            extended_div(i, *r, ((BIGNUM * )ADDRESS(obj))->val, &((BIGNUM * )ADDRESS(obj))->val, r);
        } else
         { 
            *r = ((BIGNUM * )ADDRESS(obj))->val % i;
            ((BIGNUM * )ADDRESS(obj))->val /= i;
        }
    } else
     { 
        div_int_big(((BIGNUM * )ADDRESS(obj))->next.obj, i, r);
        extended_div(i, *r, ((BIGNUM * )ADDRESS(obj))->val, &(((BIGNUM * )ADDRESS(obj))->val), r);
    }
}


