/*
 * cons.c - 
 */

/*
 * $Id: cons.c,v 1.7 2014/10/28 02:22:55 komiya Exp $
 *
 *   pair function
 *   FILE NAME : cons.c
 *   car, cdr, cadr, ... , cddddr  by Yass.
 *   memq, memvq, member by Yass.
 *   assq, assv, assoc by M.Yamamoto.
 *   find-if  by M.Yamamoto.
 *   lista by Yass.
 */

static char *RCSid="$Id: cons.c,v 1.7 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"


/********************************/
/*                              */
/*  Internal useful procedures  */
/*                              */
/********************************/

void set_car(cons_obj, obj)
int cons_obj;  /* cons_obj must be pair */
int obj;
{
    ((CARCDR * )ADDRESS(cons_obj))->car.obj = obj;
}


void set_cdr(cons_obj, obj)
int cons_obj;  /* cons_obj must be pair */
int obj;
{
    ((CARCDR * )ADDRESS(cons_obj))->cdr.obj = obj;
}


int listlen(cons_obj)
int cons_obj;  /* cons_obj must be pair */
{ 
    int len = 0;

    while (GETtag(cons_obj) == LTAG_CONS) { 
        len++;
#ifdef CTHREAD
        cons_obj = touch_obj(CDRobj(cons_obj));
#else
        cons_obj = CDRobj(cons_obj);
#endif
    }
    return(len);
}


int nth(cons_obj, ith)
int cons_obj, ith;
{ 
    int i;

    for (i = 0 ; i < ith ; ++i) { 
        if (GETtag(cons_obj) == LTAG_CONS)
            cons_obj = CDRobj(cons_obj);
        else
            return((int)NULL);
    }
    return(CARobj(cons_obj));
}


unsigned int last_cons(cons_obj)
unsigned int cons_obj;
{
    while (GETtag(CDRobj(cons_obj)) == LTAG_CONS)
#ifdef TOUCH_ARG
        cons_obj = touch_obj(CDRobj(cons_obj));
#else
        cons_obj = CDRobj(cons_obj);
#endif
    return(cons_obj);
}


/***********************************/
/*                                 */
/*  Essential standard procedures  */
/*                                 */
/***********************************/

void Spairp()
{
#ifdef TOUCH_ARG
    if (!GETtag(touch_obj(getarg(1))))
#else
    if (!GETtag(getarg(1)))
#endif
        set_reg(LTAG_T);
    else
        set_reg(LTAG_FALSE);
}

void Slistp()                        /* by T.Komiya */
{
    SOBJ l, m;

#ifdef TOUCH_ARG
    l.obj = m.obj = touch_obj(getarg(1));
#else
    l.obj = m.obj = getarg(1);
#endif
    do {
        if (GETtag(l.obj) == LTAG_CONS)
#ifdef CTHREAD
            l.obj = touch_obj(CDRobj(l.obj));
#else
            l.obj = CDRobj(l.obj);
#endif
        else {
            set_reg(l.obj == LTAG_NIL ? LTAG_T : LTAG_FALSE);
            return;
        }
        if (GETtag(l.obj) == LTAG_CONS)
#ifdef CTHREAD
            l.obj = touch_obj(CDRobj(l.obj));
#else
            l.obj = CDRobj(l.obj);
#endif
        else {
            set_reg(l.obj == LTAG_NIL ? LTAG_T : LTAG_FALSE);
            return;
        }
#ifdef CTHREAD
        m.obj = touch_obj(CDRobj(m.obj));
#else
        m.obj = CDRobj(m.obj);
#endif
    } while (l.obj != m.obj);
    set_reg(LTAG_FALSE);
}

void Scons()
{
    set_reg(mk_cons(getarg(1), getarg(2)));
}


void Scar()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS)
        set_reg(CARobj(pair.obj));
    else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
	err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe  Nov. 04, 1992  */
    }		
}


void Scdr()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS)
        set_reg(CDRobj(pair.obj));
    else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
	err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe  Nov. 04, 1992  */
    }		
}


void Ssetcar()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) != LTAG_CONS){
        err_illegal_argument("1st argument must be a pair.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    set_car(pair.obj, getarg(2));
    set_reg(pair.obj);
}



void Ssetcdr()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) != LTAG_CONS){
        err_illegal_argument("1st argument must be a pair.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    set_cdr(pair.obj, getarg(2));
    set_reg(pair.obj);
}


#ifdef CTHREAD
extern int touch_obj();
#endif

void Scaar()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS)
            set_reg(CARobj(pair.obj));
        else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
            err_illegal_argument("Illegal argument.");
	    return ;
	        /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scadr()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS)
            set_reg(CARobj(pair.obj));
        else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
            err_illegal_argument("Illegal argument.");
	    return ;
	        /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scdar()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS)
            set_reg(CDRobj(pair.obj));
        else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
            err_illegal_argument("Illegal argument.");
	    return ;
	        /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scddr()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS)
            set_reg(CDRobj(pair.obj));
        else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
            err_illegal_argument("Illegal argument.");
	    return ;
	        /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scaaar()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS)
                set_reg(CARobj(pair.obj));
            else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
	            /*  By N.Watanabe  Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	        /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scaadr()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS)
                set_reg(CARobj(pair.obj));
            else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
	            /*  By N.Watanabe  Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
            err_illegal_argument("Illegal argument.");
	    return ;
	        /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scadar()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS)
                set_reg(CARobj(pair.obj));
            else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
	            /*  By N.Watanabe  Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scaddr()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS)
                set_reg(CARobj(pair.obj));
            else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
	            /*  By N.Watanabe  Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	    /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scdaar()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS)
                set_reg(CDRobj(pair.obj));
            else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
	            /*  By N.Watanabe  Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	    /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scdadr()
{ 
    SOBJ pair;
    int tag;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS)
                set_reg(CDRobj(pair.obj));
            else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
	            /*  By N.Watanabe  Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	    /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scddar()
{ 
    SOBJ pair;
    int tag;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS)
                set_reg(CDRobj(pair.obj));
            else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
	            /*  By N.Watanabe  Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	    /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scdddr()
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS)
                set_reg(CDRobj(pair.obj));
            else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
	            /*  By N.Watanabe  Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	    /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scaaaar()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CARobj(pair.obj));
#else
                pair.obj = CARobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CARobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Illegal argument.");
		    return ;
	                /*  By N.Watanabe  Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	    /*  By N.Watanabe  Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scaaadr()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CARobj(pair.obj));
#else
                pair.obj = CARobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CARobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		        /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	        /*  By N.Watanabe Nov. 04, 1992  */
	    }
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scaadar()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CARobj(pair.obj));
#else
                pair.obj = CARobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CARobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Illegal argument.");
		    return ;
		        /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	        /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scaaddr()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CARobj(pair.obj));
#else
                pair.obj = CARobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CARobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Illegal argument.");
		    return ;
		        /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	        /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scadaar()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CDRobj(pair.obj));
#else
                pair.obj = CDRobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CARobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Illegal argument.");
		    return ;
		        /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Illegal argument.");
		return ;
	        /*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Illegal argument.");
	    return ;
	        /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scadadr()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CDRobj(pair.obj));
#else
                pair.obj = CDRobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CARobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		        /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		   /*  By N.Watanabe Nov. 04, 1992  */
	    }
	} else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	       /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scaddar()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CDRobj(pair.obj));
#else
                pair.obj = CDRobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CARobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		/*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scadddr()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CDRobj(pair.obj));
#else
                pair.obj = CDRobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CARobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		/*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scdaaar()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CARobj(pair.obj));
#else
                pair.obj = CARobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CDRobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		/*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scdaadr()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CARobj(pair.obj));
#else
                pair.obj = CARobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CDRobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		/*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scdadar()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CARobj(pair.obj));
#else
                pair.obj = CARobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CDRobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		/*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scdaddr()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CARobj(pair.obj));
#else
                pair.obj = CARobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CDRobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		/*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scddaar()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CDRobj(pair.obj));
#else
                pair.obj = CDRobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CDRobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		/*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scddadr()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CARobj(pair.obj));
#else
            pair.obj = CARobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CDRobj(pair.obj));
#else
                pair.obj = CDRobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CDRobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		/*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scdddar()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CARobj(pair.obj));
#else
        pair.obj = CARobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CDRobj(pair.obj));
#else
                pair.obj = CDRobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CDRobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		/*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Scddddr()  /*  Yass. Sep. 17 '90  */
{ 
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
        pair.obj = touch_obj(CDRobj(pair.obj));
#else
        pair.obj = CDRobj(pair.obj);
#endif
        if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
            pair.obj = touch_obj(CDRobj(pair.obj));
#else
            pair.obj = CDRobj(pair.obj);
#endif
            if (GETtag(pair.obj) == LTAG_CONS) { 
#ifdef CTHREAD
                pair.obj = touch_obj(CDRobj(pair.obj));
#else
                pair.obj = CDRobj(pair.obj);
#endif
                if (GETtag(pair.obj) == LTAG_CONS)
                    set_reg(CDRobj(pair.obj));
                else if (pair.obj == LTAG_NIL)
                    set_reg(LTAG_NIL);
                else {
		    err_illegal_argument("Argument must be a pair or ().");
		    return ;
		    /*  By N.Watanabe Nov. 04, 1992  */
		}
            } else if (pair.obj == LTAG_NIL)
                set_reg(LTAG_NIL);
            else {
		err_illegal_argument("Argument must be a pair or ().");
		return ;
		/*  By N.Watanabe Nov. 04, 1992  */
	    }
        } else if (pair.obj == LTAG_NIL)
            set_reg(LTAG_NIL);
        else {
	    err_illegal_argument("Argument must be a pair or ().");
	    return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
	}
    } else if (pair.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else {
        err_illegal_argument("Argument must be a pair or ().");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Snullp()  /*  Yass. Sep. 17 '90  */
{
#ifdef TOUCH_ARG
    if (touch_obj(getarg(1)) == LTAG_NIL)
#else
    if (getarg(1) == LTAG_NIL)
#endif
	set_reg(LTAG_T);
    else
	set_reg(LTAG_FALSE);
}


void Slist()  /*  Yass. Sep. 18 '90  */
{ 
    CARCDR * top, *cell;
    int args, i;

    args = getargnum();
    if (args == 0)
        set_reg(LTAG_NIL);
    else
     { 
        top = cell = (CARCDR * )mk_nconsp(args);
        for (i = 1 ; i <= args ; i++) { 
            cell->car.obj = getarg(i);
            cell++;
        }
        set_reg(SSERDDA(top) | LTAG_CONS);
    }
}


void Slength()  /*  Yass. Sep. 18 '90  */
{ 
    SOBJ list;
    int len = 0;

#ifdef TOUCH_ARG
    list.obj = touch_obj(getarg(1));
#else
    list.obj = getarg(1);
#endif
    while (GETtag(list.obj) == LTAG_CONS) { 
        len++;
#ifdef CTHREAD
        list.obj = touch_obj(CDRobj(list.obj));
#else
        list.obj = CDRobj(list.obj);
#endif
    }
    if (list.obj == LTAG_NIL)
        set_reg(mk_fix(len));
    else {
        err_illegal_argument("Argument must be a list.");
	return ;
	    /*  By N.Watanabe  Nov. 04, 1992  */
    }
}

void append()
{
    unsigned int tmp;

    if (reg1 == LTAG_NIL)
        return;
    else if (GETtag(reg1) == LTAG_CONS) {
        s_push(reg_reg);
        reg_reg = mk_cons(CARobj(reg1), LTAG_NIL);
        s_push(reg_reg);
        for (;;) {
#ifdef CTHREAD
            reg1 = touch_obj(CDRobj(reg1));
#else
            reg1 = CDRobj(reg1);
#endif
            if (reg1 == LTAG_NIL)
                break;
            else if (GETtag(reg1) != LTAG_CONS) {
                addtop(2);
                reg1 = LTAG_NIL;
                err_illegal_argument("Argument must be a list.");
            }
            tmp = mk_cons(CARobj(reg1), LTAG_NIL);
            set_cdr(reg_reg, tmp);
            reg_reg = CDRobj(reg_reg);
        }
        set_cdr(reg_reg, getsvt(1));
    } else {
        reg1 = LTAG_NIL;
        err_illegal_argument("Argument must be a list.");
    }
    reg_reg = s_pop();
    addtop(1);
}

void Sappend()
{
    int args, i;

    args = getargnum();
    if (args < 2) {
        if (args == 1)
            set_reg(getarg(1));
        else
            set_reg(LTAG_NIL);
        return;
    }
#ifdef TOUCH_ARG
    for (i = args-1, reg_reg = touch_obj(getarg(args)); i >= 1; i--) {
        reg1 = touch_obj(getarg(i));
#else
    for (i = args-1, reg_reg = getarg(args); i >= 1; i--) {
        reg1 = getarg(i);
#endif
        append();
    }
    reg1 = LTAG_NIL;
}

void Sreverse()  /*  Yass. Sep. 18 '90  */
{ 
    SOBJ   list;
    CARCDR * top, *cell;
    int len = 0;

#ifdef TOUCH_ARG
    list.obj = touch_obj(getarg(1));
#else
    list.obj = getarg(1);
#endif
    if (GETtag(list.obj) != LTAG_CONS && list.obj != LTAG_NIL){
	err_illegal_argument("Argument must be a list.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    if (list.obj == LTAG_NIL)
        set_reg(LTAG_NIL);
    else
     { 
        do
         { 
            s_push(CARobj(list.obj));
#ifdef CTHREAD
            list.obj = touch_obj(CDRobj(list.obj));
#else
            list.obj = CDRobj(list.obj);
#endif
            len++;
        } while (GETtag(list.obj) == LTAG_CONS);
        if (list.obj != LTAG_NIL) { 
            addtop(len);
	    err_illegal_argument("Argument must be a list.");
	    return ;
    	        /*  By N.Watanabe Nov. 04, 1992  */
	}
        else
         { 
            top = cell = (CARCDR * )mk_nconsp(len);
            for (; 0 < len ; len--) { 
                cell->car.obj = s_pop();
                cell++;
            }
            set_reg(SSERDDA(top) | LTAG_CONS);
        }
    }
}


void Smemq()  /*  Yass. Sep. 18 '90  */
{ 
    SOBJ obj, list;
    int flag = FALSE;
#ifdef CTHREAD
    int touch_obj();
#endif

#ifdef TOUCH_ARG
    list.obj = touch_obj(getarg(2));
#else
    list.obj = getarg(2);
#endif
    if (GETtag(list.obj) != LTAG_CONS && list.obj != LTAG_NIL){
	err_illegal_argument("2nd argument must be a list.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    if (list.obj == LTAG_NIL)
        set_reg(LTAG_FALSE);
    else
     { 
#ifdef TOUCH_ARG
        obj.obj  = touch_obj(getarg(1));
#else
        obj.obj  = getarg(1);
#endif
        do {
#ifdef CTHREAD
	    if (flag == FALSE && EQ(obj.obj, touch_obj(CARobj(list.obj)))) { 
#else
	    if (flag == FALSE && EQ(obj.obj, CARobj(list.obj))) { 
#endif
                set_reg(list.obj);
                flag = TRUE;
            }
#ifdef CTHREAD
            list.obj = touch_obj(CDRobj(list.obj));
#else
            list.obj = CDRobj(list.obj);
#endif
        } while (GETtag(list.obj) == LTAG_CONS);
#ifdef CTHREAD
        if (touch_obj(list.obj) != LTAG_NIL){
#else
        if (list.obj != LTAG_NIL){
#endif
	    err_illegal_argument("2nd argument must be a list.");
	    return ;
	        /*  By N.Watanabe Nov. 04, 1992  */
	}
        if (flag == FALSE)
            set_reg(LTAG_FALSE);
    }
}


void Smemv()  /*  Yass. Sep. 18 '90  */
{ 
    SOBJ obj, list;
    int flag = FALSE;
#ifdef CTHREAD
    int touch_obj();
#endif

#ifdef TOUCH_ARG
    list.obj = touch_obj(getarg(2));
#else
    list.obj = getarg(2);
#endif
    if (GETtag(list.obj) != LTAG_CONS && list.obj != LTAG_NIL){
	err_illegal_argument("2nd argument must be a list.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    if (list.obj == LTAG_NIL)
        set_reg(LTAG_FALSE);
    else
     { 
#ifdef TOUCH_ARG
        obj.obj = touch_obj(getarg(1));
#else
        obj.obj = getarg(1);
#endif
        do { 
#ifdef CTHREAD
            if (flag == FALSE && eqv(obj.obj, touch_obj(CARobj(list.obj)))) {
#else
            if (flag == FALSE && eqv(obj.obj, CARobj(list.obj))) {
#endif
                set_reg(list.obj);
                flag = TRUE;
            }
#ifdef CTHREAD
            list.obj = touch_obj(CDRobj(list.obj));
#else
            list.obj = CDRobj(list.obj);
#endif
        } while (GETtag(list.obj) == LTAG_CONS);
#ifdef CTHREAD
        if (touch_obj(list.obj) != LTAG_NIL){
#else
        if (list.obj != LTAG_NIL){
#endif
	    err_illegal_argument("2nd argument must be a list.");
	    return ;
	        /*  By N.Watanabe Nov. 04, 1992  */
	}
        if (flag == FALSE)
            set_reg(LTAG_FALSE);
    }
}



void Smember()  /*  Yass. Sep. 18 '90  */
{ 
    SOBJ obj, list;
    int flag = FALSE;
#ifdef CTHREAD
    int touch_obj();
#endif

#ifdef TOUCH_ARG
    list.obj = touch_obj(getarg(2));
#else
    list.obj = getarg(2);
#endif
    if (GETtag(list.obj) != LTAG_CONS && list.obj != LTAG_NIL){
	err_illegal_argument("2nd argument must be a list.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    if (list.obj == LTAG_NIL)
        set_reg(LTAG_FALSE);
    else
     { 
        obj.obj = getarg(1);
        do { 
            if (flag == FALSE && equal(obj.obj, CARobj(list.obj))) { 
                set_reg(list.obj);
                flag = TRUE;
            }
#ifdef CTHREAD
            list.obj = touch_obj(CDRobj(list.obj));
#else
            list.obj = CDRobj(list.obj);
#endif
        } while (GETtag(list.obj) == LTAG_CONS);
#ifdef CTHREAD
        if (touch_obj(list.obj) != LTAG_NIL){
#else
        if (list.obj != LTAG_NIL){
#endif
	    err_illegal_argument("2nd argument must be a list.");
	    return ;
	        /*  By N.Watanabe Nov. 04, 1992  */
	}
        if (flag == FALSE)
            set_reg(LTAG_FALSE);
    }
}


/*
 *  assq, assv, assoc
 *  original version by Yass.
 *  2nd version by Mikio Yamamoto.  Oct. 1990.
 */

eq(a, b)
int a, b;
{
    if (a == b)
        return(TRUE);
    else
        return(FALSE);
}


int eqv();
int equal();

void Sassq()
{
    switch (GETtag(getarg(2))) {
    case LTAG_CONS:
        set_reg(assoc(eq, getarg(1), getarg(2)));
        return;
    case LTAG_NIL:
        set_reg(LTAG_FALSE);
        return;
    default:
	err_illegal_argument("2nd argument must be a list.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Sassv()
{
    switch (GETtag(getarg(2))) {
    case LTAG_CONS:
        set_reg(assoc(eqv, getarg(1), getarg(2)));
        return;
    case LTAG_NIL:
        set_reg(LTAG_FALSE);
        return;
    default:
	err_illegal_argument("2nd argument must be a list.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


void Sassoc()
{
    switch (GETtag(getarg(2))) {
    case LTAG_CONS:
        set_reg(assoc(equal, getarg(1), getarg(2)));
        return;
    case LTAG_NIL:
        set_reg(LTAG_FALSE);
        return;
    default:
	err_illegal_argument("2nd argument must be a list.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
}


assoc(eq_proc, key, alist)
int (*eq_proc)();
int key;
int alist;
{
    int a_alist;
#ifdef CTHREAD
    int touch_obj();
#endif

#ifdef TOUCH_ARG
    alist = touch_obj(alist);
    key = touch_obj(key);
#endif
    if (alist == LTAG_NIL)
        return(LTAG_FALSE);
    else {
        while (GETtag(alist) == LTAG_CONS) {
#ifdef CTHREAD
            a_alist = touch_obj(CARobj(alist));
#else
            a_alist = CARobj(alist);
#endif
            switch (GETtag(a_alist)) {
            case LTAG_CONS :
#ifdef CTHREAD
                if (eq_proc(key, touch_obj(CARobj(a_alist))))
#else
                if (eq_proc(key, CARobj(a_alist)))
#endif
                    return(a_alist);
                break;
            case LTAG_NIL :
                break;
            default :
                bokkan("Elements of association list must be pair.");
            }
#ifdef CTHREAD
            alist = touch_obj(CDRobj(alist));
#else
            alist = CDRobj(alist);
#endif
        }
#ifdef CTHREAD
        if (GETtag(touch_obj(alist)) != LTAG_NIL)
#else
        if (GETtag(alist) != LTAG_NIL)
#endif
            bokkan("Elements of association list must be pair.");
        else 
            return(LTAG_FALSE);
    }
}



/*
 *  find-if
 *  written by M.Yamamoto.
 *
 */

void find_if2();

void Sfind_if()
{
#ifdef TOUCH_ARG
    if ( !FUNp(touch_obj(getarg(1)))) {
#else
    if ( !FUNp(getarg(1)) ){
#endif
	err_illegal_argument("1st argument must be a procedure.");
	return ;
	    /*  By N.Watanabe Nov. 04, 1992  */
    }
    open_frame2(find_if2, LTAG_NIL, 0);
#ifdef TOUCH_ARG
    s_push(touch_obj(getarg(1)));
#else
    s_push(getarg(1));
#endif
#ifdef TOUCH_ARG
    switch (GETtag(touch_obj(getarg(2)))) {
#else
    switch (GETtag(getarg(2))) {
#endif
    case LTAG_CONS :
#ifdef TOUCH_ARG
        s_push(CARobj(touch_obj(getarg(2))));
#else
        s_push(CARobj(getarg(2)));
#endif
        break;
    case LTAG_NIL  :
/*        reg_reg = LTAG_NIL; */
        reg_reg = LTAG_FALSE;	/* T.Komiya */
        return;
    default :
        err_illegal_argument("2nd argument must be a list.");
	return ;
	    /*  By N.Watanabe  Nov. 04, 1992  */
    }
    set_up_stack();
}


void find_if2()
{
    unsigned *p;

    s_pop();
    s_pop();
#ifdef CTHREAD
    reg_reg = touch_obj(reg_reg);
#endif
    if (reg_reg != LTAG_FALSE && reg_reg != LTAG_NIL) {
#ifdef CTHREAD
        reg_reg = CARobj(touch_obj(getarg(2)));
#else
        reg_reg = CARobj(getarg(2));
#endif
        xreturn();
    } else {
        open_frame2(find_if2, LTAG_NIL, 0);
#ifdef TOUCH_ARG
        s_push(touch_obj(getarg(1)));
#else
        s_push(getarg(1));
#endif
        p = active_reg - 3;
#ifdef CTHREAD
        switch (GETtag(touch_obj(CDRobj(touch_obj(*p))))) {
#else
        switch (GETtag(CDRobj(*p))) {
#endif
        case LTAG_CONS :
#ifdef CTHREAD
            s_push(CARobj(touch_obj(CDRobj(touch_obj(*p)))));
            *p = CDRobj(touch_obj(*p));
#else
            s_push(CARobj(CDRobj(*p)));
            *p = CDRobj(*p);
#endif
            xcall();
            break;
        case LTAG_NIL  :
/*            reg_reg = LTAG_NIL; */
            reg_reg = LTAG_FALSE;	/* T.Komiya */
            xreturn();
            return;
        default :
            err_illegal_argument("Arguments after 2nd must be lists.");
	    return ;
	        /*  By N.Watanabe  Nov. 04, 1992  */
        }
    }
}

#ifdef CALLIOC

void find_if2_ep()
{
    unsigned *p;

    ((unsigned *)stack_bottom)[5]--;
    s_pop();
    s_pop();
#ifdef CTHREAD
    reg_reg = touch_obj(reg_reg);
#endif
    if (reg_reg != LTAG_FALSE && reg_reg != LTAG_NIL) {
#ifdef CTHREAD
        reg_reg = CARobj(touch_obj(getarg(2)));
#else
        reg_reg = CARobj(getarg(2));
#endif
        xreturn();
    } else {
        open_frame2(find_if2, LTAG_NIL, 0);
#ifdef TOUCH_ARG
        s_push(touch_obj(getarg(1)));
#else
        s_push(getarg(1));
#endif
        p = active_reg - 3;
#ifdef CTHREAD
        switch (GETtag(touch_obj(CDRobj(touch_obj(*p))))) {
#else
        switch (GETtag(CDRobj(*p))) {
#endif
        case LTAG_CONS :
#ifdef CTHREAD
            s_push(CARobj(touch_obj(CDRobj(touch_obj(*p)))));
            *p = CDRobj(touch_obj(*p));
#else
            s_push(CARobj(CDRobj(*p)));
            *p = CDRobj(*p);
#endif
            xcall();
            break;
        case LTAG_NIL  :
/*            reg_reg = LTAG_NIL; */
            reg_reg = LTAG_FALSE;	/* T.Komiya */
            xreturn();
            return;
        default :
            err_illegal_argument("Arguments after 2nd must be lists.");
	    return ;
	        /*  By N.Watanabe  Nov. 04, 1992  */
        }
    }
}

#endif

/*************************/
/*                       */
/*  Original procedures  */
/*                       */
/*************************/

void Slista()  /*  Yass. Sep. 18 '90  */
{ 
    CARCDR * top, *cell;
    int args, i;

    args = getargnum();
    if (args == 0)
        set_reg(LTAG_NIL);
    else if (args == 1)
        set_reg(getarg(1));
    else
     { 
        top = cell = (CARCDR * )mk_nconsp(args - 1);
        for (i = 1 ; i < args ; i++) { 
            cell->car.obj = getarg(i);
            cell++;
        }
        cell--;
        cell->cdr.obj = getarg(args);
        set_reg(SSERDDA(top) | LTAG_CONS);
    }
}

unsigned int list_tail(unsigned int pair, int k)
{
    while (k-- != 0) {
        if (GETtag(pair) != LTAG_CONS)
            if (GETtag(pair) == LTAG_NIL)
                return 0;       /* out of list length */
            else
                return 1;       /* pair is not a list */
#ifdef TOUCH_ARG
        pair = touch_obj(CDRobj(pair));
#else
        pair = CDRobj(pair);
#endif
    }
/*
    if (GETtag(pair) == LTAG_CONS || GETtag(pair) == LTAG_NIL)
        return pair;
    else
        return 1;               *//* pair is not a list */
    return pair;
}

int fix_to_int();

unsigned int aux_Slist_tail()
{
    SOBJ pair, k;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) != LTAG_CONS && GETtag(pair.obj) != LTAG_NIL) {
        err_illegal_argument("1st argument must be a list.");
	return 0;
    }

#ifdef TOUCH_ARG
    k.obj = touch_obj(getarg(2));
#else
    k.obj = getarg(2);
#endif
    if (GETtag(k.obj) != LTAG_FIXNUM) {
        err_illegal_argument_type(k.obj, "positive fixnum number");
	return 0;
    }

    pair.obj = list_tail(pair.obj, fix_to_int(k.obj));
    switch (pair.obj) {
    case 0:
        err_illegal_argument("2nd argument is out of list length.");
        return 0;
    case 1:
#ifdef TOUCH_ARG
        err_illegal_argument_type(touch_obj(getarg(1)), "list");
#else
        err_illegal_argument_type(getarg(1), "list");
#endif
        return 0;
    }
    return pair.obj;
}

void Slist_tail()
{
    SOBJ pair;

    pair.obj = aux_Slist_tail();
    if (pair.obj)
        set_reg(pair.obj);
    else
        return;                 /* error is detected */
}

void Slist_ref()
{
    SOBJ pair;

    pair.obj = aux_Slist_tail();
    if (!pair.obj) return;      /* error is detected */

    switch (GETtag(pair.obj)) {
    case LTAG_CONS:
#ifdef TOUCH_ARG
        set_reg(CARobj(touch_obj(pair.obj)));
#else
        set_reg(CARobj(pair.obj));
#endif
        break;
    case LTAG_NIL:
        err_illegal_argument("2nd argument is out of list length.");
        return;
    default:
#ifdef TOUCH_ARG
        err_illegal_argument_type(touch_obj(getarg(1)), "list");
#else
        err_illegal_argument_type(getarg(1), "list");
#endif
        return;
    }
}

unsigned int nthcdr(unsigned int pair, int k)
{
    while (k-- != 0) {
        switch (GETtag(pair)) {
        case LTAG_CONS:
#ifdef TOUCH_ARG
            pair = touch_obj(CDRobj(pair));
#else
            pair = CDRobj(pair);
#endif
            break;
        case LTAG_NIL:
            return LTAG_NIL;
        default:
            return 0;           /* pair is not a list */
        }
    }
    return pair;
}

unsigned int aux_Snthcdr()
{
    SOBJ pair, k;

#ifdef TOUCH_ARG
    k.obj = touch_obj(getarg(1));
#else
    k.obj = getarg(1);
#endif
    if (GETtag(k.obj) != LTAG_FIXNUM) {
        err_illegal_argument_type(k.obj, "positive fixnum number");
	return 0;
    }

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(2));
#else
    pair.obj = getarg(2);
#endif
    if (GETtag(pair.obj) == LTAG_NIL)
        return LTAG_NIL;
        
    if (GETtag(pair.obj) != LTAG_CONS){
        err_illegal_argument("2nd argument must be a list.");
	return 0;
    }

    if (!(pair.obj = nthcdr(pair.obj, fix_to_int(k.obj)))) {
#ifdef TOUCH_ARG
        err_illegal_argument_type(touch_obj(getarg(2)), "list");
#else
        err_illegal_argument_type(getarg(2), "list");
#endif

        return 0;
    }
    return pair.obj;
}

void Snthcdr()
{
    SOBJ pair;

    pair.obj = aux_Snthcdr();
    if (pair.obj)
        set_reg(pair.obj);
    else
        return;                 /* error is detected */
}

void Snth()
{
    SOBJ pair;

    pair.obj = aux_Snthcdr();
    if (!pair.obj) return;      /* error is detected */

    switch (GETtag(pair.obj)) {
    case LTAG_CONS:
#ifdef TOUCH_ARG
        set_reg(CARobj(touch_obj(pair.obj)));
#else
        set_reg(CARobj(pair.obj));
#endif
        break;
    case LTAG_NIL:
        set_reg(LTAG_NIL);
        break;
    default:
#ifdef TOUCH_ARG
        err_illegal_argument_type(touch_obj(getarg(2)), "list");
#else
        err_illegal_argument_type(getarg(2), "list");
#endif
        return;
    }
}

void Slast()
{
    SOBJ pair;

#ifdef TOUCH_ARG
    pair.obj = touch_obj(getarg(1));
#else
    pair.obj = getarg(1);
#endif
    if (GETtag(pair.obj) == LTAG_NIL)
        set_reg(LTAG_NIL);
    else if (GETtag(pair.obj) != LTAG_CONS){
        err_illegal_argument("argument must be a list.");
	return;
    } else
        set_reg(last_cons(pair.obj));
}
