/*
 * future.c - 
 */

/*
 * $Id: future.c,v 1.11 2018/10/31 03:17:26 komiya Exp $
 *
 *       Parallel Constructs (future.c)
 *      written by T.Komiya Sep 17 1992
 */

static char *RCSid="$Id: future.c,v 1.11 2018/10/31 03:17:26 komiya Exp $";

#include "scheme.h"

#ifdef THREAD
#ifdef SUNSOL2
#include <sched.h>
#endif
#endif


extern int top();
extern int iretaddrx();

/*
 *    Future
 */

#ifdef CTHREAD

# ifdef LUNA
# include "/usr/sys/mach/luna/thread_status.h"
# endif

# ifdef PTHREAD
#  ifdef __PTL__
void pthread_init(){}
#  endif

mutex_t mutex_alloc()
{
    pthread_mutex_t *m = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
    pthread_mutex_init(m, NULL);
    return (mutex_t)m;
}

void mutex_free(m)
mutex_t m;
{
    pthread_mutex_destroy((pthread_mutex_t *)m);
    free((char *)m);
}

condition_t condition_alloc()
{
    pthread_cond_t *c = (pthread_cond_t *)malloc(sizeof(pthread_cond_t));
    pthread_cond_init(c, NULL);
    return (condition_t)c;
}

void condition_free(c)
condition_t c;
{
    pthread_cond_destroy((pthread_cond_t *)c);
    free((char *)c);
}

int mutex_try_lock(m)
mutex_t m;
{
   if (pthread_mutex_trylock(m) == 0)
       return(1);
   else
       return(0);
}

cthread_t cthread_fork(func, arg)
void *(*func)();
void *arg;
{
#ifdef PTHREAD
    extern pthread_attr_t pattr;
#endif
    pthread_t thread;

    pthread_create(&thread, &pattr, func, arg);
/*
#if defined(PTHREAD) && defined(sparc) && defined(USE_REG)
    pthread_create(&thread, &pattr, func, arg);
#else
    pthread_create(&thread, NULL, func, arg);
#endif
*/
    return(thread);
}

#ifdef LTC
#ifdef CONDITIONAL_FUTURE
mtus_thread_t mtus_thread_fork_secondary(func, arg)
void *(*func)();
void *arg;
{
    mtus_thread_t thread;
# if defined(PTHREAD) && defined(CONDITIONAL_FUTURE)
    extern pthread_attr_t sattr;
# endif

    pthread_create(&thread, &sattr, func, arg);
/*
# if defined(PTHREAD) && defined(sparc) && defined(USE_REG)
    pthread_create(&thread, &sattr, func, arg);
# else
    pthread_create(&thread, NULL, func, arg);
# endif
*/

    return(thread);
}
#endif
#endif

pthread_key_t pthread_key;

# if !defined(SUNSOL2) && !defined(FREEBSD)
void *cthread_data(t)
cthread_t t;
{
    void *data;

#  if defined(PGCC) || defined(LINUX)
    data = pthread_getspecific(pthread_key);
#  else
    pthread_getspecific(pthread_key, &data);
#  endif
    return(data);
}
# endif
#endif /*  PTHREAD */

mutex_t cthread_fork_lock;

extern REGS saved_context;
extern mutex_t addsymbol_lock;
extern mutex_t memget_lock;
extern mutex_t print_lock;
extern mutex_t read_lock;
#ifdef SHE
extern thread_info_t pri_act, sub_act;
#endif

unsigned int touch_obj(obj)
unsigned int obj;
{
    TOUCH(obj, {});
    return(obj);
}


restore_active_frame()
{
    SOBJ actframe;
    unsigned int *stackobj;
    int fsize;
    unsigned int active;

    if (more.obj != LTAG_NIL) {
        actframe.obj = CARobj(more.obj);
        more.obj = CDRobj(more.obj);
        while (1) {
            stackobj = (unsigned int *)ADDRESS(CARobj(actframe.obj));
            fsize = (0x0ffffff & (*stackobj >> 2)) - 2;
            stackobj++;
            for (; fsize > 0; fsize--)
                s_push(*(stackobj++));
            s_push((int)(fix_to_int(*(stackobj++)))); /* pc */
            s_push(fix_to_int(*(stackobj++))); /* ret */
            if ((actframe.obj = CDRobj(actframe.obj)) == LTAG_NIL)
                break;
            active = (unsigned int)IMM(CARobj(actframe.obj)) + (unsigned int)sp_reg;
            s_push(active);
            s_push(base_reg);
            base_reg = sp_reg;
            actframe.obj = CDRobj(actframe.obj);
        }
        s_pop();
        s_pop();
        s_pop();
    }
}

#ifndef SHE

void bc_intrp_future2()
{
    FUTURE *fut;

    if (GETtag(*(sp_reg+1)) != LTAG_FUTURE)
        bokkan("not future!! in bc_intrp_future2()");
    s_pop();
    fut = (FUTURE *)(ADDRESS(s_pop()));
    mutex_lock(fut->lock);
    if (fut->status == DETERMINED) {
        fut->value.obj = reg_reg;
/*      cthread_detach(cthread_self()); */
        condition_broadcast(fut->done);
    } else {
        fut->value.obj = reg_reg;
        fut->status = DETERMINED;
/*      cthread_detach(cthread_self()); */
        condition_broadcast(fut->done);
    }
    mutex_unlock(fut->lock);
    if (optcnt.obj != LTAG_NIL) {
        if (GETtag(CARobj(optcnt.obj)) == LTAG_CONTINUATION) {
            more.obj = ((CONTIN *)ADDRESS(CARobj(optcnt.obj)))->body.obj;
            optcnt.obj = CDRobj(optcnt.obj);
            sp_reg = (unsigned int *)stack_bottom;
            top();
            return;
        } else {
            open_frame2(bc_intrp_future2, LTAG_FUTURE | SSERDDA(fut), 0);
            s_push(CARobj(optcnt.obj));
            optcnt.obj = CDRobj(optcnt.obj);
            s_push(reg_reg);
            xcall();
            return;
        }
    }
    /*
     * epilogue(memory deallocation, thread termination)
     */
#ifdef PTHREAD
    /* release stack and regs by cleanup handler */
#else
    free(stack_bottom - CHILD_STACK_SIZE + 28);
    free(cthread_data(cthread_self()));
#endif
/*  need to push fut */
/*
    mutex_free(fut->lock);
    condition_free(fut->done);
*/
    mutex_lock(cthread_fork_lock);
    current_thread_no--;
    mutex_unlock(cthread_fork_lock);
    cthread_exit(NULL);
}

#ifdef CALLIOC

void bc_intrp_future2_ep()
{
    FUTURE *fut;

    s_pop();
    fut = (FUTURE *)ADDRESS(*((unsigned int *)ADDRESS(s_pop())+1));
    mutex_lock(fut->lock);
    if (fut->status == DETERMINED) {
        fut->value.obj = reg_reg;
/*      cthread_detach(cthread_self()); */
        condition_broadcast(fut->done);
    } else {
        fut->value.obj = reg_reg;
        fut->status = DETERMINED;
/*      cthread_detach(cthread_self()); */
        condition_broadcast(fut->done);
    }
    mutex_unlock(fut->lock);
    if (optcnt.obj != LTAG_NIL) {
        if (GETtag(CARobj(optcnt.obj)) == LTAG_CONTINUATION) {
            more.obj = ((CONTIN *)ADDRESS(CARobj(optcnt.obj)))->body.obj;
            optcnt.obj = CDRobj(optcnt.obj);
            sp_reg = (unsigned int *)stack_bottom;
            top();
            return;
        } else {
            open_frame2(bc_intrp_future2, LTAG_FUTURE | SSERDDA(fut), 0);
            s_push(CARobj(optcnt.obj));
            optcnt.obj = CDRobj(optcnt.obj);
            s_push(reg_reg);
            xcall();
            return;
        }
    }
    /*
     * epilogue(memory deallocation, thread termination)
     */
#ifdef PTHREAD
    /* release stack and regs by cleanup handler */
#else
    free(stack_bottom - CHILD_STACK_SIZE + 28);
    free(cthread_data(cthread_self()));
#endif
/*  need to push fut */
/*
    mutex_free(fut->lock);
    condition_free(fut->done);
*/
    mutex_lock(cthread_fork_lock);
    current_thread_no--;
    mutex_unlock(cthread_fork_lock);
    cthread_exit(NULL);
}

#endif

#ifdef PTHREAD
void free_context(dummy)
int dummy;
{
    free(stack_bottom - CHILD_STACK_SIZE + 28);
    free((char *)cthread_data(cthread_self()));
}
#endif


void bc_interprete_future(future)
unsigned int *future;
{
    FUTURE *fut;
    REGS *regsp;
    extern int exec_sysfun();

    if (GETtag(future) != LTAG_FUTURE)
        bokkan("not future object!");

    fut = (FUTURE *)ADDRESS(future);

#ifndef PTHREAD
    fut->thread_id = thread_self();
#endif

#if defined(sparc) && defined(USE_REG)
    main_thread_regs = (REGS *)malloc(sizeof(REGS));
#else
    regsp = (REGS *)malloc(sizeof(REGS));
    cthread_set_data(cthread_self(), (any_t)regsp);
#endif

    fut->status = UNDETERMINED;
    /*
    while (fut->status == NOT_READY)
        ;
    */
    stack_top = (char *)malloc(CHILD_STACK_SIZE);
    stack_bottom = stack_top + CHILD_STACK_SIZE - 28;
    stack_top = stack_top + CHILD_GETA_OF_STACK;

#ifdef PTHREAD
# ifndef LTC
    pthread_cleanup_push(free_context, NULL);
# endif
#endif

    mutex_unlock(fut->lock);

    *((unsigned int **)stack_bottom + 1) = (unsigned int *)LTAG_NIL;
    *((unsigned int **)stack_bottom + 2) = 0;
    *((unsigned int **)stack_bottom + 3) = 0;

    stack_area_bottom = stack_bottom;
    set_top(stack_bottom);
    s_push(top);
    s_push(0);
    s_push(0);
    set_base(top_val());
    set_active(top_val());
    set_a2(LTAG_NIL);
    set_reg(LTAG_NIL);

    old_top = (unsigned int *)NULL;
    old_top1 = (unsigned int *)NULL;
    reg1 = LTAG_NIL;
    reg2 = LTAG_NIL;
    reg3 = LTAG_NIL;
    call_cc_arg.obj = LTAG_NIL;
    top_of_aflk.obj = LTAG_NIL;
    end_of_aflk.obj = LTAG_NIL;
    af_tmp.obj = LTAG_NIL;
    more.obj = LTAG_NIL;
    optcnt.obj = LTAG_NIL;

    open_frame2(bc_intrp_future2, future, 0);
    s_push(fut->body.obj);
    if (fut->env.obj != LTAG_NIL)
        s_push(fut->env.obj);
    active_reg = base_reg;
    bcpc_reg = fix_to_int(fut->start_addr.obj);
    bcbody_reg =
        (unsigned int *)ADDRESS(((BCCLOS *)ADDRESS(fut->body.obj))->body.obj);

    while (1)
        ((int (*)())(*(bcbody_reg + bcpc_reg)))();

#ifdef PTHREAD
# ifndef LTC
    pthread_cleanup_pop(0);     /* dummy */
# endif
#endif

}

void bc_intrp_not_future2()
{
    FUTURE *fut;
    REGS *save;

    if (GETtag(*(sp_reg+1)) != LTAG_FUTURE)
        bokkan("not future!! in bc_intrp_not_future2()");
    save = (REGS *)s_pop();
    fut = (FUTURE *)(ADDRESS(s_pop()));
    mutex_lock(fut->lock);
    if (fut->status == DETERMINED) {
        fut->value.obj = reg_reg;
        condition_broadcast(fut->done);
    } else {
        fut->value.obj = reg_reg;
        fut->status = DETERMINED;
        condition_broadcast(fut->done);
    }
    mutex_unlock(fut->lock);
    if (optcnt.obj != LTAG_NIL) {
        if (GETtag(CARobj(optcnt.obj)) == LTAG_CONTINUATION) {
            more.obj = ((CONTIN *)ADDRESS(CARobj(optcnt.obj)))->body.obj;
            optcnt.obj = CDRobj(optcnt.obj);
            sp_reg = (unsigned int *)stack_bottom;
            top();
            return;
        } else {
            open_frame2(bc_intrp_not_future2,
                        LTAG_FUTURE | SSERDDA(fut),
                        (int)save);
            s_push(CARobj(optcnt.obj));
            optcnt.obj = CDRobj(optcnt.obj);
            s_push(reg_reg);
            xcall();
            return;
        }
    }
    free(stack_bottom - CHILD_STACK_SIZE + 28);
    free(cthread_data(cthread_self()));
    cthread_set_data(cthread_self(), (any_t)save);
#ifdef USE_REG
# if defined(LUNA)
/*    active_reg = save->_active_reg; */
    base_reg = save->_base_reg;
    sp_reg = save->_sp_reg;
    bcpc_reg = save->_bcpc_reg;
    bcbody_reg = save->_bcbody_reg;
# elif defined(SUN)
    *main_thread_regs = *save;
# endif
#endif
    reg_reg = LTAG_FUTURE | SSERDDA(fut);
    bcpc_reg += 2;
}

#ifdef CALLIOC

void bc_intrp_not_future2_ep()
{
    FUTURE *fut;
    REGS *save;

    save = (REGS *)s_pop();
    fut = (FUTURE *)ADDRESS(*((unsigned int *)ADDRESS(s_pop())+1));
    mutex_lock(fut->lock);
    if (fut->status == DETERMINED) {
        fut->value.obj = reg_reg;
        condition_broadcast(fut->done);
    } else {
        fut->value.obj = reg_reg;
        fut->status = DETERMINED;
        condition_broadcast(fut->done);
    }
    mutex_unlock(fut->lock);
    if (optcnt.obj != LTAG_NIL) {
        if (GETtag(CARobj(optcnt.obj)) == LTAG_CONTINUATION) {
            more.obj = ((CONTIN *)ADDRESS(CARobj(optcnt.obj)))->body.obj;
            optcnt.obj = CDRobj(optcnt.obj);
            sp_reg = (unsigned int *)stack_bottom;
            top();
            return;
        } else {
            open_frame2(bc_intrp_not_future2,
                        LTAG_FUTURE | SSERDDA(fut),
                        (int)save);
            s_push(CARobj(optcnt.obj));
            optcnt.obj = CDRobj(optcnt.obj);
            s_push(reg_reg);
            xcall();
            return;
        }
    }
    free(stack_bottom - CHILD_STACK_SIZE + 28);
    free(cthread_data(cthread_self()));
    cthread_set_data(cthread_self(), (any_t)save);
/*    active_reg = save->_active_reg; */
    base_reg = save->_base_reg;
    sp_reg = save->_sp_reg;
    reg_reg = LTAG_FUTURE | SSERDDA(fut);
    bcpc_reg = save->_bcpc_reg;
    bcbody_reg = save->_bcbody_reg;
    bcpc_reg += 2;
}

#endif

void bc_interprete_not_future(future, save)
unsigned int *future, *save;
{
    FUTURE *fut;
    REGS *regsp;
    extern int exec_sysfun();

    /* for DEBUG */
    printf("*INLINED*");
    fflush(stdout);

    if (GETtag(future) != LTAG_FUTURE)
        bokkan("not future object!");

    fut = (FUTURE *)ADDRESS(future);

#if defined(sparc) && defined(USE_REG)
    main_thread_regs = (REGS *)malloc(sizeof(REGS));
#else
    regsp = (REGS *)malloc(sizeof(REGS));
    cthread_set_data(cthread_self(), (any_t)regsp);
#endif

    fut->status = UNDETERMINED;

#ifndef PTHREAD
    fut->thread_id = NULL;
#endif

    stack_top = (char *)malloc(CHILD_STACK_SIZE);
    stack_bottom = stack_top + CHILD_STACK_SIZE - 28;
    stack_top = stack_top + CHILD_GETA_OF_STACK;

    mutex_unlock(fut->lock);

    *((unsigned int **)stack_bottom + 1) = (unsigned int *)LTAG_NIL;
    *((unsigned int **)stack_bottom + 2) = 0;
    *((unsigned int **)stack_bottom + 3) = 0;

    stack_area_bottom = stack_bottom;
    set_top(stack_bottom);
    s_push(top);
    s_push(0);
    s_push(0);
    set_base(top_val());
    set_active(top_val());
    set_a2(LTAG_NIL);
    set_reg(LTAG_NIL);

    old_top = (unsigned int *)NULL;
    old_top1 = (unsigned int *)NULL;
    reg1 = LTAG_NIL;
    reg2 = LTAG_NIL;
    reg3 = LTAG_NIL;
    call_cc_arg.obj = LTAG_NIL;
    top_of_aflk.obj = LTAG_NIL;
    end_of_aflk.obj = LTAG_NIL;
    af_tmp.obj = LTAG_NIL;
    more.obj = LTAG_NIL;
    optcnt.obj = LTAG_NIL;

    open_frame2(bc_intrp_not_future2, future, (int)save);
    s_push(fut->body.obj);
    if (fut->env.obj != LTAG_NIL)
        s_push(fut->env.obj);
    active_reg = base_reg;
    bcpc_reg = fix_to_int(fut->start_addr.obj);
    bcbody_reg = (unsigned int *)ADDRESS(((BCCLOS *)ADDRESS(fut->body.obj))->body.obj);
}

#endif /* ifndef SHE */

#ifndef SHE

/*
 *    Touch-set!
 */

/**
void terminate_fut()
{
    printf("terminate process\n");
    fflush(stdout);
    cthread_exit(NULL);
}
**/

Stouch_setq()
{
    SOBJ obj;
    FUTURE *fut;

#ifdef LTC
    obj.obj = getarg(1);
    if (GETtag(obj.obj) != LTAG_FUTURE)
        set_reg(LTAG_FALSE);
    else {
        fut = (FUTURE *)(ADDRESS(obj.obj));
        mutex_lock(fut->lock);
        fut->value.obj = getarg(2);
        fut->status = DETERMINED;
        condition_broadcast(fut->done);
        mutex_unlock(fut->lock);
        set_reg(fut->value.obj);
    }
#else
/**
    struct luna88k_thread_state state;
    register struct luna88k_thread_state *ts = &state;
**/

    obj.obj = getarg(1);
    if (GETtag(obj.obj) != LTAG_FUTURE)
        bokkan("1st argument must be promise.");
    fut = (FUTURE *)(ADDRESS(obj.obj));
    mutex_lock(fut->lock);
    if (fut->status == DETERMINED) {
        set_reg(LTAG_FALSE);
        condition_broadcast(fut->done);
    } else if (fut->status == NO_PROCESS) {
    /* for promise that generated by make-placeholder */
        fut->value.obj = getarg(2);
        fut->status = DETERMINED;
        condition_broadcast(fut->done);
        set_reg(fut->value.obj);
    } else {
        fut->value.obj = getarg(2);
        fut->status = DETERMINED;
/*      cthread_detach(fut->proc_id); */
        if (cthread_self() == fut->proc_id) {
            condition_broadcast(fut->done);
            set_reg(fut->value.obj);
            mutex_lock(cthread_fork_lock);
            current_thread_no--;
            mutex_unlock(cthread_fork_lock);
            mutex_unlock(fut->lock);
# ifdef PTHREAD
            /* release stack and regs by cleanup handler */
# else
            free(stack_bottom - CHILD_STACK_SIZE + 28);
            free(cthread_data(cthread_self()));
# endif
            cthread_exit(NULL);
        }
        mutex_lock(read_lock);
        mutex_lock(print_lock);
        mutex_lock(addsymbol_lock);
        mutex_lock(memget_lock);
        mutex_lock(cthread_fork_lock);

# ifdef PTHREAD
        pthread_cancel(fut->proc_id);
# else
        if (thread_suspend(fut->thread_id) != KERN_SUCCESS)
            print_con("Can't suspend process!!!");
        if (thread_abort(fut->thread_id) != KERN_SUCCESS)
            print_con("Can't abort process!!!");
/**
        thread_get_state(fut->thread_id,
                         LUNA88K_THREAD_STATE,
                         (thread_state_t)&state,
                         LUNA88K_THREAD_STATE_COUNT);
        ts->fpcr = terminate_fut;
        thread_set_state(fut->thread_id,
                         LUNA88K_THREAD_STATE,
                         (thread_state_t)&state,
                         LUNA88K_THREAD_STATE_COUNT);
        if (thread_resume(fut->thread_id) != KERN_SUCCESS)
            print_con("Can't resume process!!!");
**/
        if (thread_terminate(fut->thread_id) != KERN_SUCCESS)
            print_con("Can't terminate process!!!");
# endif

        condition_broadcast(fut->done);
        set_reg(fut->value.obj);
        current_thread_no--;
# ifdef PTHREAD
        /* release stack and regs by cleanup handler */
# else
        free(((REGS *)cthread_data(fut->proc_id))->_stack_bottom
             - CHILD_STACK_SIZE + 28);
        free(cthread_data(fut->proc_id));
# endif

        mutex_unlock(cthread_fork_lock);
        mutex_unlock(memget_lock);
        mutex_unlock(addsymbol_lock);
        mutex_unlock(print_lock);
        mutex_unlock(read_lock);
    }
    mutex_unlock(fut->lock);
    /* deallocate stack */
    /* free(...); */
    /* free(cthread_data(fut->proc_id)); */
#endif /* LTC */
}

#endif /* ifndef SHE */

/*
 *    Yield
 */

Syield()
{
#ifdef PTHREAD
# ifdef PGCC
    pthread_yield();
# elif defined(__PTL__)
# elif defined(SUNSOL2)
    sched_yield();
# elif defined(FREEBSD)
    pthread_yield();
# else
    pthread_yield(NULL);
# endif
#endif
    set_reg(LTAG_T);
}

/*
 *    Suspend
 */

Ssuspend()
{
#ifdef PTHREAD
/*    pthread_suspend_np(pthread_self()); */
    printf("Suspend is not supported\n");
#else
    if (thread_suspend(thread_self()) != KERN_SUCCESS)
        print_con("Can't suspend!!!");
#endif
}

/*
 *    Quit
 */

cthread_t main_cthread_id;

Squit()
{
    mutex_lock(cthread_fork_lock);
    current_thread_no--;
    mutex_unlock(cthread_fork_lock);
    if (cthread_self() == main_cthread_id) {
#ifdef PTHREAD
/*        pthread_suspend_np(pthread_self()); */
        printf("Suspend is not supported\n");

#else
        if (thread_suspend(thread_self()) != KERN_SUCCESS)
            print_con("Can't suspend!!!");
#endif
    } else
        cthread_exit(NULL);
}

/*
 *    Mutual exclusion
 */

int Smake_mutex()
{
    set_reg(mk_mutex());
}

int Smutex_lock()
{
    SOBJ obj;
    LOCK *lk;

#ifdef TOUCH_ARG
    obj.obj = touch_obj(getarg(1));
#else
    obj.obj = getarg(1);
#endif
    if (GETtag(obj.obj) != LTAG_LOCK)
        bokkan("Argment must be mutex object.");
    lk = (LOCK *)ADDRESS(obj.obj);

/*    mutex_lock(lk->lock1); */
    while (!mutex_try_lock(lk->lock1))
        ;
    while (!mutex_try_lock(lk->lock2))
        condition_wait(lk->c1, lk->lock1);
    set_reg(LTAG_T);
    mutex_unlock(lk->lock1);
}

int Smutex_unlock()
{
    SOBJ obj;
    LOCK *lk;

#ifdef TOUCH_ARG
    obj.obj = touch_obj(getarg(1));
#else
    obj.obj = getarg(1);
#endif
    if (GETtag(obj.obj) != LTAG_LOCK)
        bokkan("Argment must be mutex object.");
    lk = (LOCK *)(ADDRESS(obj.obj));
/*    mutex_lock(lk->lock1); */
    while (!mutex_try_lock(lk->lock1))
        ;
    mutex_unlock(lk->lock2);
    condition_signal(lk->c1);
    set_reg(LTAG_T);
    mutex_unlock(lk->lock1);
}

Swait()
{
    SOBJ obj;
    LOCK *lk;

#ifdef TOUCH_ARG
    obj.obj = touch_obj(getarg(1));
#else
    obj.obj = getarg(1);
#endif
    if (GETtag(obj.obj) != LTAG_LOCK)
        bokkan("Argument must be mutex object.");
    lk = (LOCK *)ADDRESS(obj.obj);
    mutex_lock(lk->lock1);
    mutex_unlock(lk->lock2);
    condition_signal(lk->c1);
    condition_wait(lk->c2, lk->lock1);
#ifdef TOUCH_ARG
    lk = (LOCK *)ADDRESS(touch_obj(getarg(1))); /* for GC (cf. plambda) */
#else
    lk = (LOCK *)ADDRESS(getarg(1)); /* for GC (cf. plambda) */
#endif
    while (!mutex_try_lock(lk->lock2)) {
        condition_wait(lk->c1, lk->lock1);
#ifdef TOUCH_ARG
        lk = (LOCK *)ADDRESS(touch_obj(getarg(1))); /* for GC (cf. plambda) */
#else
        lk = (LOCK *)ADDRESS(getarg(1)); /* for GC (cf. plambda) */
#endif
    }
    set_reg(lk->val.obj);
    mutex_unlock(lk->lock1);
}

Ssignal()
{
    SOBJ obj;
    LOCK *lk;

#ifdef TOUCH_ARG
    obj.obj = touch_obj(getarg(1));
#else
    obj.obj = getarg(1);
#endif
    if (GETtag(obj.obj) != LTAG_LOCK)
        bokkan("Argument must be mutex object.");
    lk = (LOCK *)ADDRESS(obj.obj);
    mutex_lock(lk->lock1);
    lk->val.obj = getarg(2);
    set_reg(lk->val.obj);
    condition_signal(lk->c2);
    mutex_unlock(lk->lock1);
}

/*
 *    Misc.
 */

Sfuturep()
{
    if (GETtag(getarg(1)) == LTAG_FUTURE)
        set_reg(LTAG_T);
    else
        set_reg(LTAG_FALSE);
}

#ifdef SHE

int Sdeterminedp()
{
    ENTITYP fut;

    if (GETtag(getarg(1)) == LTAG_FUTURE) {
        fut = (ENTITYP)ADDRESS(getarg(1));
        if (fut->future.sstfr == NULL)
            set_reg(LTAG_T);
        else 
            set_reg(LTAG_FALSE);
    } else {
        set_reg(LTAG_T);
    }

    return 0;
}

#else

int Sdeterminedp()
{
    ENTITYP fut;

    if (GETtag(getarg(1)) == LTAG_FUTURE) {
        fut = (ENTITYP)ADDRESS(getarg(1));
        if (fut->future.status == DETERMINED)
            set_reg(LTAG_T);
        else 
            set_reg(LTAG_FALSE);
    } else {
        set_reg(LTAG_T);
    }

    return 0;
}

#endif

Sthe_num_of_current_thread()
{
    set_reg(mk_fix(current_thread_no));
}

Sthread_id()
{
#ifdef PTHREAD
    set_reg(mk_fix(pthread_self()));
#else
    set_reg(mk_fix(thread_self()));
#endif
}

Stouch()
{
    set_reg(touch_obj(getarg(1)));
}

#ifndef SHE

Suntouch()
{
    SOBJ obj;
    FUTURE *fut;

    obj.obj = getarg(1);
    if (GETtag(obj.obj) != LTAG_FUTURE)
        bokkan("1st argument must be promise.");
    fut = (FUTURE *)(ADDRESS(obj.obj));
    mutex_lock(fut->lock);
    if (fut->status == DETERMINED) {
        fut->status = NO_PROCESS;
        set_reg(LTAG_T);
    } else
        set_reg(LTAG_FALSE);
    mutex_unlock(fut->lock);
}

#endif

Smutexp()
{
#ifdef TOUCH_ARG
    if (GETtag(touch_obj(getarg(1))) == LTAG_LOCK)
#else
    if (GETtag(getarg(1)) == LTAG_LOCK)
#endif
        set_reg(LTAG_T);
    else
        set_reg(LTAG_FALSE);
}

/*
 *    Make future object
 */

mk_mutex()
{
    LOCK *lock;

    lock = (LOCK *)memget(sizeof(LOCK));
    lock->head = LTAG_BLOCK1 | (unsigned int)24;
    lock->val.obj = LTAG_NIL;
    lock->block.obj = LTAG_BLOCK2 | 16;
    lock->lock1 = mutex_alloc();
    lock->lock2 = mutex_alloc();
    lock->c1 = condition_alloc();
    lock->c2 = condition_alloc();
    return(LTAG_LOCK | SSERDDA(lock));
}

get_cc_one()             /* get one active frame of current continuation */
{
    SOBJ ret_val;
    SOBJP a_frame;
    int (*ret_addr)();
    int frame_size, tmp_frame, pc, active, i;
    unsigned int *save_active_reg, *save_base_reg, *save_sp_reg;

    if (top_val() != (unsigned int *)stack_bottom - 3) { /* by T.Komiya */
        top_of_aflk.obj = end_of_aflk.obj = mk_cons(LTAG_NIL, LTAG_NIL);
        af_tmp.obj = LTAG_NIL;
        save_sp_reg = sp_reg;
        save_active_reg = active_reg;
        save_base_reg = base_reg;
/*   open_frame2(LTAG_NIL, LTAG_NIL, 0); */
        *(--save_sp_reg) = LTAG_NIL;
        *(--save_sp_reg) = 0;        
        *(--save_sp_reg) = LTAG_NIL;
        *(--save_sp_reg) = (unsigned int)save_active_reg;
        *(--save_sp_reg) = (unsigned int)save_base_reg;
        save_base_reg = save_sp_reg;

        save_base_reg = (unsigned int *)*save_sp_reg++;    /* recovering a base frame */
        save_active_reg = (unsigned int *)*save_sp_reg++;  /* recovering an active frame */
        do /* loop for active links */
        {
            ret_addr = (int (*)())(*save_sp_reg++);
            pc = *save_sp_reg++;
            frame_size = ((unsigned int)save_base_reg - (unsigned int)save_sp_reg) / 4;
            a_frame = (SOBJP)mk_evecp(frame_size + 2);
            (a_frame + frame_size + 2)->obj = mk_fix((int)ret_addr);
            (a_frame + (frame_size + 1))->obj = mk_fix(pc);
            /* loop for elements in the frame */
            for (; frame_size > 0 ; frame_size--)
                (a_frame + frame_size)->obj = *save_sp_reg++;
            af_tmp.obj = mk_cons((LTAG_VECTOR | SSERDDA(a_frame)),
                                 af_tmp.obj);
            if ((unsigned int)save_sp_reg < (unsigned int)save_active_reg) {
                save_base_reg = (unsigned int *)*save_sp_reg++;
                save_active_reg = (unsigned int *)*save_sp_reg++;
                af_tmp.obj = mk_cons(mk_fix(((unsigned int)active - 
                                             (unsigned int)save_sp_reg) / 4),
                                     af_tmp.obj);
            } else
                break;
        } while (1);
        set_car(end_of_aflk.obj, af_tmp.obj);
        set_cdr(end_of_aflk.obj, more.obj);
    } else
        return(LTAG_NIL);
    ret_val.obj = top_of_aflk.obj;
    top_of_aflk.obj = end_of_aflk.obj = af_tmp.obj = LTAG_NIL;
    return(ret_val.obj);
}

#ifndef SHE

/*
** for Lazy Task Creation
**   06 Aug 1997
*/
#ifdef LTC
FUTURE *mk_plchld()
#else
FUTURE *mk_plchld(body, env, start_address)
unsigned int *body, *env;
int start_address;
#endif
{
    FUTURE *fut;

#ifndef LTC
    s_push(env);
    s_push(body);
#endif
    fut = (FUTURE *)memget(sizeof(FUTURE));
    fut->head = LTAG_BLOCK1 | (unsigned int)44;
#ifndef LTC
    fut->body.obj = s_pop();
    fut->start_addr.obj = mk_fix(start_address);
    fut->env.obj = s_pop();
#endif
    fut->value.obj = unbound_obj.obj;
    fut->block.obj = LTAG_BLOCK2 | 24;
    fut->status = NO_PROCESS;
    fut->lock = mutex_alloc();
    fut->done = condition_alloc();
    return(fut);
}


free_mem()
{
    REGS *save;

    printf("free!\n");
    save = (REGS *)s_pop();
    s_pop();
    free(stack_bottom - CHILD_STACK_SIZE + 28);
    free(cthread_data(cthread_self()));
    cthread_set_data(cthread_self(), (any_t)save);
    return(s_pop());
}

mk_future(body, env, start_address)
unsigned int *body, *env;
int start_address;
{
    extern void bc_interprete_future();
    FUTURE *fut;
    REGS *save;

    fut = mk_plchld(body, env, start_address);
    fut->status = NOT_READY;
    /* regsp = (REGS *)malloc(sizeof(REGS)); */

    mutex_lock(fut->lock);

    if (current_thread_no < MAX_THREAD_NO) {
        s_push(LTAG_FUTURE | SSERDDA(fut));

#ifdef PTHREAD
        fut->proc_id = cthread_fork(bc_interprete_future,
                                    LTAG_FUTURE | SSERDDA(fut));
        cthread_detach(fut->proc_id);
#else
        cthread_detach(fut->proc_id
                       = cthread_fork(bc_interprete_future,
                                      LTAG_FUTURE | SSERDDA(fut)));
#endif

        mutex_lock(cthread_fork_lock);
        current_thread_no++;
        mutex_unlock(cthread_fork_lock);
        /*
        regsp->_more.obj = LTAG_NIL;
        cthread_set_data(fut->proc_id, (any_t)regsp);
        fut->status = UNDETERMINED;
        */
        return(s_pop());

    } else {
        /*
        fut->status = UNDETERMINED;
        regsp->_more.obj = LTAG_NIL;
        */
#if !defined(sparc) || !defined(USE_REG)
        save = (REGS *)cthread_data(cthread_self());
#endif

#ifdef USE_REG
# if defined(LUNA)
        /* save->_active_reg = active_reg; */
        save->_base_reg = base_reg;
        save->_sp_reg = sp_reg;
        save->_reg_reg = reg_reg;
        save->_bcpc_reg = bcpc_reg;
        save->_bcbody_reg = bcbody_reg;
# elif defined(sparc)
        *save = *main_thread_regs;
# endif
#endif

        /* cthread_set_data(cthread_self(), (any_t)regsp); */
        fut->proc_id = cthread_self();
        bc_interprete_not_future(LTAG_FUTURE | SSERDDA(fut), save);
        bcpc_reg -= 2;
        return NULL;
    }
}


mk_spawn(prom, body, env, start_address)
unsigned int *prom, *body, *env;
int start_address;
{
    extern void bc_interprete_future();
    FUTURE *fut;
    REGS *regsp;

    if (GETtag(prom) != LTAG_FUTURE)
        bokkan("not future!! in mk_spawn().");
    fut = (FUTURE *)(ADDRESS(prom));
    fut->body.obj = (unsigned int)body;
    fut->env.obj = (unsigned int)env;
    fut->start_addr.obj = mk_fix(start_address);
    fut->status = NOT_READY;
    regsp = (REGS *)malloc(sizeof(REGS));

    mutex_lock(fut->lock);
    s_push(LTAG_FUTURE | SSERDDA(fut));
#ifdef PTHREAD
    fut->proc_id = cthread_fork(bc_interprete_future,
                                LTAG_FUTURE | SSERDDA(fut));
    cthread_detach(fut->proc_id);
#else
    cthread_detach(fut->proc_id
                   = cthread_fork(bc_interprete_future,
                                  LTAG_FUTURE | SSERDDA(fut)));
#endif
    mutex_lock(cthread_fork_lock);
    current_thread_no++;
    mutex_unlock(cthread_fork_lock);
    return(s_pop());
}

#endif /* ifndef SHE */

#endif
