/*
 * inet.c - 
 */

/*
 * $Id: inet.c,v 1.10 2014/10/28 02:22:55 komiya Exp $
 *
 * Networking functions.
 *
 *		N.Kurabayashi
 */

static char *RCSid="$Id: inet.c,v 1.10 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"

#ifdef TCPIP
#include <errno.h>

#include "inet.h"

#ifndef __CYGWIN32__
#include <protocols/rwhod.h>
#endif
#include <sys/time.h>
#include <sys/timeb.h>
#ifdef SUNOS4
#include <dirent.h>
#endif
#include <fcntl.h>
#include <string.h>
//#include <varargs.h>
#include <stdarg.h>

#if defined(__CYGWIN32__) || defined(linux)
sigjmp_buf timeout_env;
#else
jmp_buf timeout_env;
#endif
extern char *ourhostname;
extern int errno;

#define STRING(p)    mk_str (ORDINARY_STRING, A_STRING, strlen (p), p)

/* Size of queue buffer for the socket in bytes
 * (under 4.3BSD, upper limit is 52000).  */
static int RCVBUFSIZ = 8192;	/* receive queue buffer */
static int SNDBUFSIZ = 8192;	/* send queue buffer */

unsigned mk_fix();

vbokkan (char *fmt, ...)
{
  va_list args;
  char msgbuf[128];

  va_start (args, fmt);
  vsprintf (msgbuf, fmt, args);
  va_end (args);

  bokkan (msgbuf);
}

/*  old C version
vbokkan (va_alist)
     va_dcl
{
  va_list args;
  char *fmt;
  char msgbuf[128];

  va_start (args);
  fmt = va_arg (args, char *);
  vsprintf (msgbuf, fmt, args);
  va_end (args);

  bokkan (msgbuf);
}
*/

int
mk_netport (kind, sockfd, addrp)
     int kind, sockfd;
     struct sockaddr_in *addrp;
{
  ENTITYP port;
  int netport;
  int weakpair;
  char addrbuf[sizeof (struct sockaddr_in) + 1];

  bcopy ((char *) addrp, addrbuf, sizeof (*addrp));
  addrbuf[sizeof (*addrp) + 1] = '\0';

  reg1 = mk_str (ORDINARY_STRING, A_STRING,
		 sizeof (struct sockaddr_in), addrbuf);

  port = (ENTITYP) memget (sizeof (PORT));

  port->port.head = LTAG_BLOCK1 | (unsigned)24;

  port->port.slot.obj = reg1;

  port->port.block.com = LTAG_BLOCK2 | (unsigned)16;

  port->port.fp    = (FILE *) sockfd;
  port->port.kind  = kind;
  port->port.bufstatus = 0;
  port->port.dummy = 0;		/* should be cleared for UDP_BOUND */
  port->port.buf2  = 0;
  port->port.buf1  = 0;

#ifdef NIHONGO
  port->port.kanji_input_code =
	  fix_to_int(get_sval(get_sval(default_network_kanji_input_code_sym.obj)));
  port->port.kanji_output_code =
	  fix_to_int(get_sval(get_sval(default_network_kanji_output_code_sym.obj)));
#endif

#if 0
  /* link to port-weak-list */
  reg1 = netport = LTAG_PORT | SSERDDA(port);
  weakpair = mk_wp (reg1, get_sval (portlist.obj));
  set_sval (portlist.obj, weakpair);
  set_sval (wplist.obj, weakpair);
  reg1 = LTAG_NIL;

  return netport;

#else

  return LTAG_PORT | SSERDDA(port);
#endif
}


struct sockaddr_in *
getnetaddr (port)
     int port;
{
  int *pe;

  pe = (int *) ADDRESS (port);
  return (struct sockaddr_in *) ((char *) ADDRESS (*(pe + 1)) + 12);
}


unsigned
timeoutsec (svar)
     int svar;
{
  SOBJ timeout;

  timeout.obj = get_sval (svar);
  if (GETtag (timeout.obj) == LTAG_FIXNUM)
    return (unsigned) fix_to_int (timeout.obj);
  else
    return (unsigned) 0;
}


set_socket_options (sd)
     int sd;
{
  if (setsockopt (sd, SOL_SOCKET, SO_RCVBUF,
		  (char *) &RCVBUFSIZ, sizeof (RCVBUFSIZ)) < 0)
    vbokkan (";setsockopt failed (SO_RCVBUF) (errno = %d).",
	     errno);

  if (setsockopt (sd, SOL_SOCKET, SO_SNDBUF,
		  (char *) &SNDBUFSIZ, sizeof (SNDBUFSIZ)) < 0)
    vbokkan (";setsockopt failed (SO_SNDBUF) (errno = %d).",
	     errno);
}


#define PASSIVE_OPEN	1
#define ACTIVE_OPEN	2


int
setup_netaddr (mode, ptype, addrp, hostaddr)
     int mode, ptype;
     struct sockaddr_in *addrp;
     long hostaddr;
{
  SOBJ portnum;
  int proto, pn, sd;
  char *protoname, *servname;
  struct servent *sp;
  struct sockaddr_in addr;
  extern char *strobj_to_str ();

  switch (ptype)
    {
    case UDP_DGRAM:
      proto = SOCK_DGRAM;
      protoname = "udp";
      break;
    case TCP_STREAM:
    case TCP_BOUND:
      proto = SOCK_STREAM;
      protoname = "tcp";
      break;
    default:
      vbokkan (";Internal error at open_netport (mode = %d).", mode);
    }

#ifdef TOUCH_ARG
  portnum.obj = touch_obj(getarg (1));
#else
  portnum.obj = getarg (1);
#endif
  switch (GETtag (portnum.obj))
    {
    case LTAG_FIXNUM:	/* positive fixnum only */
      if ((pn = fix_to_int (portnum.obj)) > IPPORT_ULIMIT ||
	  (mode != ACTIVE_OPEN && pn < IPPORT_RESERVED))
	vbokkan ("%d isn't a valid port number.", pn);
      pn = htons (pn);
      break;
    case LTAG_STRING:
    case LTAG_SSTRING:
      servname = strobj_to_str (portnum.obj);
      if ((sp = getservbyname (servname, protoname)) == NULL)
	vbokkan ("%s isn't a %s service.", servname, protoname);
      pn = sp->s_port;
      break;
    default:
      err_illegal_argument ("1st argument must be a positive integer or a string.");
      return 0;                 /* never reach here */
    }

  if ((sd = socket (AF_INET, proto, 0)) < 0)
    {
#ifdef TCP_DEBUG
      perror ("setup_netaddr : socket");
#endif
      return -1;
    }

  set_socket_options (sd);

  bzero ((char *) addrp, sizeof (*addrp));
  addrp->sin_family = AF_INET;
  addrp->sin_addr.s_addr = hostaddr;
  addrp->sin_port = pn;

  return sd;
}


int
p_open_netport (ptype)
     int ptype;
{
  int sd, val;
  struct sockaddr_in addr;

#if defined(__CYGWIN32__) || defined(linux)
  if (sigsetjmp (timeout_env, 1) == 0)
#else
  if (setjmp (timeout_env) == 0)
#endif
    {
      alarm (timeoutsec (conn_timeout.obj));

      sd = setup_netaddr (PASSIVE_OPEN, ptype, &addr, htonl (INADDR_ANY));
      if (sd < 0)
	val = LTAG_FALSE;
      else if (bind (sd, (struct sockaddr *)&addr, sizeof (addr)) < 0)
	{
#ifdef TCP_DEBUG
	  perror ("p_open_netport : bind");
#endif
	  close (sd);
	  val = LTAG_FALSE;
	}
      else if (ptype == TCP_BOUND &&
	       listen (sd, SOMAXCONN) < 0)
	{
#ifdef TCP_DEBUG
	  perror ("p_open_netport : listen");
#endif
	  close (sd);
	  val = LTAG_FALSE;
	}
      else
	val = mk_netport (ptype, sd, &addr);
      alarm (0);
      return val;
    }
  else
    return LTAG_FALSE;
}


int
a_open_netport (ptype)
     int ptype;
{
  SOBJ host;
  char *hostname;
  struct hostent *hp;
  long hostaddr;
  int sd, val;
  struct sockaddr_in addr;
  extern char *strobj_to_str ();

  switch (getargnum ())
    {
    case 0:
      err_illegal_argument ("Too few arguments.");
      return 0;                 /* never reach here */
    case 1:
      hostaddr = htonl (INADDR_ANY);
      break;
    case 2:
#ifdef TOUCH_ARG
      host.obj = touch_obj(getarg (2));
#else
      host.obj = getarg (2);
#endif
      if (GETstr_tag (host.obj) != LTAG_STRING){
	  err_illegal_argument ("2nd argument must be a host name string.");
	  return 0;             /* never reach here */
      }
      hostname = strobj_to_str (host.obj);
      if ((hp = gethostbyname (hostname)) == NULL)
	vbokkan ("%s isn't a valid host name.", hostname);
      if (hp->h_addrtype != AF_INET)
	vbokkan ("Sorry, address type of `%s' isn't supported yet.",
		 hostname);
      if (hp->h_length != 4)
	vbokkan ("Sorry, address length of `%s' isn't supported yet.",
		 hostname);
      bcopy (hp->h_addr, (char *) &hostaddr, hp->h_length);
      break;
    default:
      err_illegal_argument ("Too many arguments.");
      return 0;                 /* never reach here */
    }

#if defined(__CYGWIN32__) || defined(linux)
  if (sigsetjmp (timeout_env, 1) == 0)
#else
  if (setjmp (timeout_env) == 0)
#endif
    {
      alarm (timeoutsec (conn_timeout.obj));
      if ((sd = setup_netaddr (ACTIVE_OPEN, ptype, &addr, hostaddr)) < 0)
	val = LTAG_FALSE;
      else
	{
	  if (ptype == UDP_DGRAM)
	    {
	      struct sockaddr_in myaddr;

	      bzero ((char *) &myaddr, sizeof (myaddr));
	      myaddr.sin_family = AF_INET;
	      myaddr.sin_addr.s_addr = htonl (INADDR_ANY);
	      myaddr.sin_port = htons (0);
	      if (bind (sd, (struct sockaddr *)&myaddr, sizeof (myaddr)) < 0)
		{
#ifdef TCP_DEBUG
		  perror ("a_open_netport : bind");
#endif
		  close (sd);
		  val = LTAG_FALSE;
		  goto done;
		}
	    }
	  else if (connect (sd, (struct sockaddr *)&addr, sizeof (addr)) < 0)
	    {
#ifdef TCP_DEBUG
	      perror ("a_open_netport : connect");
#endif
	      close (sd);
	      val = LTAG_FALSE;
	      goto done;
	    }

	  val = mk_netport (ptype, sd, &addr);
	}
    done:
      alarm (0);
      return val;
    }
  else
    return LTAG_FALSE;
}


void Sbind_udp_port ()
{
  SOBJ port;

  if ((port.obj = p_open_netport (UDP_DGRAM)) != LTAG_FALSE)
    ((PORT *) ADDRESS (port.obj))->dummy = UDP_BOUND;

  set_reg (port.obj);
}


void Sbind_tcp_port ()
{
  set_reg (p_open_netport (TCP_BOUND));
}


void Sopen_udp_port ()
{
  set_reg (a_open_netport (UDP_DGRAM));
}


void Sopen_tcp_port ()
{
  set_reg (a_open_netport (TCP_STREAM));
}


void Saccept ()
{
  SOBJ port;
  struct sockaddr_in cli_addr;
  int clilen = sizeof (cli_addr);
  int sd0, sd;

#ifdef TOUCH_ARG
  port.obj = touch_obj(getarg (1));
#else
  port.obj = getarg (1);
#endif
  if (!boundportp (port.obj)){
      err_illegal_argument ("Argument must be a bound TCP port.");
      return ;
  }
#if defined(__CYGWIN32__) || defined(linux)
  if (sigsetjmp (timeout_env, 1) == 0)
#else
  if (setjmp (timeout_env) == 0)
#endif
    {
      alarm (timeoutsec (accept_timeout.obj));
      sd0 = (int) ((PORT *) ADDRESS (port.obj))->fp;
      if ((sd = accept (sd0, (struct sockaddr *)&cli_addr, &clilen)) < 0)
	vbokkan (";accept failed (errno = %d).", errno);
      alarm (0);
      set_reg (mk_tcpport (sd, &cli_addr));
    }
  else
    set_reg (LTAG_FALSE);
}


void Sbound_udp_port_p ()
{
  SOBJ port;

#ifdef TOUCH_ARG
  port.obj = touch_obj(getarg (1));
#else
  port.obj = getarg (1);
#endif
  if (!udpportp (port.obj))
    set_reg (LTAG_FALSE);
  else if (((PORT *) ADDRESS (port.obj))->dummy == UDP_BOUND)
    set_reg (LTAG_T);
  else
    set_reg (LTAG_FALSE);
}


void Sbound_tcp_port_p ()
{
#ifdef TOUCH_ARG
  set_reg (boundportp (touch_obj(getarg (1))) ? LTAG_T : LTAG_FALSE);
#else
  set_reg (boundportp (getarg (1)) ? LTAG_T : LTAG_FALSE);
#endif
}


void Sudp_port_p ()
{
#ifdef TOUCH_ARG
  set_reg (udpportp (touch_obj(getarg (1))) ? LTAG_T : LTAG_FALSE);
#else
  set_reg (udpportp (getarg (1)) ? LTAG_T : LTAG_FALSE);
#endif
}


void Stcp_port_p ()
{
#ifdef TOUCH_ARG
  set_reg (tcpportp (touch_obj(getarg (1))) ? LTAG_T : LTAG_FALSE);
#else
  set_reg (tcpportp (getarg (1)) ? LTAG_T : LTAG_FALSE);
#endif
}


void Sacceptable_p ()
{
  SOBJ port;
  fd_set readfds;
  struct timeval timeout;
  int fd;

#ifdef TOUCH_ARG
  port.obj = touch_obj(getarg (1));
#else
  port.obj = getarg (1);
#endif
  if (!boundportp (port.obj)){
      err_illegal_argument ("Argument must be a bound TCP port.");
      return ;
  }

  fd = (int) ((PORT *) ADDRESS (port.obj))->fp;
  FD_ZERO (&readfds);
  FD_SET (fd, &readfds);
  timeout.tv_sec = 0L;
  timeout.tv_usec = 0L;

  if (select (getdtablesize (), &readfds, NULL, NULL, &timeout) < 0)
    vbokkan (";select failed (errno = %d).", errno);

  set_reg (FD_ISSET (fd, &readfds) ? LTAG_T : LTAG_FALSE);
}


int
close_netport (ptype, errmsg)
     int ptype;
     char *errmsg;
{
  SOBJ port;
  int kind;

#ifdef TOUCH_ARG
  port.obj = touch_obj(getarg (1));
#else
  port.obj = getarg (1);
#endif
  if (GETtag (port.obj) == LTAG_PORT)
    {
      if ((kind = ((PORT *) ADDRESS (port.obj))->kind) == CLOSE)
	return LTAG_FALSE;
      else if (kind == ptype)
	{
	  closeport (port.obj);
	  return LTAG_T;
	}
    }
  vbokkan ("Argument must be a %s port.", errmsg);
}


void Sclose_b_udp_port ()
{
  SOBJ port;

#ifdef TOUCH_ARG
  port.obj = touch_obj(getarg (1));
#else
  port.obj = getarg (1);
#endif
  if (udpportp (port.obj) &&
      ((PORT *) ADDRESS (port.obj))->dummy == UDP_BOUND)
    set_reg (close_netport (UDP_DGRAM, "bound UDP"));
  else {
      err_illegal_argument ("Argument must be a bound UDP port.");
      return ;
  }
}


void Sclose_b_tcp_port ()
{
  set_reg (close_netport (TCP_BOUND, "bound TCP"));
}


void Sclose_udp_port ()
{
  set_reg (close_netport (UDP_DGRAM, "UDP"));
}


void Sclose_tcp_port ()
{
  set_reg (close_netport (TCP_STREAM, "TCP"));
}


void Sport_number ()
{
  SOBJ port;
  struct sockaddr_in *addrp;

#ifdef TOUCH_ARG
  port.obj = touch_obj(getarg (1));
#else
  port.obj = getarg (1);
#endif
  if (GETtag (port.obj) == LTAG_PORT &&
      (_netportp (port.obj) || _boundportp (port.obj)))
    {
      addrp = getnetaddr (port.obj);
      set_reg (mk_fix (ntohs (addrp->sin_port)));
    }
  else {
      err_illegal_argument ("Argument must be a network port or a bound port.");
      return ;
  }
}


void Shost_name ()
{
  SOBJ port;
  struct sockaddr_in *addrp;
  struct hostent *hp;

#ifdef TOUCH_ARG
  port.obj = touch_obj(getarg (1));
#else
  port.obj = getarg (1);
#endif
  if (GETtag (port.obj) == LTAG_PORT &&
      (_netportp (port.obj) || _boundportp (port.obj)))
    {
      addrp = getnetaddr (port.obj);
#ifdef FREEBSD
      if ((hp = gethostbyaddr (&(addrp->sin_addr.s_addr), 4, AF_INET)) == NULL)
#else
      if ((hp = gethostbyaddr ((void *)&(addrp->sin_addr), 4, AF_INET)) == NULL)
#endif
	set_reg (LTAG_FALSE);
      else
	set_reg (mk_str (ORDINARY_STRING, A_STRING,
			 strlen (hp->h_name), hp->h_name));
    }
  else {
      err_illegal_argument ("Argument must be a network port or a bound port.");
      return ;
  }
}


void Speer ()
{
  SOBJ port, head;
  struct sockaddr_in addr;
  struct hostent *hp;
  int sd, addrlen;

#ifdef TOUCH_ARG
  port.obj = touch_obj(getarg (1));
#else
  port.obj = getarg (1);
#endif
  if (!netportp (port.obj)){
      err_illegal_argument ("Argument must be a network port.");
      return ;
  }
  sd = (int) ((PORT *) ADDRESS (port.obj))->fp;
  addrlen = sizeof (addr);

  if (getpeername (sd, (struct sockaddr *)&addr, &addrlen) < 0 ||
      (addrlen = sizeof (addr),
       (hp = gethostbyaddr ((void *)&addr, addrlen, AF_INET)) == NULL))
    set_reg (LTAG_FALSE);
  else
    {
#define LIST2(a, b)  mk_cons (a, mk_cons (b, LTAG_NIL))

      set_reg (LIST2 (STRING (hp->h_name), mk_fix (ntohs (addr.sin_port))));
    }
}

#ifdef SUNOS4

#define statdir		"/var/spool/rwho/"
#define statbase	"whod."

#define MAYBEDOWN	(6 * 60)	/* in seconds */
#define down_p(now,wd)	((now) - (wd).wd_recvtime > MAYBEDOWN)


void Shost_up_p ()
{
  SOBJ host;
  char statfile[MAXPATHLEN];
  int fd;
  time_t now;
  struct whod wd;
  char *strobj_to_str ();

#ifdef TOUCH_ARG
  host.obj = touch_obj(getarg (1));
#else
  host.obj = getarg (1);
#endif
  if (GETstr_tag (host.obj) != LTAG_STRING){
      err_illegal_argument ("Argument must be a host name string.");
      return ;
  }
  strcpy (statfile, statdir);
  strcat (statfile, statbase);
  strcat (statfile, strobj_to_str (host.obj));

  if ((fd = open (statfile, O_RDONLY)) < 0)
    set_reg (LTAG_FALSE);
  else
    {
      time (&now);
      if (read (fd, &wd, sizeof (wd)) < 0)
	{
	  close (fd);
	  vbokkan (";read failed in host-up? (errno = %d).", errno);
	}
      close (fd);
      set_reg (down_p (now, wd) ? LTAG_FALSE : LTAG_T);
    }
}


void Suptime ()
{
  ENTITYP cell;
  time_t now;
  int nargs;
  register int i;
  ENTITYP mk_nconsp ();

  time (&now);

  if ((nargs = getargnum ()) == 0)
    all_uptime (now);
  else
    {
      cell = mk_nconsp (nargs);
      s_push (LTAG_CONS | SSERDDA(cell));
      for (i = 1; i <= nargs; i++)
	{
#ifdef TOUCH_ARG
	  if (GETstr_tag (touch_obj(getarg (i))) != LTAG_STRING){
#else
	  if (GETstr_tag (getarg (i)) != LTAG_STRING){
#endif
	      err_illegal_argument ("Arguments must be strings.");
	      return ;
	  }
#ifdef TOUCH_ARG
	  CARobj (cell) = uptime (now, touch_obj(getarg (i)));
#else
	  CARobj (cell) = uptime (now, getarg (i));
#endif
	  cell = (ENTITYP) ADDRESS (CDRobj (cell));
	}
      set_reg (s_pop ());
    }
}


void
all_uptime (now)
     time_t now;
{
  SOBJ host;
  DIR *dirp;
  register struct dirent *dp;
  register int npush;
  int statbaselen = strlen (statbase);
  unsigned tmp;

  if ((dirp = opendir (statdir)) == NULL)
    vbokkan (";opendir failed in all_uptime (errno = %d).", errno);

  tmp = uptime (now, get_sval (localhostname.obj));
  s_push (tmp);
  npush = 1;

  while ((dp = readdir (dirp)) != NULL)
    {
      if (strncmp (dp->d_name, statbase, statbaselen) ||
	  !strcmp (dp->d_name + statbaselen, ourhostname))
	continue;

      host.obj = mk_str (ORDINARY_STRING, A_STRING,
			 strlen (dp->d_name + statbaselen),
			 dp->d_name + statbaselen);
      tmp = uptime (now, host.obj);
      s_push (tmp);
      npush++;
    }
  closedir (dirp);

  set_reg (LTAG_NIL);
  while (npush--)
    set_reg (mk_cons (s_pop (), reg_reg));
}

#define LOADAV(wd, i)	((double) (wd).wd_loadav[i] / 100.0)

int
uptime (now, rhost)
     time_t now;
     int rhost;
{
  ENTITYP cell;
  char statfile[MAXPATHLEN];
  int fd;
  struct whod wd;
  register int i;
  ENTITYP mk_nconsp ();
  char *strobj_to_str ();
  unsigned tmp;

  strcpy (statfile, statdir);
  strcat (statfile, statbase);
  strcat (statfile, strobj_to_str (rhost));

  if ((fd = open (statfile, O_RDONLY)) < 0)
    return mk_cons (rhost, LTAG_NIL);

  if (read (fd, &wd, sizeof (wd)) < 0)
    {
      close (fd);
      vbokkan (";read failed in uptime (errno = %d).", errno);
    }
  close (fd);

  if (down_p (now, wd))
    return mk_cons (rhost, LTAG_NIL);

  tmp = LTAG_CONS | SSERDDA(cell = mk_nconsp (4));
  s_push (tmp);
  CARobj (cell) = rhost;
  cell = (ENTITYP) ADDRESS (CDRobj (cell));

  for (i = 0; i < 3; i++)
    {
      CARobj (cell) = mk_flt (LOADAV (wd, i));
      cell = (ENTITYP) ADDRESS (CDRobj (cell));
    }

  return s_pop ();
}

#endif

void Sserv_to_portnum ()
{
  SOBJ serv, proto;
  char *protoname;
  struct servent *sp;
  extern char *strobj_to_str ();

#ifdef TOUCH_ARG
  serv.obj = touch_obj(getarg (1));
  proto.obj = touch_obj(getarg (2));
#else
  serv.obj = getarg (1);
  proto.obj = getarg (2);
#endif
  if (GETstr_tag (serv.obj) != LTAG_STRING ||
      GETstr_tag (proto.obj) != LTAG_STRING){
      err_illegal_argument ("Arguments must be strings.");
      return ;
  }

  protoname = strobj_to_str (proto.obj);
  if (strcmp (protoname, "udp") &&
      strcmp (protoname, "tcp")){
      err_illegal_argument ("2nd argument must be a protocol string.");
      return ;
  }

  if ((sp = getservbyname (strobj_to_str (serv.obj),
			   protoname)) == NULL)
    set_reg (LTAG_FALSE);
  else
    set_reg (mk_fix (ntohs (sp->s_port)));
}


void Sget_host_list ()
{
  register int npush;
  struct hostent *hp;
  unsigned tmp;

  s_push (localhostname.obj);
  npush = 1;

  sethostent (1);	/* stay open */
#if defined(MACHTEN) || defined(__CYGWIN32__)
  while ((hp = NULL) != NULL)
#else
  while ((hp = gethostent ()) != NULL)
#endif
    {
      if (!strcmp (hp->h_name, ourhostname))
	continue;
      tmp = STRING (hp->h_name);
      s_push (tmp);
      npush++;
    }
  endhostent ();

  set_reg (LTAG_NIL);
  while (npush--)
    set_reg (mk_cons (s_pop (), reg_reg));
}


void Sselect ()
{
  SOBJ arg1, arg, pair;
  int tsec, nargs, margs, fd;
  struct timeval timeout;
  fd_set readfds;

  if ((nargs = margs = getargnum ()) < 2){
      err_illegal_argument ("Too few arguments.");
      return ;
  }

#ifdef TOUCH_ARG
  arg1.obj = touch_obj(getarg (1));
#else
  arg1.obj = getarg (1);
#endif
  if (GETtag (arg1.obj) == LTAG_FIXNUM)
    {
      tsec = fix_to_int (arg1.obj);
      timeout.tv_sec = (long) tsec;
      timeout.tv_usec = 0L;
    }
  else
    tsec = -1;

  FD_ZERO (&readfds);
  for (; nargs > 1; nargs--)
    {
#ifdef TOUCH_ARG
      arg.obj = touch_obj(getarg (nargs));
#else
      arg.obj = getarg (nargs);
#endif
      if (!netportp (arg.obj) && !boundportp (arg.obj)){
	  err_illegal_argument ("Arguments (except first) must be network ports.");
	  return ;
      }
      fd = (int) ((PORT *) ADDRESS (arg.obj))->fp;
      FD_SET (fd, &readfds);
    }

  if (select (getdtablesize (), &readfds, NULL,
	      NULL, tsec != -1 ? &timeout : NULL) < 0)
    vbokkan (";select failed (errno = %d).", errno);

  for (pair.obj = LTAG_NIL; margs > 1; margs--)
    {
#ifdef TOUCH_ARG
      fd = (int) ((PORT *) ADDRESS (touch_obj(getarg (margs))))->fp;
#else
      fd = (int) ((PORT *) ADDRESS (getarg (margs)))->fp;
#endif
      if (FD_ISSET (fd, &readfds))
#ifdef TOUCH_ARG
	set_reg (pair.obj = mk_cons (touch_obj(getarg (margs)), pair.obj));
#else
	set_reg (pair.obj = mk_cons (getarg (margs), pair.obj));
#endif
    }

  if (pair.obj == LTAG_NIL)
    set_reg (LTAG_FALSE);
}


#define timeout_break(val)	\
{				\
  error_prologue ();		\
  set_reg (val);		\
}


void Sread_fix ()
{
  SOBJ port;
  struct sockaddr_in *addrp;
  int sd, addrlen, x;

  switch (getargnum ())
    {
    case 0:
      if (!netportp (c_in_port.obj))
	bokkan ("Current input port isn't a network port.");
      port.obj = c_in_port.obj;
      break;
    case 1:
#ifdef TOUCH_ARG
      port.obj = touch_obj(getarg (1));
#else
      port.obj = getarg (1);
#endif
      if (!netportp (port.obj)){
	  err_illegal_argument ("Argument must be a network port.");
	  return ;
      }
      break;
    default:
      err_illegal_argument ("Too many arguments.");
      return ;
    }

#if defined(__CYGWIN32__) || defined(linux)
  if (sigsetjmp (timeout_env, 1) == 0)
#else
  if (setjmp (timeout_env) == 0)
#endif
    {
      if (_tcpportp (port.obj))
	alarm (timeoutsec (tcp_timeout.obj));
      else if (_udpportp (port.obj))
	alarm (timeoutsec (udp_timeout.obj));

      sd = (int) ((PORT *) ADDRESS (port.obj))->fp;

      addrp = getnetaddr (port.obj);
      addrlen = sizeof (*addrp);
      if (recvfrom (sd, (void *)&x, sizeof (x), 0,
		    (struct sockaddr *)addrp, &addrlen) < 0)
	set_reg (mk_fix (x));
      alarm (0);
    }
  else
    timeout_break (eof_obj.obj);
}


void Swrite_fix ()
{
  SOBJ value, port;
  struct sockaddr_in *addrp;
  int sd, x;

  switch (getargnum ())
    {
    case 0:
	err_illegal_argument ("Too few arguments.");
	return ;
    case 1:
      if (!netportp (c_out_port.obj))
	bokkan ("Current output port isn't a network port.");
      port.obj = c_out_port.obj;
      break;
    case 2:
#ifdef TOUCH_ARG
      port.obj = touch_obj(getarg (2));
#else
      port.obj = getarg (2);
#endif
      if (!netportp (port.obj)){
	  err_illegal_argument ("2nd argument must be a network port.");
	  return ;
      }
      break;
    default:
	err_illegal_argument ("Too many arguments.");
	return ;
    }

#ifdef TOUCH_ARG
  value.obj = touch_obj(getarg (1));
#else
  value.obj = getarg (1);
#endif
  if (GETfix_tag (value.obj) != LTAG_FIXNUM){
      err_illegal_argument ("1st arugment must be a fixnum.");
      return ;
  }
  x = fix_to_int (value.obj);

  sd = (int) ((PORT *) ADDRESS (port.obj))->fp;
  addrp = getnetaddr (port.obj);
  if (sendto (sd, (void *)&x, sizeof (x), 0,
	      (struct sockaddr *)addrp, sizeof (*addrp)) < 0)
    vbokkan (";sendto failed in write-fixnum (errno = %d).", errno);
  set_reg (value.obj);
}


void Sread ()
{
  int nargs;

  switch (nargs = getargnum ())
    {
    case 0:
      break;
    case 1:
      s_push (c_in_port.obj);
#ifdef TOUCH_ARG
      c_in_port.obj = touch_obj(getarg (1));
#else
      c_in_port.obj = getarg (1);
#endif
      if (!inputportp (c_in_port.obj) && !netportp (c_in_port.obj))
	{
	  c_in_port.obj = s_pop ();
	  err_illegal_argument ("1st argument must be an input port or a network port.");
	  return ;
	}
      break;
    default:
      err_illegal_argument ("Too many arguments.");
      return ;
    }

#if defined(__CYGWIN32__) || defined(linux)
  if (sigsetjmp (timeout_env, 1) == 0)
#else
  if (setjmp (timeout_env) == 0)
#endif
    {
      if (_tcpportp (c_in_port.obj))
	alarm (timeoutsec (tcp_timeout.obj));
      else if (_udpportp (c_in_port.obj))
	alarm (timeoutsec (udp_timeout.obj));
      set_reg (readexp ());
      alarm (0);
      if (nargs)
	c_in_port.obj = s_pop ();
    }
  else
    timeout_break (eof_obj.obj);
}

void readch(int);
void readbyte(int);

void Sreadch ()
{
  readch (0);
}

void Sreadbyte()
{
    readbyte(0);
}

void Speekch ()
{
  readch (1);
}


void readch (peek_flag)
     int peek_flag;	/* non-zero if peek */
{
  int ch1, ch2;
  int nargs;
  int isnet;	/* non-zero if argument is a network port */

  switch (nargs = getargnum ())
    {
    case 0:
      break;
    case 1:
      s_push (c_in_port.obj);
#ifdef TOUCH_ARG
      c_in_port.obj = touch_obj(getarg (1));
#else
      c_in_port.obj = getarg (1);
#endif
      if (!inputportp (c_in_port.obj) && !netportp (c_in_port.obj))
	{
	  c_in_port.obj = s_pop ();
	  err_illegal_argument ("1st argument must be an input port or a network port.");
	  return ;
	}
      break;
    default:
      err_illegal_argument ("Too many arguments.");
      return ;
    }

  isnet = _netportp (c_in_port.obj) ? 1 : 0;

#if defined(__CYGWIN32__) || defined(linux)
  if (sigsetjmp (timeout_env, 1) == 0)
#else
  if (setjmp (timeout_env) == 0)
#endif
    {
      if (_tcpportp (c_in_port.obj))
	alarm (timeoutsec (tcp_timeout.obj));
      else if (_udpportp (c_in_port.obj))
	alarm (timeoutsec (udp_timeout.obj));

    retry:
      if ((ch1 = getc2 ()) == EOF)
	{
/*
	  if (isnet)
	    goto retry;
	  else
*/
	    {
	      set_reg (eof_obj.obj);
	      goto ok;
	    }
	}
      else if (ch1 & 0x80)
	{
	  if ((ch2 = getc2 ()) == EOF)
	    {
	      c_in_port.obj = s_pop ();
	      bokkan ("Incomplete KANJI character (unexpected EOF).");
	    }
	  set_reg (mk_ch ((char) ch1, (char) ch2));
	}
      else
	  set_reg (mk_ch ('\0', (char) ch1));

      if (peek_flag)
	{
	    if (ch1 & 0x80)
		ungetc2 (ch2);
	    ungetc2 (ch1);
	}
    ok:
      alarm (0);

      if (nargs)
	c_in_port.obj = s_pop ();
    }
  else
    timeout_break (eof_obj.obj);
}

void readbyte(int peek_flag) /* non-zero if peek */
{
  int ch;
  int nargs;
  int isnet;	/* non-zero if argument is a network port */

  switch (nargs = getargnum ())
    {
    case 0:
      break;
    case 1:
      s_push (c_in_port.obj);
#ifdef TOUCH_ARG
      c_in_port.obj = touch_obj(getarg (1));
#else
      c_in_port.obj = getarg (1);
#endif
      if (!inputportp (c_in_port.obj) && !netportp (c_in_port.obj))
	{
	  c_in_port.obj = s_pop ();
	  err_illegal_argument ("1st argument must be an input port or a network port.");
	  return ;
	}
      break;
    default:
      err_illegal_argument ("Too many arguments.");
      return ;
    }

  isnet = _netportp (c_in_port.obj) ? 1 : 0;

#if defined(__CYGWIN32__) || defined(linux)
  if (sigsetjmp (timeout_env, 1) == 0)
#else
  if (setjmp (timeout_env) == 0)
#endif
    {
      if (_tcpportp (c_in_port.obj))
	alarm (timeoutsec (tcp_timeout.obj));
      else if (_udpportp (c_in_port.obj))
	alarm (timeoutsec (udp_timeout.obj));

    retry:
      if ((ch = getc2 ()) == EOF)
	{
/*
	  if (isnet)
	    goto retry;
	  else
*/
	    {
	      set_reg (eof_obj.obj);
	      goto ok;
	    }
	}
      set_reg(mk_fix(ch));

      if (peek_flag)
	{
	    ungetc2 (ch);
	}
    ok:
      alarm (0);

      if (nargs)
	c_in_port.obj = s_pop ();
    }
  else
    timeout_break (eof_obj.obj);
}

#endif /* TCPIP */
