/*
 * kanji.c - 
 */

/*
 * $Id: kanji.c,v 1.5 2014/10/28 02:22:55 komiya Exp $
 */

static char *RCSid="$Id: kanji.c,v 1.5 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"

#ifdef TCPIP
#include "inet.h"
extern struct sockaddr_in *getnetaddr ();
#endif

#ifdef NIHONGO
#define		KANJI(c)	((c) & 0x80)

#define		KANJI2(c)	((c) & 0x00008080)
#define		KANJI1(c)	(!((c) & 0x00008000) && ((c) & 0x000000080))

void Skanjip()
{
	int	tmp;

#ifdef TOUCH_ARG
	if ((tmp = GETtag(touch_obj(getarg(1)))) == LTAG_KCHAR)
#else
	if ((tmp = GETtag(getarg(1))) == LTAG_KCHAR)
#endif
		set_reg(LTAG_T);
	else
		set_reg(LTAG_FALSE);
}

void Skanji_input_code()
{
	PORT	*port;

	switch (getargnum()) {
		case	0:
			port = (PORT *)ADDRESS(c_in_port.obj);
			break;
		case	1:
#ifdef TOUCH_ARG
			if (GETtag(touch_obj(getarg(1))) != LTAG_PORT) {
#else
			if (GETtag(getarg(1)) != LTAG_PORT){
#endif
				err_illegal_argument("1st argument must be a port.");
				return ;
			    }
#ifdef TOUCH_ARG
			port = (PORT *)ADDRESS(touch_obj(getarg(1)));
#else
			port = (PORT *)ADDRESS(getarg(1));
#endif
			if (port->kind == STRING_INPUT){
				err_illegal_argument("1st argument may not be a string port.");
				return ;
			    }
			break;
		default:
			err_illegal_argument("Too many arguments.");
			return ;
	}

	switch (port->kanji_input_code) {
		case	EUC_CODE:
			set_reg(euc_sym.obj);
			break;
		case	JIS_CODE:
			set_reg(jis_sym.obj);
			break;
		case	SJIS_CODE:
			set_reg(sjis_sym.obj);
			break;
		case	NOC_CODE:
			set_reg(noc_sym.obj);
			break;
		default:
			bokkan("Undefined symbol");
	}
}

void Skanji_output_code()
{
	PORT	*port;

	switch (getargnum()) {
		case	0:
			port = (PORT *)ADDRESS(c_out_port.obj);
			break;
		case	1:
#ifdef TOUCH_ARG
			if (GETtag(touch_obj(getarg(1))) != LTAG_PORT) {
#else
			if (GETtag(getarg(1)) != LTAG_PORT){
#endif
				err_illegal_argument("1st argument must be a port.");
				return ;
			    }
#ifdef TOUCH_ARG
			port = (PORT *)ADDRESS(touch_obj(getarg(1)));
#else
			port = (PORT *)ADDRESS(getarg(1));
#endif
			if (port->kind == STRING_OUTPUT){
				err_illegal_argument("1st argument may not be a string port.");
				return ;
			    }
			break;
		default:
			err_illegal_argument("Too many arguments.");
			return ;
	}

	switch (port->kanji_output_code) {
		case	EUC_CODE:
			set_reg(euc_sym.obj);
			break;
		case	JIS_CODE:
			set_reg(jis_sym.obj);
			break;
		case	SJIS_CODE:
			set_reg(sjis_sym.obj);
			break;
		case	NOC_CODE:
			set_reg(noc_sym.obj);
			break;
		default:
			bokkan("Undefined symbol");
	}
}

void Schange_kanji_input_code()
{
	PORT	*port;
	SOBJ	code;
	int	argnum = getargnum();

	switch	(argnum) {
		case	0:
			err_illegal_argument("Too few arguments.");
			return ;
		case	1:
			port = (PORT *)ADDRESS(c_in_port.obj);
			break;
		case	2:
#ifdef TOUCH_ARG
			if (GETtag(touch_obj(getarg(1))) != LTAG_PORT) {
#else
			if (GETtag(getarg(1)) != LTAG_PORT){
#endif
				err_illegal_argument("1st argument must be a port.");
				return ;
			    }
#ifdef TOUCH_ARG
			port = (PORT *)ADDRESS(touch_obj(getarg(1)));
#else
			port = (PORT *)ADDRESS(getarg(1));
#endif
			if (port->kind == STRING_INPUT){
				err_illegal_argument("1st argument may not be a string port.");
				return ;
			    }
			break;
		default:
			err_illegal_argument("Too many arguments.");
			return ;
	}

	code.obj = getarg(argnum);
	if (GETsym_tag(code.obj) != LTAG_SYMBOL)
		if (argnum == 1){
			err_illegal_argument("1st argument must be a symbol.");
			return ;
		    }
		else{
			err_illegal_argument("2nd argument must be a symbol.");
			return ;
		    }

	if (code.obj == euc_sym.obj || code.obj == jis_sym.obj ||
	    code.obj == sjis_sym.obj || code.obj == noc_sym.obj)
		port->kanji_input_code = fix_to_int(get_sval(code.obj));
	else{
		bokkan("Undefined code symbol");
	    }

	set_reg(getarg(argnum));
}

void Schange_kanji_output_code()
{
	PORT	*port;
	SOBJ	code;
	int	argnum = getargnum();

	switch (argnum) {
		case	0:
			err_illegal_argument("Too few arguments.");
			return ;
		case	1:
			port = (PORT *)ADDRESS(c_out_port.obj);
			break;
		case	2:
#ifdef TOUCH_ARG
			if (GETtag(touch_obj(getarg(1))) != LTAG_PORT) {
#else
			if (GETtag(getarg(1)) != LTAG_PORT){
#endif
				err_illegal_argument("1st argument must be a port.");
				return ;
			    }
#ifdef TOUCH_ARG
			port = (PORT *)ADDRESS(touch_obj(getarg(1)));
#else
			port = (PORT *)ADDRESS(getarg(1));
#endif
			if (port->kind == STRING_OUTPUT){
				err_illegal_argument("1st argument may not be a string port.");
				return ;
			    }
			break;
		default:
			err_illegal_argument("Too many arguments.");
				return ;
	}

	code.obj = getarg(argnum);
	if (GETsym_tag(code.obj) != LTAG_SYMBOL)
		if (argnum == 1){
			err_illegal_argument("1st argument must be a symbol.");
			return ;
		    }
		else{
			err_illegal_argument("2nd argument must be a symbol.");
			return ;
		    }

	if (code.obj == euc_sym.obj || code.obj == jis_sym.obj ||
	    code.obj == sjis_sym.obj || code.obj == noc_sym.obj)
		port->kanji_output_code = fix_to_int(get_sval(code.obj));
	else{
		bokkan("Undefined code symbol");
	    }

	set_reg(getarg(argnum));
}

void	jis2euc(c1, c2, c11, c22)
int	c1, c2, *c11, *c22;
{
	*c11 = c1 + 0x80;
	*c22 = c2 + 0x80;
}

void	euc2jis(c1, c2, c11, c22)
unsigned char	c1, c2, *c11, *c22;
{
	*c11 = c1 - 0x80;
	*c22 = c2 - 0x80;
}

void	sjis2euc(c1, c2, c11, c22)
int	c1, c2, *c11, *c22;
{
	if (c2 >= 0x9f) {
		c1 = c1 * 2 - ((c1 >= 0xe0) ? 0xe0: 0x60);
		c2 = c2 + 2;
	} else {
		c1 = c1 * 2 - ((c1 >= 0xe0) ? 0xe1: 0x61);
		c2 = c2 + ((c2 >= 0x7f) ? 0x60: 0x61);
	}

	*c11 = c1;
	*c22 = c2;
}

void	euc2sjis(c1, c2, c11, c22)
unsigned char	c1, c2, *c11, *c22;
{
	if (c1 & 1) {
		c1 = c1 / 2 + ((c1 < 0xdf) ? 0x31: 0x71);
		c2 = c2 - ((c2 >= 0xe0) ? 0x60: 0x61);
	} else {
		c1 = c1 / 2 + ((c1 < 0xdf) ? 0x30: 0x70);
		c2 = c2 - 2;
	}

	*c11 = c1;
	*c22 = c2;
}

static	int	_gsp = 0, _gstack[5];

static	int	_kgetc1(port)
PORT	*port;
{
#ifdef TCPIP
	char c;
	struct sockaddr_in *addrp;
	int addrlen = 0;
#endif

	if (_gsp > 0)
		return (_gstack[--_gsp]);
	else {
		switch (port->kind) {
#ifdef TCPIP			
		case UDP_DGRAM:
			addrp = getnetaddr (c_in_port.obj);
			addrlen = sizeof (struct sockaddr_in);
		case TCP_STREAM:
			switch (recvfrom ((int) port->fp, &c, 1, 0, (struct sockaddr *)addrp, &addrlen)) {
			case -1:
			case 0:
				return (EOF);
			default:
				return ((int) c);
			}
#endif
		default:
#ifdef TK
			return (Getc(port->fp));
#else
			return (getc(port->fp));
#endif
		}
	}
}

static	void	_unkgetc1(c)
int	c;
{
	_gstack[_gsp++] = c;
}

int	kgetc_euc(port)
PORT	*port;
{
	return (_kgetc1(port));
}

int	kgetc_jis(port)
PORT	*port;
{
	int	c, c1, c2;
	static	int	second = OFF;

 	if ((c = _kgetc1(port)) == EOF)
		return (EOF);

	if (second) {
		second = OFF;
		return (c);
	}

	if (c == ESC_CODE) {
		c1 = _kgetc1(port);
		c2 = _kgetc1(port);
		if (c1 == SHIFT_IN1 && (c2 == SHIFT_IN2_NEW || c2 == SHIFT_IN2_OLD)) {
			port->shift = ON;
			c = _kgetc1(port);
		} else if (c1 == SHIFT_OUT1
                           && (c2 == SHIFT_OUT2 || c2 == SHIFT_OUT2_ASC)) {
			port->shift = OFF;
			c = _kgetc1(port);
		} else {
			_unkgetc1(c2);
			_unkgetc1(c1);
		}
	}

	if (port->shift) {
		jis2euc(c, _kgetc1(port), &c1, &c2);
		_unkgetc1(c2);
		second = ON;
		return (c1);
	} else
		return (c);
}

int	kgetc_sjis(port)
PORT	*port;
{
	int	c, c1, c2;
	static	int	second = OFF;

	if ((c = _kgetc1(port)) == EOF)
		return (EOF);

	if (second) {
		second = OFF;
		return (c);
	}

	if (c & 0x80) {
		sjis2euc(c, _kgetc1(port), &c1, &c2);
		_unkgetc1(c2);
		second = ON;
		return (c1);
	} else
		return (c);
}

int	kgetc_noc(port)
PORT	*port;
{
	return (_kgetc1(port));
}

int	kgetc(port)
PORT	*port;
{
	static	int	(*_kgetc[])() = {kgetc_euc, kgetc_jis, kgetc_sjis, kgetc_noc};

	return ((*_kgetc[port->kanji_input_code])(port));
}

static	void	_kputc1(c, port)
char	c;
PORT	*port;
{
	switch (port->kind) {
#ifdef TCPIP			
	case UDP_DGRAM:
	case TCP_STREAM:
		if (sendto ((int) port->fp, &c, 1, 0,
			    (struct sockaddr *)getnetaddr (c_out_port.obj),
			    sizeof (struct sockaddr_in)) < 0)
			bokkan (";sendto failed (at kputc)");
		break;
#endif
	default:
		putc(c, port->fp);
	}
}

void	kputc_euc(c, port)
char	c;
PORT	*port;
{
	if (c == EOF)
		return;
	else
		_kputc1(c, port);
}

void	kputc_jis(c, port)
char	c;
PORT	*port;
{
	static	char	hc = (char)NULL;
	char	c1, c2;

	if (c == EOF) {
		if (port->shift) {
			_kputc1(ESC_CODE, port);
			_kputc1(SHIFT_OUT1, port);
			_kputc1(SHIFT_OUT2, port);
			port->shift = OFF;
		}
		return;
 	} else if (KANJI(c)) {
		if (!hc) {
			if (!port->shift) {
				_kputc1(ESC_CODE, port);
				_kputc1(SHIFT_IN1, port);
				_kputc1(SHIFT_IN2, port);
				port->shift = ON;
			}
			hc = c;
		} else {
			euc2jis(hc, c, (unsigned char *)&c1, (unsigned char *)&c2);
			_kputc1(c1, port);
			_kputc1(c2, port);
			hc = (char)NULL;
		}
	} else {
		if (port->shift) {
			_kputc1(ESC_CODE, port);
			_kputc1(SHIFT_OUT1, port);
			_kputc1(SHIFT_OUT2, port);
			port->shift = OFF;
		}
		_kputc1(c, port);
	}
}			

void	kputc_sjis(c, port)
char	c;
PORT	*port;
{
	static	char	hc = (char)NULL;
	char	c1, c2;

	if (c == EOF)
		return;
 	else if (KANJI(c)) {
		if (!hc) {
			hc = c;
		} else {
			euc2sjis(hc, c, (unsigned char *)&c1, (unsigned char *)&c2);
			_kputc1(c1, port);
			_kputc1(c2, port);
			hc = (char)NULL;
		}
	} else {
		_kputc1(c, port);
	}
}			

void	kputc_noc(c, port)
char	c;
PORT	*port;
{
	_kputc1(c, port);
}

void	kputc(c, port)
char	c;
PORT	*port;
{
	static	void	(*_kputc[])() = {kputc_euc, kputc_jis, kputc_sjis, kputc_noc};

	(*_kputc[port->kanji_output_code])(c, port);
}
#endif /* NIHONGO */
