/*
 * logical.c - 
 */

/*
 * $Id: logical.c,v 1.4 2014/10/28 02:22:55 komiya Exp $
 */

/************************************/
/*                                  */
/*  File Name : lpgical.c           */
/*  Update    : Yass.  Jul. 19 '90  */
/*              Yass.  Jul. 18 '90  */
/*              Myama. ?            */
/*                                  */
/*    Logical operations            */
/*                                  */
/************************************/

static char *RCSid="$Id: logical.c,v 1.4 2014/10/28 02:22:55 komiya Exp $";

#include <math.h>
#include "scheme.h"

/*  bitwise logical operations on bignum */
#define FUJIWARA
#define FUJIWARA2

/*************************/
/*                       */
/*  Original procedures  */
/*                       */
/*************************/

#ifdef FUJIWARA
int
and_op(int i, int j);
#endif

logand2(x, y)
int x, y;
{
    int val1, val2;

#ifdef FUJIWARA
    int big_flag = 0;
#endif

    switch (GETtag(x)) {
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val1 = fix_to_int(x);
        break;
    case LTAG_BIGNUM  :
#ifdef FUJIWARA
        big_flag = 1;
	break;
#else
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
#endif
    default :
        err_illegal_argument("Both arguments must be integers.");
	return 0;               /* never reach here */
        break;
    }
    switch (GETtag(y)) {
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val2 = fix_to_int(y);
        break;
    case LTAG_BIGNUM  :
#ifdef FUJIWARA
        big_flag = 1;
	break;
#else
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
#endif
    default :
        err_illegal_argument("Both arguments must be integers.");
	return 0;               /* never reach here */
        break;
    }

#ifdef FUJIWARA
    if (big_flag == 0){
#endif

    val1 &= val2;
    return((int)mk_fix(val1));

#ifdef FUJIWARA
    }
    else {
      if (GETfix_tag(x) == LTAG_FIXNUM){
	s_push(y);
	x = mk_bign(fix_to_int(x));       
	y = s_pop();  
      }
      return(log2_big(x,y,and_op));
    }
#endif

}

#ifdef FUJIWARA
/*
  References:
        /home/lab4/intern1/fujiwara/tus/c/bignum.c
        /home/lab4/komiya/kcl-rb/kcl-nise-return/c/num_log.c
	20010613
*/
/*
	x must be BIGNUM.
	y must be either BIGNUM or FIXNUM.
	log2_big() makes a new BIGNUM object for the result
	and converts it into FIXNUM if possible.
*/
int log2_big(int x, int y, int (*op)())
{
    SOBJ new_obj;

    int	sign_x, sign_y;
    int	ext_x, ext_y;
    int	end_x, end_y;
    int	i, j;

    int first_flg;

    /* MEMO fix -> fix_to_int */
    /*      big_sign -> big_sign */
    /*      MASK -> INT_MASK */

    /* ext_x and ext_y are numbers to fill extended BIGNUM. */
    if (big_sign(x) < 0) {
        sign_x = ~INT_MASK;
	ext_x = INT_MASK;
    } else
        sign_x = ext_x = 0;

    if (GETfix_tag(y) == LTAG_FIXNUM){

      if (fix_to_int(y) < 0) {
	  sign_y = ~INT_MASK;
	  ext_y = INT_MASK;
      } else
	  sign_y = ext_y = 0;
    }
    /* case of BIGNUM */
    else {
        if (big_sign(y) < 0) {
	    sign_y = ~INT_MASK;
	    ext_y = INT_MASK;
	} else
	    sign_y = ext_y = 0;
    }
    
    end_x = end_y = 0;
    first_flg = 1;
    while ((end_x == 0) || (end_y == 0)) {
        if (end_x == 0)
	    i = (((BIGNUM * )ADDRESS(x))->val) & INT_MASK;
	else
	    i = ext_x;
	if (end_y == 0)
	    if (GETfix_tag(y) == LTAG_FIXNUM)
	        j = (fix_to_int(y)) & INT_MASK;
	    else
	        j = (((BIGNUM * )ADDRESS(y))->val) & INT_MASK;
	else
	    j = ext_y;
      
	i = (*op)(i, j);

	/* substitution of the result */
	if (first_flg == 1){
	    s_push(x);
	    s_push(y);
	    new_obj.obj = mk_bign(i);
	    y = s_pop();  
	    x = s_pop();
	    
	    /* will pop new_obj.obj before return */
	    s_push(new_obj.obj);

	    first_flg = 0;
      }
      else {
          s_push(x);
	  s_push(y);
      
	  new_obj.obj = stretch_big(new_obj.obj, i);
      
	  y = s_pop();  
	  x = s_pop();
      }
      if (((BIGNUM * )ADDRESS(x))->next.obj != LTAG_NIL)
          x = ((BIGNUM * )ADDRESS(x))->next.obj;
      else
	  end_x = 1;
      if (GETfix_tag(y) == LTAG_FIXNUM)
	  end_y = 1;
      else if (((BIGNUM * )ADDRESS(y))->next.obj != LTAG_NIL)
	  y = ((BIGNUM * )ADDRESS(y))->next.obj;
      else
	  end_y = 1;
    }

    /* Now x points ths last sell of bignum.
       We must set the sign bit according to operation.
       Sign bit of x is already masked out in previous
       while-iteration */
    ((BIGNUM * )ADDRESS(new_obj.obj))->val |=
      ((*op)(sign_x, sign_y) & ~INT_MASK);
    
    new_obj.obj = s_pop();
    
    normalize_big(new_obj.obj);
    
    /* convert into FIXNUM if possible */
    if (((BIGNUM * )ADDRESS(new_obj.obj))->next.obj == LTAG_NIL &&
	((BIGNUM * )ADDRESS(new_obj.obj))->val <= FIX_MAX &&
	((BIGNUM * )ADDRESS(new_obj.obj))->val >= FIX_MIN){
        new_obj.obj = mk_fix(((BIGNUM * )ADDRESS(new_obj.obj))->val);
    }

    return((int)new_obj.obj);
}

/*
        20010629
	x must be BIGNUM.
*/
int lognot_big(int x)
{
    SOBJ new_obj;

    int	end_x;
    int	i;

    int first_flg;

    end_x = 0;
    first_flg = 1;
    while (end_x == 0) {
        i = (((BIGNUM * )ADDRESS(x))->val);

	i = ~i;

	/* i must not be negative until x reaches the end. */
	if (((BIGNUM * )ADDRESS(x))->next.obj != LTAG_NIL){
	    x = ((BIGNUM * )ADDRESS(x))->next.obj;
	    i = i & INT_MASK;
	}
	else
	    end_x = 1;

	/* substitution of the result */
	if (first_flg == 1){
	    s_push(x);
	    new_obj.obj = mk_bign(i);
	    x = s_pop();
	    
	    /* will pop new_obj.obj before return */
	    s_push(new_obj.obj);
	    
	    first_flg = 0;
	}
	else {
	    s_push(x);
	    
	    new_obj.obj = stretch_big(new_obj.obj, i);
      
	    x = s_pop();
	}
    }

    new_obj.obj = s_pop();

    normalize_big(new_obj.obj);
    
    /* convert into FIXNUM if possible */
    if (((BIGNUM * )ADDRESS(new_obj.obj))->next.obj == LTAG_NIL &&
	((BIGNUM * )ADDRESS(new_obj.obj))->val <= FIX_MAX &&
	((BIGNUM * )ADDRESS(new_obj.obj))->val >= FIX_MIN){
        new_obj.obj = mk_fix(((BIGNUM * )ADDRESS(new_obj.obj))->val);
    }

    return((int)new_obj.obj);
}

int
ior_op(i, j)
int	i, j;
{
    return(i | j);
}

int
xor_op(i, j)
int	i, j;
{
    return(i ^ j);
}

int
and_op(i, j)
int	i, j;
{
    return(i & j);
}

#endif

Slogand()  /*  Yass. Jul. 19 '90  */
{ 
    int argnum, i, val1, val2;

    argnum = getargnum();
    val1 = mk_fix(-1);
    for (i = 1; i <= argnum; i++) {
#ifdef TOUCH_ARG
	val2 = touch_obj(getarg(i));
#else
	val2 = getarg(i);
#endif
	val1 = logand2(val1, val2);
    }

    set_reg(val1);
}

logior2(x, y)
int x, y;
{
    int val1, val2;

#ifdef FUJIWARA
    int big_flag = 0;
#endif

    switch (GETtag(x)) {
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val1 = fix_to_int(x);
        break;
    case LTAG_BIGNUM  :
#ifdef FUJIWARA
        big_flag = 1;
	break;
#else
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
#endif
    default :
        err_illegal_argument("Both arguments must be integers.");
	return 0;               /* never reach here */
        break;
    }
    switch (GETtag(y)) {
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val2 = fix_to_int(y);
        break;
    case LTAG_BIGNUM  :
#ifdef FUJIWARA
        big_flag = 1;
	break;
#else
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
#endif
    default :
        err_illegal_argument("Both arguments must be integers.");
	return 0;                /* never reach here */
        break;
    }

#ifdef FUJIWARA
    if (big_flag == 0){
#endif

    val1 |= val2;
    return((int)mk_fix(val1));

#ifdef FUJIWARA
    }
    else {
        if (GETfix_tag(x) == LTAG_FIXNUM){
	    s_push(y);
	    x = mk_bign(fix_to_int(x));       
	    y = s_pop();  
	}
	return(log2_big(x,y,ior_op));
    }
#endif

}

void Slogior()  /*  Yass. Jul. 19 '90  */
{ 
    int argnum, i, val1, val2;

    argnum = getargnum();
    val1 = SCOBJ_0;
    for (i = 1; i <= argnum; i++) {
#ifdef TOUCH_ARG
	val2 = touch_obj(getarg(i));
#else
	val2 = getarg(i);
#endif
	val1 = logior2(val1, val2);
    }
    set_reg(val1);
}

void Slognot()  /*  Yass. Jul. 19 '90  */
{ 
    SOBJ obj;
    int val;

#ifdef FUJIWARA
    int big_flag = 0;
    int tmp;
#endif

#ifdef TOUCH_ARG
    obj.obj = touch_obj(getarg(1));
#else
    obj.obj = getarg(1);
#endif
    switch (GETtag(obj.obj)) { 
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val = fix_to_int(obj.obj);
        break;
    case LTAG_BIGNUM  :
#ifdef FUJIWARA
        big_flag = 1;
	break;
#else
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
#endif
    default :
        err_illegal_argument("Argument must be an integer.");
	return ;
        break;
    }
#ifdef FUJIWARA
    if (big_flag == 0){
#endif

    val = ~val;
    set_reg(mk_fix(val));
#ifdef FUJIWARA
    }
    else {
        set_reg(lognot_big(obj.obj));
    }
#endif
}

#ifndef FUJIWARA2

void Slogshl()  /*  Yass. Jul. 19 '90  */
{ 
    SOBJ obj;
    unsigned val1;
    int val2;

#ifdef TOUCH_ARG
    obj.obj = touch_obj(getarg(1));
#else
    obj.obj = getarg(1);
#endif
    switch (GETtag(obj.obj)) { 
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val1 = obj.obj & 0xffffffc;
        break;
    case LTAG_BIGNUM  :
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
    default :
        err_illegal_argument("Both arguments must be integers.");
	return ;
        break;
    }
#ifdef TOUCH_ARG
    obj.obj = touch_obj(getarg(2));
#else
    obj.obj = getarg(2);
#endif
    switch (GETtag(obj.obj)) { 
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val2 = fix_to_int(obj.obj);
        break;
    case LTAG_BIGNUM  :
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
    default :
        err_illegal_argument("Both arguments must be integers.");
	return ;
        break;
    }
    if (val2 < 0)  {
        err_illegal_argument("2nd argument must be a non-negative number.");
	return ;
    }

    if (val2 >= 26)
	set_reg(SCOBJ_0);
    else {
	val1 <<= val2;
	set_reg(LTAG_NUMBER1 | (val1 & 0xffffffc));
    }
}


void Slogshr()  /*  Yass. Jul. 19 '90  */
{ 
    SOBJ obj;
    unsigned val1;
    int val2;

#ifdef TOUCH_ARG
    obj.obj = touch_obj(getarg(1));
#else
    obj.obj = getarg(1);
#endif
    switch (GETtag(obj.obj)) { 
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val1 = obj.obj & 0xffffffc;
        break;
    case LTAG_BIGNUM  :
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
    default :
        err_illegal_argument("Both arguments must be integers.");
	return ;
        break;
    }
#ifdef TOUCH_ARG
    obj.obj = touch_obj(getarg(2));
#else
    obj.obj = getarg(2);
#endif
    switch (GETtag(obj.obj)) { 
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val2 = fix_to_int(obj.obj);
        break;
    case LTAG_BIGNUM  :
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
    default :
        err_illegal_argument("Both arguments must be integers.");
	return ;
        break;
    }
    if (val2 < 0)  {
        err_illegal_argument("2nd argument must be a non-negative number.");
	return ;
    }

    if (val2 >= 26)
	set_reg(SCOBJ_0);
    else {
	val1 >>= val2;
	set_reg(LTAG_NUMBER1 | (val1 & 0xffffffc));
    }
}

#endif

#ifdef FUJIWARA2

/*
	double_shift(h, l, w, hp, lp) shifts the int h & l ( 31 bits)
	w bits to left ( w > 0) or to right ( w < 0).
	result is returned in *hp and *lp.
*/
void double_shift(int h, int l, int w, int *hp, int *lp)
{
    if (w >= 0) {
        *lp = (l << w) & INT_MASK;
	*hp = ((h << w) & INT_MASK) | ((l & INT_MASK) >> (31 - w));
    } else {
        w = -w;
	*hp = (h & INT_MASK) >> w;
	*lp = ((h << (31 - w)) & INT_MASK) | ((l & INT_MASK) >> w);
    }
}

#define FIXWIDTH 26

int
shift_integer(int x,int w)
{
    int y;
    int	cell, bits, sign, i;
    int	ext, h, l, nh, nl, end_x;
    
    if (GETfix_tag(x) == LTAG_FIXNUM){
        i = fix_to_int(x);
	if (abs(w) < FIXWIDTH-1) {
	    if (w < 0) {
	        if (i >= 0)
		    return(mk_fix(i >> -w));
		else
		    return(mk_fix(~((~i) >> -w)));
	    } if (i >= 0) {
	        if (((-1<<(FIXWIDTH-1-w)) & i) == 0)
		    /* if (((~INT_MASK >> w) & i) == 0) */
		    return(mk_fix(i << w));
	    } else {
	        if (((-1<<(FIXWIDTH-1-w)) & ~i) == 0)
		    /* if (w < 32 && ((~INT_MASK >> w) & ~i) == 0) */
		    return(mk_fix(i << w));
	    }
	}
	/* case of left shift over 25 bit */
	x = mk_bign(i);
    }
    
    if ((sign = big_sign(x)) < 0)
        ext = INT_MASK;
    else
        ext = 0;
    
    s_push(x);
    y = mk_bign(0);
    x = s_pop();
    
    /* will pop y before return */
    s_push(y);
    
    cell = w / 31;
    bits = w % 31;

    if (w >= 0){
      /*  LEFT: */
        s_push(x);
        while (cell-- > 0)
	  y = stretch_big(y, 0);
	x = s_pop();

	l = 0;
	h = ((BIGNUM * )ADDRESS(x))->val;
	end_x = 0;
    }
    else {
      /* RIGHT: */
        end_x = 0;
	h = ((BIGNUM * )ADDRESS(x))->val;
	while (cell++ <= 0) {
            l = h;
	    if (end_x == 1) break;
	    if (((BIGNUM * )ADDRESS(x))->next.obj != LTAG_NIL) {
	        x = ((BIGNUM * )ADDRESS(x))->next.obj;
		h = ((BIGNUM * )ADDRESS(x))->val;
	    } else {
	        end_x = 1;
		h = ext;
	    }
	}
    }
    /* COMMON */
    for (;;) {
        double_shift(h, l, bits, &nh, &nl);
	if (w < 0)
	    ((BIGNUM * )ADDRESS(y))->val = nl;
	else
	    ((BIGNUM * )ADDRESS(y))->val = nh;
	if (end_x == 1) break;
	l = h;
	if (((BIGNUM * )ADDRESS(x))->next.obj != LTAG_NIL) {
  	    x = ((BIGNUM * )ADDRESS(x))->next.obj;
	    h = ((BIGNUM * )ADDRESS(x))->val;
	} else {
	    h = ext;
	    end_x = 1;
	}
	s_push(x);
	y = stretch_big(y, 0);
	x = s_pop();
    }
    /* set sign bit */
    if (sign < 0) ((BIGNUM * )ADDRESS(y))->val |= ~INT_MASK;
    
    y = s_pop();
    normalize_big(y);
    
    /* convert into FIXNUM if possible */
    if (((BIGNUM * )ADDRESS(y))->next.obj == LTAG_NIL &&
	((BIGNUM * )ADDRESS(y))->val <= FIX_MAX &&
	((BIGNUM * )ADDRESS(y))->val >= FIX_MIN){
        y = mk_fix(((BIGNUM * )ADDRESS(y))->val);
    }

    return(y);
}

/* if lr = 1  then left  shift
   if lr = -1 then right shift */
int shift_LR(SOBJ obj1,SOBJ obj2,int lr)
{
    int r;
    int sign_val1,val1;
    
    if (GETfix_tag(obj2.obj) == LTAG_FIXNUM){
        r = shift_integer((int)obj1.obj, lr * fix_to_int(obj2.obj));
    }
    else {

        /*
	  bit position represented by bignum is probably
	  out of our address space. So, result is returned
	  according to sign of integer.
	*/
        /* LARGE_SHIFT */
      
        if (GETfix_tag(obj1.obj) == LTAG_FIXNUM){
	    val1 = fix_to_int(obj1.obj);
	    if (val1 > 0)
	        sign_val1 = 1;
	    else if (val1 == 0)
                sign_val1 = 0;
	    else
	        sign_val1 = -1;
	}
	else /* case of BIGNUM */
	    sign_val1 = big_sign(obj1.obj);
      
	/* shift right too much */
	if (lr * big_sign(obj2.obj) < 0){
	    if (sign_val1 < 0)
	        r = mk_fix(-1);
	    else
	        r = mk_fix(0);
	}
	/* shift left too much */
	else if (sign_val1 == 0)
	    r = mk_fix(0);
	else
	    bokkan("Insufficient memory.");
    }
    return r;
}

/* NEW 20010629*/
/* Slogshl and Slogshr just check given parameters */
void Slogshl()
{
    SOBJ obj1,obj2;
    
#ifdef TOUCH_ARG
    obj1.obj = touch_obj(getarg(1));
#else
    obj1.obj = getarg(1);
#endif
    switch (GETtag(obj1.obj)) { 
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        break;
    case LTAG_BIGNUM  :
        break;
    default :
        err_illegal_argument("Both arguments must be integers.");
	return ;
        break;
    }
#ifdef TOUCH_ARG
    obj2.obj = touch_obj(getarg(2));
#else
    obj2.obj = getarg(2);
#endif
    switch (GETtag(obj2.obj)) { 
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        break;
    case LTAG_BIGNUM  :
        break;
    default :
        err_illegal_argument("Both arguments must be integers.");
	return ;
        break;
    }
    set_reg(shift_LR(obj1,obj2,1));
}

void Slogshr()
{
    SOBJ obj1,obj2;
    
#ifdef TOUCH_ARG
    obj1.obj = touch_obj(getarg(1));
#else
    obj1.obj = getarg(1);
#endif
    switch (GETtag(obj1.obj)) { 
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        break;
    case LTAG_BIGNUM  :
        break;
    default :
        err_illegal_argument("Both arguments must be integers.");
	return ;
        break;
    }
#ifdef TOUCH_ARG
    obj2.obj = touch_obj(getarg(2));
#else
    obj2.obj = getarg(2);
#endif
    switch (GETtag(obj2.obj)) { 
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        break;
    case LTAG_BIGNUM  :
        break;
    default :
        err_illegal_argument("Both arguments must be integers.");
	return ;
        break;
    }
    set_reg(shift_LR(obj1,obj2,-1));
}

#endif

logxor2(x, y)
int x, y;
{
    int val1, val2;

#ifdef FUJIWARA
    int big_flag = 0;
#endif

    switch (GETtag(x)) {
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val1 = fix_to_int(x);
        break;
    case LTAG_BIGNUM  :
#ifdef FUJIWARA
        big_flag = 1;
	break;
#else
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
#endif
    default :
        err_illegal_argument("Both arguments must be integers.");
	return 0;               /* never reach here */
        break;
    }
    switch (GETtag(y)) {
    case LTAG_FIXNUM  :
    case LTAG_FIXNUM1 :
    case LTAG_FIXNUM2 :
    case LTAG_FIXNUM3 :
        val2 = fix_to_int(y);
        break;
    case LTAG_BIGNUM  :
#ifdef FUJIWARA
        big_flag = 1;
	break;
#else
        bokkan("Sorry, bignum type is not implimented yet.");
        break;
#endif
    default :
        err_illegal_argument("Both arguments must be integers.");
	return 0;               /* never reach here */
        break;
    }

#ifdef FUJIWARA
    if (big_flag == 0){
#endif

    val1 ^= val2;
    return((int)mk_fix(val1));

#ifdef FUJIWARA
    }
    else {
        if (GETfix_tag(x) == LTAG_FIXNUM){
	    s_push(y);
	    x = mk_bign(fix_to_int(x));       
	    y = s_pop();  
	}
	return(log2_big(x,y,xor_op));
    }
#endif

}

void Slogxor()  /*  Yass. Jul. 19 '90  */
{ 
    int argnum, i, val1, val2;

    argnum = getargnum();
    val1 = SCOBJ_0;
    for (i = 1; i <= argnum; i++) {
#ifdef TOUCH_ARG
	val2 = touch_obj(getarg(i));
#else
	val2 = getarg(i);
#endif
	val1 = logxor2(val1, val2);
    }
    set_reg(val1);
}
