/*
 * ms-gc.c - 
 */

/*
 * $Id: ms-gc.c,v 1.3 2014/10/28 02:22:55 komiya Exp $
 */

/*********************************************/
/*                                           */
/*  File Name : ms-gc.c                      */
/*     written by Mikio Yamamoto, 1990.7?    */
/*                                           */
/*    Mark & Sweep method Gabage Collection  */
/*                                           */
/*********************************************/

static char *RCSid="$Id: ms-gc.c,v 1.3 2014/10/28 02:22:55 komiya Exp $";

#include "scheme.h"
#ifdef GC_TIME
#include <sys/time.h>
#include <sys/resource.h>
#endif

#define     GC_SUCCESS  1
#define     GC_FAIL     0

/* global variable */
/* 
entity_base : 
entity_fin :
entity_area_size :
max_eintity_area_size :
entity_end :
extent_rate :
*/
/*

   /|\        /|\               ----------------- entity_base
    |          |                |///////////////|
    |          |                |///////////////|
    |          |                |///////////////|
    |          |                |               |
    |          |                |               |
    |          |                |///////////////|
    |          |                |///////////////|
    |          |                |               |
    |          |                |///////////////|
    |          |                |///////////////|
    |          |                |               |
    |   entity_area_size        |               |
    |          |                |               |
    |          |                |///////////////|
    |          |                |///////////////|
    |          |                |///////////////|
    |          |                |               |
    |          |                |///////////////|
    |          |                |///////////////|
    |          |                |               |
    |         \|/               |---------------| entity_end
    |                           |               |
    |                           |               |
    |                           |               |
    |                           |               |
    |                           |               |
  max_entity_area_size          |               |
    |                           |               |
    |                           |               |
    |                           |               |
    |                           |               |
   \|/               entity_fin -----------------
*/

#if defined(THINK_C) || defined(MACHTEN)
static unsigned a2, a4, a5, d3;
#else
unsigned a2, a4, a5, d3;
#endif

extern int  message_level;
extern sharp_eq_context_max;
extern SHARP_EQ_CONTEXT_STRUCT sharp_eq_context[];

/* for debug -------------------------------------------------------- */

#if defined(THINK_C) || defined(MACHTEN)
char src_tmp_tmp[8];
unsigned *src_tmp;
extern char *dummy_dummy;
extern unsigned *dummy_source;
void mark_unit(int *);
#endif

extern int  gctraceflag;
extern FILE *debug_stream;
int sweepertraceflag;
int bcitraceflag = 0;

bci_debug_for_msgc()
{
    if (bcitraceflag > 0) {
        printf("a2 = %x, d4 = 0x%x(%d)\n", a2_val(), d4_val(), d4_val());
    }
}


a_gc_debug_go_ahead(obja2, adda2, obja3, adda3)
{
    fprintf(debug_stream, "(go ahead)(a2) = %x, a2 = %x, (a3) = %x, a3 = %x\n",
        obja2, adda2, obja3, adda3);
    if (GETtag(obja3) == LTAG_SPECIAL) {
        fprintf(debug_stream, "((a3)) = %x\n", *((unsigned *)ADDRESS(obja3)));
    }
    fflush(debug_stream);
}


Ssweeper_trace_on()
{
    sweepertraceflag = 1;
    set_reg(LTAG_T);
}


Ssweeper_trace_off()
{
    sweepertraceflag = 0;
    set_reg(LTAG_NIL);
}


Sbci_trace_on()
{
    bcitraceflag = 1;
    set_reg(LTAG_T);
}


Sbci_trace_off()
{
    bcitraceflag = 0;
    set_reg(LTAG_NIL);
}


a_gc_debug_sweeper(obja2, adda2)
{
    if (sweepertraceflag == 1) {
        printf("(go ahead)(a2) = %x, a2 = %x\n", obja2, adda2);
    }
}


/* for debug end -------------------------------------------------------- */

#ifdef CALLIOC
unsigned *marked_stack_list;
#endif

mark_and_sweep_gc()
{
    int vacancy_size;
    int a3, d4;
    unsigned *tmp;
    char    messtr[256];

#ifdef GC_TIME
    struct rusage ru;
    long sec, usec;
    extern gc_time_sec, gc_time_usec;

    getrusage(RUSAGE_SELF, &ru);
    sec = ru.ru_utime.tv_sec;
    usec = ru.ru_utime.tv_usec;
#endif
#ifdef CALLIOC
    marked_stack_list = (unsigned *)LTAG_NIL;
    ((unsigned *)stack_bottom)[2] = 1;
#endif

#if defined(THINK_C) || defined(MACHTEN)
    dummy_source = (unsigned *)(dummy_dummy + ((unsigned)dummy_dummy % 4));
    src_tmp = (unsigned *)(src_tmp_tmp + ((unsigned)src_tmp_tmp % 4));
#endif
    message_level = fix_to_int(get_sval(gc_message_level_sym.obj));

    /*    
    debug_stream = fopen("mark-trace" , "w");
*/
    if (message_level)
        print_con("\nMARK and SWEEP GC start:\n");
    if (message_level >= 2) {
        sprintf(messtr, "base = 0x%x, end = 0x%x fin = 0x%x,\nc_size = 0x%x(%d), max_size = 0x%x(%d)\n",
            (int)entity_base,
            (int)entity_end,
            (int)entity_fin,
            entity_area_size, entity_area_size,
            max_entity_area_size, max_entity_area_size);
        print_con(messtr);
    }

    a2 = (unsigned)a2_val();
    if (a2 != LTAG_NIL)
        a2 = SSERDDA(a2) | LTAG_STRING;
    a3 = (unsigned)active_val();
    a4 = (unsigned)base_val();
    a5 = (unsigned)top_val();
    d3 = (unsigned)reg_val();
    d4 = (unsigned)d4_val();

    if ((tmp = (unsigned *)get_sval(wplist.obj)) != (unsigned *)LTAG_NIL)
        set_sval(wplist.obj, ADDRESS(tmp) | LTAG_NIL);

    /*  if(gctraceflag != 0){
    gctraceflag = 0;
function_debug();
    mark_init_hash();
    gctraceflag = 1; }
  else{
function_debug();
    mark_init_hash();
  } */

    mark_init_hash();
    mark_init_global_var();
    {
	int i;

	for (i = 0; i < sharp_eq_context_max; i++) {
	    mark_unit(&(sharp_eq_context[i].sharp_index));
	    mark_unit(&(sharp_eq_context[i].sharp_eq));
	    mark_unit(&(sharp_eq_context[i].sharp_sharp));
	}
    }
    if (c_strbuf != strbuf) {
        c_strbuf = (char *)(SSERDDA(c_strbuf) | LTAG_STRING);
        mark_unit((unsigned *)&c_strbuf);
    }

    if (message_level >= 2)
        print_con("mark & sweep gc init a2\n");
    mark_unit(&a2);
    if (message_level >= 2)
        print_con("mark & sweep gc init reg\n");
    mark_unit(&d3);
    if (message_level >= 2)
        print_con("mark & sweep gc init stack\n");
    /* stackdump2(1000); */
#ifdef CALLIOC
    mark_init_stack((unsigned *)stack_bottom);
{
    unsigned *tmp;

    while ((unsigned int)marked_stack_list != LTAG_NIL) {
        tmp = marked_stack_list;
        marked_stack_list = (unsigned *)marked_stack_list[2];
        mark_init_stack(tmp);
    }
}
#else
    mark_init_stack((unsigned *)a4, (unsigned *)a5);
#endif
    if (message_level)
        print_con("sweep start\n");
    vacancy_size = sweeper();

#ifdef CALLIOC
    {
        unsigned *sb;

        for (sb = active_stacks; sb != NULL; sb = (unsigned *)(sb[3]))
            if (!sb[2])
                free_stack(sb);
            else
                sb[2] = 0;
    }
#endif

    if ((tmp = (unsigned *)get_sval(wplist.obj)) != (unsigned *)LTAG_NIL)
        set_sval(wplist.obj, LTAG_WEAKPAIR | SSERDDA(ADDRESS(tmp)));

    if (c_strbuf != strbuf)
        c_strbuf = (char *)ADDRESS(c_strbuf);
    if (a2 == LTAG_NIL)
        set_a2(LTAG_NIL);
    else
        set_a2(ADDRESS(a2));
    set_active(a3);
    set_base(a4);
    set_top(a5);
    set_reg(d3);
    set_pc(d4);

    if (message_level)
        print_con("MARK and SWEEP GC end:\n");
    if (message_level >= 2) {
        sprintf(messtr, "             end = 0x%x, c_size = 0x%x(%d)\n",
            (int)entity_end,
            entity_area_size, entity_area_size);
        print_con(messtr);
    }

/*    fclose(debug_stream); */
#ifdef GC_TIME
    getrusage(RUSAGE_SELF, &ru);
    gc_time_sec += (ru.ru_utime.tv_sec - sec);
    gc_time_usec += (ru.ru_utime.tv_usec - usec);
#endif
    return(vacancy_size);
}


void extent_entity_area(vacancy_size, req_size)
int vacancy_size, req_size;
{
    int i, quarter_vacancy;
    char    messtr[256];

    if (message_level)
        print_con("\nextent_entity_area\n\n");
    last_entity = (unsigned *)compaction(vacancy_size, 0);

    for (i = 0; i < BLOCK_TABLE_SIZE; i++)
        block_table[i] = 0;
    big_block = 0;

    if ((entity_end = (unsigned *)(IMM(entity_area_size * 1.5) + (unsigned)entity_base + IMM(req_size))) > entity_fin)
        entity_end = entity_fin;
    entity_area_size = (unsigned)entity_end - (unsigned)entity_base;
    quarter_vacancy = (unsigned)entity_end - (unsigned)last_entity;
    if (quarter_vacancy < 4) {
#ifdef X
        draw_free_list();
#endif      
        return;
    }
    *last_entity = LTAG_BLOCK2 | quarter_vacancy - 4;
    quarter_vacancy /= 4;
    if (quarter_vacancy < BLOCK_TABLE_SIZE) {
        block_table[quarter_vacancy] = last_entity;
        if (quarter_vacancy > 1)
            *(last_entity + 1) = 0;
    } else {
        big_block = last_entity;
        *(last_entity + 1) = 0;
    }
    if (message_level >= 2) {
        sprintf(messtr, "vacancy = %d byte\n", quarter_vacancy << 2);
        print_con(messtr);
        sprintf(messtr, "entity_base = %x, last_entity = %x, entity_end = %x\nentity_end - last_entity = %x\n", (int)entity_base,
                         (int)last_entity,
            (int)entity_end, (unsigned)entity_end - (unsigned)last_entity);
        print_con(messtr);
    }
#ifdef X
    draw_free_list();
#endif      
}


mark_init_hash()
{
#ifdef ASMGC
    int i;
#else
    register int    i;
#endif

    if (message_level >= 2)
        print_con("mark & sweep gc init hash tabel\n");
    for (i = 0; i < hash_table_size; ++i)
        mark_unit(hash_table + i);
    for (i = 0; i < bchashsize; ++i)
        mark_unit(bchash + i);
}


mark_init_global_var()
{
    if (message_level >= 2)
        print_con("mark & sweep gc int global var\n");
    /* 1st : the symbol which the scheme system hold  initialize. */
    mark_unit((unsigned *)&quotesym);
    mark_unit((unsigned *)&qquotesym);
    mark_unit((unsigned *)&atsym);
    mark_unit((unsigned *)&commasym);

    /*
  mark_unit((unsigned *)&loadsym);
  mark_unit((unsigned *)&dumpsym);
  mark_unit((unsigned *)&setsym);
  mark_unit((unsigned *)&endsym);
  mark_unit((unsigned *)&breakfunsym);
  mark_unit((unsigned *)&breaknumsym);
  mark_unit((unsigned *)&tracesym);
  mark_unit((unsigned *)&untracesym);
  mark_unit((unsigned *)&gctracesym);
  mark_unit((unsigned *)&ungctracesym);
  mark_unit((unsigned *)&pctracesym);
  mark_unit((unsigned *)&unpctracesym);
  mark_unit((unsigned *)&settracesym);
  mark_unit((unsigned *)&breakpointsym);
  mark_unit((unsigned *)&dumppointsym);
  mark_unit((unsigned *)&stacksym);
*/

    mark_unit((unsigned *)&definesym);
    mark_unit((unsigned *)&macrosym);
    mark_unit((unsigned *)&xreturnsym);
    mark_unit((unsigned *)&lreturnsym);
    mark_unit((unsigned *)&xcallsym);
    mark_unit((unsigned *)&c_f_sym);
    mark_unit((unsigned *)&mkncsym);
    mark_unit((unsigned *)&sctopsym);
    mark_unit((unsigned *)&errsym);		/* By N.Watanabe 92/11/19 */
    mark_unit((unsigned *)&recomsym);	/* By A243 */
    mark_unit((unsigned *)&recomlevel);	/* By A243 */
#ifdef PVM3
    mark_unit((unsigned *)&bctran_exec_form_sym);	/* By A243 */    
    mark_unit((unsigned *)&sharp_closure_pvm);	/* By A243 */    
    mark_unit((unsigned *)&pvm_recv_function);	/* By A243 */    
    mark_unit((unsigned *)&pvm_reduce_function);	/* By A243 */    
    mark_unit((unsigned *)&pvm_get_unbound_obj);	/* By A243 */    
    mark_unit((unsigned *)&pvm_put_unbound_obj);	/* By A243 */    
#endif
    mark_unit((unsigned *)&debugsym);
    mark_unit((unsigned *)&sys_debug_sym);
    mark_unit((unsigned *)&user_exec_sym);
    mark_unit((unsigned *)&bcloadsym);
    mark_unit((unsigned *)&bc_trace_proc_sym);
    mark_unit((unsigned *)&sysfun_tracer_body);
    mark_unit((unsigned *)&max_back_trace_depth_sym);
    mark_unit((unsigned *)&gc_message_level_sym);
    mark_unit((unsigned *)&msgc_to_copygc_sym);
    mark_unit((unsigned *)&struct_sym);
    mark_unit((unsigned *)&l_t_e_sym);
#ifdef PVM3
    mark_unit((unsigned *)&pvm_l_t_e_sym);      /* By A243 */
#endif
    mark_unit((unsigned *)&local_fun_sym);
    mark_unit((unsigned *)&beginsym);
    mark_unit((unsigned *)&bcload_with_cip_sym);

    /* 2nd: the other object which the scheme system hold */
    mark_unit((unsigned *)&unbound_obj);
    mark_unit((unsigned *)&eof_obj);
    mark_unit((unsigned *)&reg1);
    mark_unit((unsigned *)&reg2);
    mark_unit((unsigned *)&reg3);
    mark_unit((unsigned *)&stdin_port);
    mark_unit((unsigned *)&stdout_port);
    mark_unit((unsigned *)&c_in_port);
    mark_unit((unsigned *)&c_out_port);
    mark_unit((unsigned *)&bcconst);
    mark_unit((unsigned *)&xfunconst);
    mark_unit((unsigned *)&bcraw);
    mark_unit((unsigned *)&bcraw2);
    mark_unit((unsigned *)&bcbody);
    mark_unit((unsigned *)&acode);
    mark_unit((unsigned *)&tempp);
    mark_unit((unsigned *)&cdrbody);
    mark_unit((unsigned *)&bcf);
    mark_unit((unsigned *)&bcfhead);
    mark_unit((unsigned *)&call_cc_arg);
    mark_unit((unsigned *)&top_of_aflk);
    mark_unit((unsigned *)&end_of_aflk);
    mark_unit((unsigned *)&af_tmp);
    mark_unit((unsigned *)&more);
#ifdef CTHREAD
    mark_unit((unsigned *)&optcnt);
#endif
    mark_unit((unsigned *)&portlist);
    mark_unit((unsigned *)&ncclist);
    mark_unit((unsigned *)&wplist);

#ifdef TCPIP
    mark_unit ((unsigned *) &conn_timeout);
    mark_unit ((unsigned *) &tcp_timeout);
    mark_unit ((unsigned *) &udp_timeout);
    mark_unit ((unsigned *) &accept_timeout);
    mark_unit ((unsigned *) &localhostname);
#endif

#ifdef NIHONGO
    mark_unit((unsigned *)&default_kanji_input_code_sym);
    mark_unit((unsigned *)&default_kanji_output_code_sym);
#ifdef TCPIP
    mark_unit((unsigned *)&default_network_kanji_input_code_sym);
    mark_unit((unsigned *)&default_network_kanji_output_code_sym);
#endif
    mark_unit((unsigned *)&euc_sym);
    mark_unit((unsigned *)&jis_sym);
    mark_unit((unsigned *)&sjis_sym);
    mark_unit((unsigned *)&noc_sym);
#endif

    mark_unit((unsigned *)&print_circle_sym);	/* By ushijima */

    if (system_function_loading_flag == ON)
        mark_unit(&protect_symbol_listsym);
}


