;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                     ;;;
;;;           PVM_TUS               by A243                             ;;;
;;;                                                                     ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (pvm-set-env . data)
  (set! *global-symbol* data))

(define (hosts . host)
  (let* ((nhost (length host))
         (infos (make-list nhost #f))
         (info (p-addhosts host nhost infos)))
    (if (= nhost info)
      #t
      (begin (for-each (lambda(x)
			 (if (negative? x) (pvm-error x 'hosts))) infos)
        (change-num-sym infos)))))

(define (spawntus fun . arg)
  (apply raw-spawntus PvmTaskDefault "" fun arg))

(define (spawntus-on-host host fun . arg)
  (apply raw-spawntus PvmTaskHost host fun arg))

(define (spawntus-in-arch arch fun . arg)
  (apply raw-spawntus PvmTaskArch arch fun arg))

(define (raw-spawntus flag where fun . arg)
  (let* ((tids '(#f))
         (task "pvmtus")
         (argv (list "-e(remote-eval)"))
         (ntask 1)
         (numt (p-spawn task argv flag where ntask tids)))
    (if (negative? numt)
      (pvm-error numt 'spawntus)
      (let ((global-symbol-value '())
            (source (get-source fun))
            (name (get-fun-name fun))
            (env (get-closure-env fun))
            (stad (get-closure-stad fun))
            (arginfo (get-closure-arginfo fun)))
        (if (not *pvm-on-demand*)
        (for-each (lambda (x)
                    (let ((value (get-symbol-value x)))
                      (set! global-symbol-value
                        (cons (if (function? value)
                                (if (sysfun? x)
                                  (cons x (cons 'get-gval 
                                    (string->symbol (get-fun-name value))))
                                  (cons #f (cons (get-closure-env value)
                                            (change-pvm-l-t-e-sym value))))
                                (cons x (cons 'get-const value)))
                            global-symbol-value)))) *global-symbol*))
        (pvm-initsend 0)
        (pvm-pk *global-symbol*)
        (pvm-pk (cons env global-symbol-value))
        (pvm-pk (change-source-pvm source))
        (pvm-pk name)
        (pvm-pk stad)
        (pvm-pk arginfo)
        (pvm-pk arg)
        (pvm-send (car tids) 1)
        (car tids)))))

(define (change-source-pvm source)
  (if (eq? (car source) 'load-time-eval)
    (set-car! source 'pvm-load-time-eval)
    (if (eq? (caaddr source) 'load-time-eval)
      (set-car! (caddr source) 'pvm-load-time-eval)))
  source)

(define (change-pvm-l-t-e-sym fun)
  (put-source fun (change-source-pvm (get-source fun)))
  fun)

(define (mspawntus num fun . arg)
  (apply raw-mspawntus PvmTaskDefault "" num fun arg))

(define (mspawntus-on-host host num fun . arg)
  (apply raw-mspawntus PvmTaskHost host num fun arg))

(define (mspawntus-in-arch arch num fun . arg)
  (apply raw-mspawntus PvmTaskArch arch num fun arg))

(define (raw-mspawntus flag where num fun . arg)
  (let* ((tids (make-list num))
         (task "pvmtus")
         (argv (list "-e(remote-eval2)"))
         (flag PvmDataDefault)
         (where "")
         (ntask num)
         (numt (p-spawn task argv flag where ntask tids)))
    (if (= numt num)
      (let ((global-symbol-value '())
            (source (get-source fun))
            (name (get-fun-name fun))
            (env (get-closure-env fun))
            (stad (get-closure-stad fun))
            (arginfo (get-closure-arginfo fun)))
        (if (not *pvm-on-demand*)
        (for-each (lambda (x)
                    (let ((value (get-symbol-value x)))
                      (set! global-symbol-value
                        (cons (if (function? value)
                                (if (sysfun? x)
                                  (cons x (cons 'get-gval 
                                    (string->symbol (get-fun-name value))))
                                  (cons #f (cons (get-closure-env value)
                                            (change-pvm-l-t-e-sym value))))
                                (cons x (cons 'get-const value)))
                            global-symbol-value)))) *global-symbol*))
        (pvm-initsend 0)
        (pvm-pk *global-symbol*)
        (pvm-pk (cons env global-symbol-value))
        (pvm-pk (change-source-pvm source))
        (pvm-pk name)
        (pvm-pk stad)
        (pvm-pk arginfo)
        (pvm-mcast 1 tids)
        (if (not (null? arg))
          (let ((arg (apply map list arg)))
            (for-each (lambda (child)
                        (pvm-initsend 0)
                        (pvm-pk (car arg))
                        (pvm-send child 2)
                        (set! arg (cdr arg))) tids))
          (begin 
            (pvm-initsend 0)
            (pvm-pk arg)
            (pvm-mcast 2 tids)))
        tids)
    (do ((child tids (cdr child))
         (error-num 0))
        ((or (negative? error-num) (null? child))
         (pvm-error error-num 'mspawntus))
      (if (negative? (car child))
        (set! error-num (car child)))))))

(define (reduce func data group tag root)
  (let ((size (if (null? count) (pvm-gsize group) (car count))))
    (if (= root (pvm-getinst group (pvm-mytid)))
      (let ((recv-data (list data)))
        (dotimes (i (1- size))
          (pvm-recv #f tag)
          (set! recv-data (cons (pvm-upk) recv-data))) 
        (apply func recv-data))
      (begin
        (pvm-initsend 0)
        (pvm-pk data)
        (pvm-send (pvm-gettid group root) tag)))))

(define (send data . tid)
  (let ((bufid (p-initsend PvmDataDefault)))
    (if (negative? bufid)
      (pvm-error bufid 'send)
      (let ((info (pvm-pack data)))
        (if (negative? info)
          (pvm-error info 'send)
          (let* ((ntask (length tid))
                 (info (if (= ntask 1) 
                         (if (pair? (car tid))
                           (p-mcast (car tid) (length (car tid)) 1)
                           (p-send (car tid) 1))
                         (p-mcast tid (length tid) 1))))
            (if (negative? info)
              (pvm-error info 'send)
              #t)))))))

(define (send/t tag data . tid)
  (let ((bufid (p-initsend PvmDataDefault)))
    (if (negative? bufid)
      (pvm-error bufid 'send-t)
      (let ((info (pvm-pack data)))
        (if (negative? info)
          (pvm-error info 'send-t)
          (let* ((ntask (length tid))
                 (info (if (= ntask 1)
                         (if (pair? (car tid))
                           (p-mcast (car tid) (length (car tid)) tag)
                           (p-send (car tid) tag))
                         (p-mcast tid (length tid) tag))))
            (if (negative? info)
              (pvm-error info 'send-t)
              #t)))))))

(define (recv . tid)
  (let* ((tid (if (null? tid) -1 (car tid)))
         (bufid (p-recv tid -1)))
    (if (negative? bufid)
      (pvm-error bufid 'recv)
      (read (open-input-string (p-upkstr2))))))

(define (recv/t tag . tid)
  (let* ((tid (if (null? tid) -1 (car tid)))
         (bufid (p-recv tid tag)))
    (if (negative? bufid)
      (pvm-error bufid 'recv-t)
      (read (open-input-string (p-upkstr2))))))

(define (nrecv . tid)
  (let* ((tid (if (null? tid) -1 (car tid)))
         (bufid (p-nrecv tid 1)))
    (if (negative? bufid)
      (pvm-error bufid 'recv-t)
      (if (zero? bufid)
        #f
        (read (open-input-string (p-upkstr2)))))))

(define (nrecv/t tag . tid)
  (let* ((tid (if (null? tid) -1 (car tid)))
         (bufid (p-nrecv tid tag)))
    (if (negative? bufid)
      (pvm-error bufid 'nrecv-t)
      (if (zero? bufid)
        #f
        (read (open-input-string (p-upkstr2)))))))

(define (trecv sec usec . tid)
  (let* ((tid (if (null? tid) -1 (car tid)))
         (bufid (p-trecv tid 1 (list sec usec))))
    (if (negative? bufid)
      (pvm-error bufid 'trecv)
      (if (zero? bufid)
        #f
        (read (open-input-string (p-upkstr2)))))))

(define (trecv/t tag sec usec . tid)
  (let* ((tid (if (null? tid) -1 (car tid)))
         (bufid (p-trecv tid tag (list sec usec))))
    (if (negative? bufid)
      (pvm-error bufid 'trecv-t)
      (if (zero? bufid)
        #f
        (read (open-input-string (p-upkstr2)))))))

;;;;;;;;;;;;;;;;;;;;;;;; SYMPLE FORM ;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (mytid)
  (let ((tid (p-mytid)))
    (if (negative? tid)
      (pvm-error tid 'mytid)
      tid)))

(define (kill tid)
  (let ((info (p-kill tid)))
    (if (negative? info)
      (pvm-error tid 'kill)
      #t)))

(define (parent)
  (let ((tid (p-parent)))
    (if (negative? tid)
      (pvm-error tid 'parent)
      tid)))

(define (joingroup group)
  (let ((inum (p-joingroup group)))
    (if (negative? inum)
      (pvm-error inum 'joingroup)
      inum)))

(define (lvgroup group)
  (let ((info (p-lvgroup group)))
    (if (negative? info)
      (pvm-error info 'lvgroup)
      #t)))

(define (gsize group)
  (let ((size (p-gsize group)))
    (if (negative? size)
      (pvm-error size 'gsize)
      size)))

(define (gettid group inum)
  (let ((tid (p-gettid group inum)))
    (if (negative? tid)
      (pvm-error tid 'gettid)
      tid)))

(define (getinst group tid)
  (let ((inum (p-getinst group tid)))
    (if (negative? inum)
      (pvm-error inum 'getinst)
      inum)))

(define (barrier group count)
  (let ((info (p-barrier group count)))
    (if (negative? info)
      (pvm-error info 'barrier)
      #t)))

(define (bcast group data)
  (let ((bufid (p-initsend PvmDataDefault)))
    (if (negative? bufid)
      (pvm-error bufid 'bcast)
      (let ((info (pvm-pack data)))
        (if (negative? info)
          (pvm-error info 'bcast)
          (let ((info (p-bcast group 1)))
            (if (negative? info)
              (pvm-error info 'bcast)
              #t)))))))

(define (bcast/t msgtag group data)
  (let ((bufid (p-initsend PvmDataDefault)))
    (if (negative? bufid)
      (pvm-error bufid 'bcast)
      (let ((info (pvm-pack data)))
        (if (negative? info)
          (pvm-error info 'bcast)
          (let ((info (p-bcast group msgtag)))
            (if (negative? info)
              (pvm-error info 'bcast)
              #t)))))))

