/*
 * sctop.c - 
 */

/*
 * $Id: sctop.c,v 1.28 2016/02/23 09:06:57 komiya Exp $
 *
 * Scheme system's main routines File Name : sctop.c written by Mikio
 * Yamamoto.
 * 
 */

static char *RCSid="$Id: sctop.c,v 1.28 2016/02/23 09:06:57 komiya Exp $";

#include "scheme.h"
#include <signal.h>
#include <sys/wait.h>
#ifndef WAITPID
#include <sys/time.h>
#include <sys/resource.h>
#endif

#ifdef TK
#include <tcl.h>
#include <tk.h>
#endif

#ifdef THINK_C
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef THINK_C
#include <Events.h>
#include <Memory.h>
#endif

#ifdef TCPIP
#include <sys/param.h>
#include "inet.h"
char *ourhostname;
#endif

#ifdef PVM3
#define TUS_VERSION     "1.0j [PVM]"
#else
#define TUS_VERSION     "1.4g"
#endif

#ifdef PVM3
#define PVM_ERROR_TAG 33554431
#endif

#ifdef THREAD
#ifdef SUNSOL2
#include <unistd.h>
#endif
#endif

#ifdef LTC
#undef  TUS_VERSION
# ifdef CONDITIONAL_FUTURE
#define TUS_VERSION     "1.4g [LTC] [CONDITIONAL_FUTURE]"
# else
#define TUS_VERSION     "1.4g [LTC]"
# endif
#endif

/* global variabel */
unsigned *block_table[BLOCK_TABLE_SIZE];
unsigned *big_block;

/* initial memory allocation */

#if defined(PVM3) || defined(TK) || defined(__CYGWIN32__)
#define     DEFAULT_MAX_HEAP_SIZE   0x200000
#else
#define     DEFAULT_MAX_HEAP_SIZE   0x100000
#endif
#define     HEAP_SIZE_UNIT          0x10000

/* #define     DEFAULT_SYS_HEAP_SIZE       2048 *//* == bytes/sizeof(HEADER) */
#ifdef CALLIOC
#define     DEFAULT_SYS_HEAP_SIZE       32768 /* == bytes/sizeof(HEADER) */
#else
#define     DEFAULT_SYS_HEAP_SIZE       16384 /* == bytes/sizeof(HEADER) */
#endif

/* the symbol used a top level routine */
int pctraceflag;
int top();
#if defined(__CYGWIN32__) || defined(linux)
sigjmp_buf toplevel;
#else
jmp_buf toplevel;
#endif
int tracelevel;
int gctraceflag;
extern int  memgettraceflag;
extern int  sweepertraceflag;
extern int sharp_eq_context_max;
extern int init_system_heap;
extern int symbol_sensitive;
void system_program_loader(char *, int);
void stackdump(int, int);
#ifdef THINK_C
#define         ErrorAlert      256
#endif

#ifdef CTHREAD
cthread_t main_cthread_id;
extern mutex_t addsymbol_lock;
extern mutex_t memget_lock;
extern mutex_t cthread_fork_lock;
extern mutex_t print_lock;
extern mutex_t read_lock;
#endif

#ifdef PTHREAD
pthread_attr_t pattr;
struct sched_param pschedparam;
# ifdef CONDITIONAL_FUTURE
pthread_attr_t sattr;
struct sched_param sschedparam;
# endif
#endif

#ifdef PVM3   /* By A243 */
int kill_pvm_family_process();
int *serch_kill_process();
int sigint_root_process();
int serch_root_process();
extern int gstid;
int pvm_on_demand();
#endif

#ifdef TK
Tcl_Interp *interp;
Tk_Window mainWindow;
extern int StdinProc();
extern int bufidx, count, filled, ungetted, previous_char, Tk_initialized;
extern int apply_fun(), do_one_event();
# ifdef __CYGWIN32__
jmp_buf jb;
# endif
#endif

/* for freeze */
int freeze_flag = 0;
char    stdin_buf[BUFSIZ], stdout_buf[BUFSIZ];

char tus_self[256];
char faspath[256];

int sigchld_enable;

meminit_for_freeze()
{
    unsigned *entity;

#ifdef CALLIOC
    *(unsigned *)(stack_bottom + 4) = (unsigned)(stack_bottom - 12);
    *(unsigned *)(stack_bottom + 8) = 0;
    *(unsigned *)(stack_bottom + 12) = 0;
#else
    *(unsigned *)(stack_bottom + 4) = LTAG_NIL;
    *(stack_bottom + 8) = 0;
    *(stack_bottom + 12) = 0;
#endif
    set_top(stack_bottom);
    old_top = (unsigned *) NULL;
    old_top1 = old_top;
    hash_bit_mask = DEFAULT_HASH_B_M;
    bcraw.obj = LTAG_NIL;
    bcraw2.obj = LTAG_NIL;
    bcbody.obj = LTAG_NIL;
    tempp.obj = LTAG_NIL;
    cdrbody.obj = LTAG_NIL;
    acode.obj = LTAG_NIL;
    bcf.obj = LTAG_NIL;
    bcconst.obj = LTAG_NIL;
    xfunconst.obj = LTAG_NIL;
    bcfhead.obj = LTAG_NIL;
    set_a2(LTAG_NIL);

    reg1 = LTAG_NIL;
    reg2 = LTAG_NIL;
    reg3 = LTAG_NIL;

    call_cc_arg.obj = LTAG_NIL;
    top_of_aflk.obj = LTAG_NIL;
    end_of_aflk.obj = LTAG_NIL;
    af_tmp.obj = LTAG_NIL;
    more.obj = LTAG_NIL;
#ifdef CTHREAD
    optcnt.obj = LTAG_NIL;
#endif

    set_sval(portlist.obj, LTAG_NIL);
    set_sval(ncclist.obj, LTAG_NIL);

#ifdef TCPIP
    set_sval (conn_timeout.obj, LTAG_FALSE);
    set_sval (tcp_timeout.obj, LTAG_FALSE);
    set_sval (udp_timeout.obj, LTAG_FALSE);
    set_sval (accept_timeout.obj, LTAG_FALSE);

    if ((ourhostname = (char *) malloc (MAXHOSTNAMELEN)) == NULL)
      ourhostname = "???";
    else
      gethostname (ourhostname, MAXHOSTNAMELEN);

    set_sval (localhostname.obj,
              mk_str (ORDINARY_STRING, A_STRING,
                      strlen (ourhostname), ourhostname));
#endif

    entity = (unsigned *)ADDRESS(stdin_port.obj);
    *(entity + 1) = LTAG_NIL;
    *(entity + 3) = (unsigned)stdin;
    *(entity + 4) = INPUT;
    *(entity + 5) = 0;

    entity = (unsigned *)ADDRESS(stdout_port.obj);
    *(entity + 1) = LTAG_NIL;
    *(entity + 3) = (unsigned)stdout;
    *(entity + 4) = OUTPUT;
    *(entity + 5) = 0;

    c_in_port.obj = stdin_port.obj;
    c_out_port.obj = stdout_port.obj;
}


meminit()
{
    char    *a;
    int i;

#ifdef __CYGWIN32__
# define G_VAR_NUM (100+SHARP_EQ_CONTEXT_SIZE*3+2) /* The # of global variable and regs */
    extern unsigned *dummy_source, *src_tmp, *g_var_gc_tmp;
    extern char g_var_gc_tmp_tmp[];
    dummy_source = (unsigned *)sbrk(8);
    src_tmp = (unsigned *)sbrk(8);
    g_var_gc_tmp = (unsigned *)sbrk(G_VAR_NUM * sizeof(unsigned *) + 4);
#endif

    debug1 = 1;
    traceflag = 0;
    pctraceflag = 0;
    gctraceflag = 0;
    sweepertraceflag = 0;
    memgettraceflag = 0;

#ifdef THINK_C
/*      MaxApplZone(); */
    if ((stack_top = (char *)NewPtr(STACK_SIZE+4)) == NULL) {
        printf("not enough memory.\n");
        exit(1);
    }
    stack_top += ((unsigned)stack_top % 4);
#elif defined(CTHREAD)
    stack_top = (char *)malloc(STACK_SIZE);
#else
    stack_top = (char *)sbrk(STACK_SIZE);
#endif
    stack_bottom = stack_top + STACK_SIZE - 28;
#ifdef CALLIOC
    active_stacks = (unsigned *)stack_bottom;
#endif
    stack_top = stack_top + GETA_OF_STACK;
#ifdef CALLIOC
    *(unsigned *)(stack_bottom + 4) = (unsigned)(stack_bottom - 12);
    *(unsigned *)(stack_bottom + 8) = 0;
    *(unsigned *)(stack_bottom + 12) = 0;
#else
    *(unsigned *)(stack_bottom + 4) = LTAG_NIL;
    *(stack_bottom + 8) = 0;
    *(stack_bottom + 12) = 0;
#endif
    stack_area_bottom = stack_bottom;
    set_top(stack_bottom);
    old_top = (unsigned *) NULL;
    old_top1 = old_top;

#ifdef THINK_C
    if ((hash_table = (unsigned *)NewPtr(HASH_SIZE+4)) == NULL) {
        printf("not enough memory.\n");
        exit(1);
    }
    hash_table = (unsigned *)((char *)hash_table + ((unsigned)hash_table % 4));
#elif defined(CTHREAD)
    hash_table = (unsigned *)malloc(HASH_SIZE);
#else
    hash_table = (unsigned *)sbrk(HASH_SIZE);
#endif
    hash_table_size = HASH_SIZE >> 2;
    hash_bit_mask = DEFAULT_HASH_B_M;
    for (i = 0; i < hash_table_size; ++i)
        *(hash_table + i) = LTAG_NIL;

#ifdef THINK_C
    if ((bchash = (unsigned *)NewPtr(BCHASH_SIZE+4)) == NULL) {
        printf("not enough memory.\n");
        exit(1);
    }
    bchash = (unsigned *)((char *)bchash + ((unsigned)bchash % 4));
#elif defined(CTHREAD)
    bchash = (unsigned *)malloc(BCHASH_SIZE);
#else
    bchash = (unsigned *)sbrk(BCHASH_SIZE);
#endif
    bchashsize = BCHASH_SIZE >> 2;
    for (i = 0; i < bchashsize; ++i)
        *(bchash + i) = LTAG_NIL;
    bcraw.obj = LTAG_NIL;
    bcraw2.obj = LTAG_NIL;
    bcbody.obj = LTAG_NIL;
    tempp.obj = LTAG_NIL;
    cdrbody.obj = LTAG_NIL;
    acode.obj = LTAG_NIL;
    bcf.obj = LTAG_NIL;
    bcconst.obj = LTAG_NIL;
    xfunconst.obj = LTAG_NIL;
    bcfhead.obj = LTAG_NIL;

#ifdef THINK_C
    if ((strbuf = (char *)NewPtr(STRBUFSIZE+4)) == NULL) {
        printf("not enough memory.\n");
        exit(1);
    }
    strbuf += ((unsigned)strbuf % 4);
#elif defined(CTHREAD)
    strbuf = (char *)malloc(STRBUFSIZE);
#else
    strbuf = (char *)sbrk(STRBUFSIZE);
#endif
    c_strbuf = strbuf;
    transcript_flag = TRANSCRIPT_OFF;

    bcdatainit();
#if (!defined(THINK_C) && !defined(CTHREAD) && !defined(PVM3) && !defined(TK) \
     && !defined(__CYGWIN32__))
    sheap_start = (HEADER * )sbrk(DEFAULT_SYS_HEAP_SIZE * sizeof(HEADER));
    sheap_end = sheap_start + DEFAULT_SYS_HEAP_SIZE;
    sheap_start->size = DEFAULT_SYS_HEAP_SIZE;
#endif
#ifdef THINK_C
    if ((entity_base = (unsigned *) NewPtr(max_entity_area_size+4)) == NULL) {
        printf("not enough memory.\n");
        exit(1);
    }
    entity_base = (unsigned *)((char *)entity_base + ((unsigned)entity_base % 4));
#elif defined(CTHREAD)
    entity_base = (unsigned *)malloc(max_entity_area_size);
#else
    entity_base = (unsigned *) sbrk(max_entity_area_size);
#endif
    entity_fin = (unsigned *) (((char *) entity_base) + max_entity_area_size);
    entity_top = entity_base;
    entity_middle = (unsigned *)(((char *) entity_base) + (entity_area_size / 2));
    entity_bottom = (unsigned *)(((char *) entity_middle) + (entity_area_size / 2));
    entity_start = entity_base;
    entity_end = entity_middle;
    last_entity = entity_base;
    entity_position = HIGH;
    extent_rate = 1.8;
    gctype = COPY_GC;

    set_a2(LTAG_NIL);

    reg1 = LTAG_NIL;
    reg2 = LTAG_NIL;
    reg3 = LTAG_NIL;

    call_cc_arg.obj = LTAG_NIL;
    top_of_aflk.obj = LTAG_NIL;
    end_of_aflk.obj = LTAG_NIL;
    af_tmp.obj = LTAG_NIL;
    more.obj = LTAG_NIL;
#ifdef CTHREAD
    optcnt.obj = LTAG_NIL;
#endif
    init_system_heap = 1;       /* T.Komiya */
}


int Ssysfun_tracer();       /* in debug.c */
int exec_sysfun();      /* in bci.c   */

define_special_object()
{
    unsigned *p;
    unbound_obj.obj = mkcons(LTAG_NIL, LTAG_NIL);
/*    eof_obj.obj = mkcons(LTAG_NIL, LTAG_NIL); */
    eof_obj.obj = LTAG_EOF;
    p = (unsigned *) memget(12);
    *p = LTAG_BLOCK2 | 0x8;
    *(p + 1) = (unsigned) exec_sysfun;
    *(p + 2) = (unsigned) Ssysfun_tracer;
    sysfun_tracer_body.obj = LTAG_STRING | SSERDDA(p);

    stdin_port.obj = mkstdinport();
    stdout_port.obj = mkstdoutport();
    c_in_port.obj = stdin_port.obj;
    c_out_port.obj = stdout_port.obj;
}


define_symbol()
{
    sctopsym = define_a_ordinary_symbol("sctop");
    errsym   = define_a_ordinary_symbol("error"); /* By N.Watanabe 92/11/19 */
    recomsym = define_a_ordinary_symbol("recom"); /* By A243 */
    recomlevel = define_a_ordinary_symbol("*re-compile-level*"); /* By A243 */
#ifdef PVM3   /* By A243 */
    bctran_exec_form_sym = define_a_ordinary_symbol("bctran-exec-form"); 
    sharp_closure_pvm = define_a_ordinary_symbol("sharp-closure-pvm"); 
    pvm_recv_function = define_a_ordinary_symbol("*pvm-recv-function*"); 
    pvm_reduce_function = define_a_ordinary_symbol("*pvm-reduce-function*"); 
    pvm_get_unbound_obj = define_a_ordinary_symbol("pvm-get-unbound-obj");
    pvm_put_unbound_obj = define_a_ordinary_symbol("pvm-put-unbound-obj");
#endif
    debugsym = define_a_ordinary_symbol("*debug-level*");
    user_exec_sym = define_a_ordinary_symbol("*user-exec*");
    sys_debug_sym = define_a_ordinary_symbol("*system-debug*");
    set_sval(debugsym, SCOBJ_0);
#ifdef PVM3
    set_sval(sharp_closure_pvm, LTAG_NIL); /* By A243 */
    set_sval(pvm_recv_function, SCOBJ_0); /* By A243 */
    set_sval(pvm_reduce_function, SCOBJ_0); /* By A243 */
#endif
    set_sval(recomlevel, SCOBJ_0);      /* By A243 */
    set_sval(user_exec_sym, LTAG_T);
    set_sval(sys_debug_sym, LTAG_NIL);
    protect_symbol_listsym = define_a_ordinary_symbol("*protect-symbols*");
    set_sval(protect_symbol_listsym, LTAG_NIL);

    bc_trace_proc_sym.obj = define_a_ordinary_symbol("bc-trace-proc");
    max_back_trace_depth_sym.obj = define_a_ordinary_symbol("*max-back-trace-depth*");
    gc_message_level_sym.obj = define_a_ordinary_symbol("*gc-message-level*");
    set_sval(gc_message_level_sym.obj, LTAG_NIL);
    msgc_to_copygc_sym.obj = define_a_ordinary_symbol("*msgc2copygc*");

    xreturnsym.obj = define_a_ordinary_symbol("xreturn");
    lreturnsym.obj = define_a_ordinary_symbol("lreturn");
    xcallsym.obj = define_a_ordinary_symbol("xcall");
    mkncsym.obj = define_a_ordinary_symbol("make-nil-closure");

    wplist.obj = define_a_ordinary_symbol("*wplist*");
    set_sval(wplist.obj, LTAG_NIL);
    portlist.obj = define_a_ordinary_symbol("*portlist*");
    set_sval(portlist.obj, LTAG_NIL);
    ncclist.obj = define_a_ordinary_symbol("*ncclist*");
    set_sval(ncclist.obj, LTAG_NIL);

#ifdef TCPIP
    conn_timeout.obj = define_a_ordinary_symbol ("*connection-timeout*");
    set_sval (conn_timeout.obj, LTAG_FALSE);
    tcp_timeout.obj = define_a_ordinary_symbol ("*tcp-timeout*");
    set_sval (tcp_timeout.obj, LTAG_FALSE);
    udp_timeout.obj = define_a_ordinary_symbol ("*udp-timeout*");
    set_sval (udp_timeout.obj, LTAG_FALSE);
    accept_timeout.obj = define_a_ordinary_symbol ("*accept-timeout*");
    set_sval (accept_timeout.obj, LTAG_FALSE);

    localhostname.obj = define_a_ordinary_symbol ("*hostname*");

    if ((ourhostname = (char *) malloc (MAXHOSTNAMELEN)) == NULL)
      ourhostname = "???";
    else
      gethostname (ourhostname, MAXHOSTNAMELEN);

    set_sval (localhostname.obj,
              mk_str (ORDINARY_STRING, A_STRING,
                      strlen (ourhostname), ourhostname));
#endif

#ifdef NIHONGOSV
    euc_sym.obj = define_a_ordinary_symbol("euc");
    set_sval(euc_sym.obj, SCOBJ_0);
    jis_sym.obj = define_a_ordinary_symbol("jis");
    set_sval(jis_sym.obj, SCOBJ_1);
    sjis_sym.obj = define_a_ordinary_symbol("sjis");
    set_sval(sjis_sym.obj, SCOBJ_2);
    noc_sym.obj = define_a_ordinary_symbol("noc");
    set_sval(noc_sym.obj, SCOBJ_3);

    default_kanji_input_code_sym.obj = define_a_ordinary_symbol("*default-kanji-input-code*");
    set_sval(default_kanji_input_code_sym.obj, SCOBJ_0);
    
    default_kanji_output_code_sym.obj = define_a_ordinary_symbol("*default-kanji-output-code*");
    set_sval(default_kanji_output_code_sym.obj, SCOBJ_0);
#endif /* NIHONGOSV */

#ifdef NIHONGO
    euc_sym.obj = define_a_ordinary_symbol("euc");
    set_sval(euc_sym.obj, SCOBJ_0);
    jis_sym.obj = define_a_ordinary_symbol("jis");
    set_sval(jis_sym.obj, SCOBJ_1);
    sjis_sym.obj = define_a_ordinary_symbol("sjis");
    set_sval(sjis_sym.obj, SCOBJ_2);
    noc_sym.obj = define_a_ordinary_symbol("noc");
    set_sval(noc_sym.obj, SCOBJ_3);

    default_kanji_input_code_sym.obj = define_a_ordinary_symbol("*default-kanji-input-code*");
    set_sval(default_kanji_input_code_sym.obj, LTAG_SYMBOL | euc_sym.obj);
    
    default_kanji_output_code_sym.obj = define_a_ordinary_symbol("*default-kanji-output-code*");
    set_sval(default_kanji_output_code_sym.obj, LTAG_SYMBOL | euc_sym.obj);

#ifdef TCPIP
    default_network_kanji_input_code_sym.obj =
            define_a_ordinary_symbol("*default-network-kanji-input-code*");
    set_sval(default_network_kanji_input_code_sym.obj, LTAG_SYMBOL | jis_sym.obj);

    default_network_kanji_output_code_sym.obj =
            define_a_ordinary_symbol("*default-network-kanji-output-code*");
    set_sval(default_network_kanji_output_code_sym.obj, LTAG_SYMBOL | jis_sym.obj);
#endif

#endif /* NIHONGO */

    print_circle_sym.obj = define_a_ordinary_symbol("*print-circle*");  /* By ushijima */
    set_sval(print_circle_sym.obj, LTAG_T);

    struct_sym.obj = define_a_ordinary_symbol("structure-size");

    define_a_special_symbol("if");
    define_a_special_symbol("case");
    define_a_special_symbol("cond");
    define_a_special_symbol("else");
    define_a_special_symbol("and");
    define_a_special_symbol("or");
    define_a_special_symbol("let");
    define_a_special_symbol("let*");
    define_a_special_symbol("letrec");
    define_a_special_symbol("do");
    define_a_special_symbol("set!");
    define_a_special_symbol("lambda");
    define_a_special_symbol("delay");

    define_a_special_symbol("trace");
    define_a_special_symbol("untrace");

    define_a_special_symbol("compile-file");
    define_a_special_symbol("call-with-input-file");
    define_a_special_symbol("call-with-output-file");
    define_a_special_symbol("open-input-file");
    define_a_special_symbol("open-output-file");
/*    define_a_special_symbol("transcript-on");
    define_a_special_symbol("transcript-off"); */
    define_a_special_symbol("file-exists?");
    quotesym.obj = define_a_special_symbol("quote");
    qquotesym.obj = define_a_special_symbol("quasiquote");
    atsym.obj = define_a_special_symbol("unquote-splicing");
    commasym.obj = define_a_special_symbol("unquote");
    bcloadsym = define_a_special_symbol("bcload");
    bcload_with_cip_sym.obj
        = define_a_special_symbol("bcload/current-input-port");

    definesym.obj = define_a_special_symbol("define");
    macrosym.obj = define_a_special_symbol("macro");
    c_f_sym.obj = define_a_special_symbol("compiled-function");
    l_t_e_sym.obj = define_a_special_symbol("load-time-eval");
#ifdef PVM3
    pvm_l_t_e_sym.obj = define_a_special_symbol("pvm-load-time-eval"); 
                                                    /* By A243 */
#endif
    local_fun_sym.obj = define_a_special_symbol("local-function");
    beginsym.obj = define_a_special_symbol("begin");

    set_sval(define_a_special_symbol("*stdout*"), stdout_port.obj);
    set_sval(define_a_special_symbol("*stdin*"), stdin_port.obj);
    set_sval(define_a_special_symbol("t"), LTAG_T);
    set_sval(define_a_special_symbol("nil"), LTAG_NIL);

    set_sval(define_a_ordinary_symbol("*faslib-path*"),
             mk_str(ORDINARY_STRING, A_STRING, strlen(faspath), faspath));
}


define_a_special_symbol(str)
char    *str;
{
    return (addsymbol(mk_str(SYMBOL_STRING, A_STRING, strlen(str), str),
        LTAG_SPECIAL));
}


define_a_ordinary_symbol(str)
char    *str;
{
    return (addsymbol(mk_str(SYMBOL_STRING, A_STRING, strlen(str), str),
        LTAG_ORDINARY));
}


strtonum(str)
char    *str;
{
    int n;
    for (n = 0; *str != '\0'; ++str)
        if ((*str >= '0') && (*str <= '9'))
            n = (n * 10) + ((int) (*str)) - '0';
    return (n);
}

#ifdef LTC
#include "ltc.h"
extern int processors;
extern thread_data_t **pthrs;
# ifdef CONDITIONAL_FUTURE
extern thread_data_t **sthrs;
# endif
#endif

#ifndef THINK_C
void sigalrm(int sig)
{
    void interr();
#ifdef TCPIP
    void timeout_err();
# ifdef SIGACTION
    struct sigaction act;
# else
    struct sigvec svec;
# endif
#endif
    
    if (!interrupt_enable) {
        signal(SIGALRM, sigalrm);
        alarm(1);
        return;
    }
    signal(SIGINT, SIG_IGN);
    alarm(0);
#ifdef TCPIP
# ifdef SIGACTION
    act.sa_handler = (void (*)())timeout_err;
    sigemptyset(&(act.sa_mask));
    sigaddset(&(act.sa_mask), SIGINT);
    act.sa_flags = 0;
    sigaction(SIGALRM, &act, (struct sigaction *)NULL);
# else
    svec.sv_handler = (void (*) ()) timeout_err;
    svec.sv_mask = SIGINT;
    svec.sv_flags = SV_INTERRUPT;
    sigvec (SIGALRM, &svec, (struct sigvec *) 0);
# endif
#endif

#ifdef LTC
    if ((unsigned int **)stack_top == pthrs[0]->stack)
        for (;;) {
            int i;

#ifdef CONDITIONAL_FUTURE
            int j;
#endif
            /* detect still-alive primary processors */
            for (i = 1; (i < processors) && (pthrs[i]->thr == NULL); i++)
                ;
#ifdef CONDITIONAL_FUTURE
            /* if all primary processors are died? */
            if (i == processors) {
                for (i = 1; i < processors; i++) {
                    free((char *)(pthrs[i]->stack) - CHILD_GETA_OF_STACK);
                    free((char *)(pthrs[i]->cstack) - CHILD_GETA_OF_STACK);
                    free(pthrs[i]->regs);
                    free(pthrs[i]->cregs);
                    free(pthrs[i]->mutex);
                    free(pthrs[i]);
                }
                /* detect still-alive secondary processors */
                for (j = 0; (j < processors) && (sthrs[j]->thr == NULL); i++)
                    ;
                /* if all secondary processors are died? */
                if (j == processors) {
                    for (j = 0; j < processors; j++) {
                        free((char *)(sthrs[j]->stack) - CHILD_GETA_OF_STACK);
                        free((char *)(sthrs[j]->cstack) - CHILD_GETA_OF_STACK);
                        free(sthrs[j]->regs);
                        free(sthrs[j]->cregs);
                        free(sthrs[j]->mutex);
                        free(sthrs[j]);
                    }
                    mutex_unlock(addsymbol_lock);
                    mutex_unlock(memget_lock);
                    mutex_unlock(cthread_fork_lock);
                    mutex_unlock(print_lock);
                    mutex_unlock(read_lock);
                    back_trace();
#if defined(__CYGWIN32__) || defined(linux)
                    siglongjmp(toplevel, 1);
#else
                    longjmp(toplevel, 1);
#endif
                } else {
                    pthread_kill(sthrs[i]->thr, SIGINT);
                    pause();
                }
            } else {
                pthread_kill(pthrs[i]->thr, SIGINT);
                pause();
            }
#else 
            /* if all primary processors are died? */
            if (i == processors) {
                for (i = 1; i < processors; i++) {
                    free((char *)(pthrs[i]->stack) - CHILD_GETA_OF_STACK);
                    free(main_thread_regs);
                    free(pthrs[i]->mutex);
                    free(pthrs[i]);
                }
                mutex_unlock(addsymbol_lock);
                mutex_unlock(memget_lock);
                mutex_unlock(cthread_fork_lock);
                mutex_unlock(print_lock);
                mutex_unlock(read_lock);
                back_trace();
#if defined(__CYGWIN32__) || defined(linux)
                siglongjmp(toplevel, 1);
#else
                longjmp(toplevel, 1);
#endif
            } else {
                pthread_kill(pthrs[i]->thr, SIGINT);
                pause();
            }
#endif
        }
    else {
        int i;
#ifdef CONDITIONAL_FUTURE
        int j;
#endif
        /* detect own processor from primary */
        for (i = 1; (i < processors) && (pthrs[i]->stack != (unsigned int **)stack_top); i++)
            ;
#ifdef CONDITIONAL_FUTURE
        /* detect own processor from secondary */
/*        for (j = 0; (i < processors) && (sthrs[i]->stack != (unsigned int **)stack_top); j++) */
        for (j = 0; (j < processors) && (sthrs[j]->stack != (unsigned int **)stack_top); j++)
            ;
#endif
        /*
         * free own thread
         */
#ifdef CONDITIONAL_FUTURE
        if (i == processors && j == processors) {
            fprintf(stderr, "I'm already killed???\n");
        }
        /* own processor is primary */
        else if (i != processors) {
            pthrs[i]->thr = NULL;
            /* detect other still-alive processors */
            for (i = 1; (i < processors) && (pthrs[i]->thr == NULL); i++)
                ;
            for (j = 0; (j < processors) && (sthrs[j]->thr == NULL); i++)
                ;
            if (i == processors && j == processors) {
                pthread_kill(pthrs[0]->thr, SIGINT);
            } else if (i != processors) {
                pthread_kill(pthrs[i]->thr, SIGINT);
            } else if (j != processors) {
                pthread_kill(sthrs[j]->thr, SIGINT);
            }
            printf("Thread #%x(p) is killed.\n", mtus_thread_self());
            pthread_exit((void *)0);
        }
        /* own processor is secondary */
        else if (j != processors) {
            sthrs[j]->thr = NULL;
            /* detect other still-alive processors */
            for (i = 1; (i < processors) && (pthrs[i]->thr == NULL); i++)
                ;
            for (j = 0; (j < processors) && (sthrs[j]->thr == NULL); i++)
                ;
            if (i == processors && j == processors) {
                pthread_kill(pthrs[0]->thr, SIGINT);
            } else if (i != processors) {
                pthread_kill(pthrs[i]->thr, SIGINT);
            } else if (j != processors) {
                pthread_kill(sthrs[j]->thr, SIGINT);
            }
            printf("Thread #%x(s) is killed.\n", mtus_thread_self());
            pthread_exit((void *)0);
        }
#else 
        if (i == processors)
            fprintf(stderr, "I'm already killed???\n");
        pthrs[i]->thr = NULL;
        /* detect other still-alive processors */
        for (i = 1; (i < processors) && (pthrs[i]->thr == NULL); i++)
            ;
        if (i == processors) {
            pthread_kill(pthrs[0]->thr, SIGINT);
        } else {
            pthread_kill(pthrs[i]->thr, SIGINT);
        }
        printf("Thread #%x is killed.\n", mtus_thread_self());
        pthread_exit((void *)0);
#endif  
    }
        
#else /* not LTC */
# ifdef CTHREAD
    mutex_unlock(addsymbol_lock);
    mutex_unlock(memget_lock);
    mutex_unlock(cthread_fork_lock);
    mutex_unlock(print_lock);
    mutex_unlock(read_lock);
# endif
    back_trace();               /* By N.Watanabe 92/11/29 */
#if defined(__CYGWIN32__) || defined(linux)
    siglongjmp(toplevel, 1);
#else
    longjmp(toplevel, 1);
#endif
#endif /* LTC */
}

#endif


#ifdef THINK_C
void
#endif
void interr(sig)
int sig;
{
    int obj ;
#ifdef TCPIP
    void timeout_err();
# ifdef SIGACTION
    struct sigaction act;
# else
    struct sigvec svec;
# endif
#endif

    if (!interrupt_enable) {
#ifndef THINK_C
        signal(SIGALRM, sigalrm);
        alarm(1);
        signal(SIGINT, interr);
#endif
        return;
    }
#ifndef THINK_C
    signal(SIGALRM, SIG_IGN);
    alarm(0);
#endif
#ifdef TCPIP
# ifdef SIGACTION
    act.sa_handler = (void (*)())timeout_err;
    sigemptyset(&(act.sa_mask));
    sigaddset(&(act.sa_mask), SIGINT);
    act.sa_flags = 0;
    sigaction(SIGALRM, &act, (struct sigaction *)NULL);
# else
    svec.sv_handler = (void (*) ()) timeout_err;
    svec.sv_mask = SIGINT;
    svec.sv_flags = SV_INTERRUPT;
    sigvec (SIGALRM, &svec, (struct sigvec *) 0);
# endif
#endif

#ifdef THINK_C
    signal(SIGINT, (void(*)())interr);
#else
    signal(SIGINT, interr);
#endif

    /* err_interrupt(); */     /* By N.Watanabe 92/11/26 */
#ifdef PVM3      /* By A243 */
if(pvm_parent() == -23)
    kill_pvm_family_process();
else
    sigint_root_process();
#endif

#ifdef LTC
    if ((unsigned int **)stack_top == pthrs[0]->stack)
        for (;;) {
            int i;

#ifdef CONDITIONAL_FUTURE
            int j;
#endif
            /* detect still-alive primary processors */
            for (i = 1; (i < processors) && (pthrs[i]->thr == NULL); i++)
                ;
#ifdef CONDITIONAL_FUTURE
            /* if all primary processors are died? */
            if (i == processors) {
                for (i = 1; i < processors; i++) {
                    free((char *)(pthrs[i]->stack) - CHILD_GETA_OF_STACK);
                    free((char *)(pthrs[i]->cstack) - CHILD_GETA_OF_STACK);
                    free(pthrs[i]->regs);
                    free(pthrs[i]->cregs);
                    free(pthrs[i]->mutex);
                    free(pthrs[i]);
                }
                /* detect still-alive secondary processors */
                for (j = 0; (j < processors) && (sthrs[j]->thr == NULL); i++)
                    ;
                /* if all secondary processors are died? */
                if (j == processors) {
                    for (j = 0; j < processors; j++) {
                        free((char *)(sthrs[j]->stack) - CHILD_GETA_OF_STACK);
                        free((char *)(sthrs[j]->cstack) - CHILD_GETA_OF_STACK);
                        free(sthrs[j]->regs);
                        free(sthrs[j]->cregs);
                        free(sthrs[j]->mutex);
                        free(sthrs[j]);
                    }
                    mutex_unlock(addsymbol_lock);
                    mutex_unlock(memget_lock);
                    mutex_unlock(cthread_fork_lock);
                    mutex_unlock(print_lock);
                    mutex_unlock(read_lock);
                    back_trace();
#if defined(__CYGWIN32__) || defined(linux)
                    siglongjmp(toplevel, 1);
#else
                    longjmp(toplevel, 1);
#endif
                } else {
                    pthread_kill(sthrs[i]->thr, SIGINT);
                    pause();
                }
            } else {
                pthread_kill(pthrs[i]->thr, SIGINT);
                pause();
            }
#else 
            /* if all primary processors are died? */
            if (i == processors) {
                for (i = 1; i < processors; i++) {
                    free((char *)(pthrs[i]->stack) - CHILD_GETA_OF_STACK);
                    free(main_thread_regs);
                    free(pthrs[i]->mutex);
                    free(pthrs[i]);
                }
                mutex_unlock(addsymbol_lock);
                mutex_unlock(memget_lock);
                mutex_unlock(cthread_fork_lock);
                mutex_unlock(print_lock);
                mutex_unlock(read_lock);
                back_trace();
#if defined(__CYGWIN32__) || defined(linux)
                siglongjmp(toplevel, 1);
#else
                longjmp(toplevel, 1);
#endif
            } else {
                pthread_kill(pthrs[i]->thr, SIGINT);
                pause();
            }
#endif
        }
    else {
        int i;
#ifdef CONDITIONAL_FUTURE
        int j;
#endif

        /* detect own processor from primary */
        for (i = 1; (i < processors) && (pthrs[i]->stack != (unsigned int **)stack_top); i++)
            ;
#ifdef CONDITIONAL_FUTURE
        /* detect own processor from secondary */
/*        for (j = 0; (i < processors) && (sthrs[i]->stack != (unsigned int **)stack_top); j++) */
        for (j = 0; (j < processors) && (sthrs[j]->stack != (unsigned int **)stack_top); j++)
            ;
#endif
        /*
         * free own thread
         */
#ifdef CONDITIONAL_FUTURE
        if (i == processors && j == processors) {
            fprintf(stderr, "I'm already killed???\n");
        }
        /* own processor is primary */
        else if (i != processors) {
            pthrs[i]->thr = NULL;
            /* detect other still-alive processors */
            for (i = 1; (i < processors) && (pthrs[i]->thr == NULL); i++)
                ;
            for (j = 0; (j < processors) && (sthrs[j]->thr == NULL); i++)
                ;
            if (i == processors && j == processors) {
                pthread_kill(pthrs[0]->thr, SIGINT);
            } else if (i != processors) {
                pthread_kill(pthrs[i]->thr, SIGINT);
            } else if (j != processors) {
                pthread_kill(sthrs[j]->thr, SIGINT);
            }
            printf("Thread #%x(p) is killed.\n", mtus_thread_self());
            pthread_exit((void *)0);
        }
        /* own processor is secondary */
        else if (j != processors) {
            sthrs[j]->thr = NULL;
            /* detect other still-alive processors */
            for (i = 1; (i < processors) && (pthrs[i]->thr == NULL); i++)
                ;
            for (j = 0; (j < processors) && (sthrs[j]->thr == NULL); i++)
                ;
            if (i == processors && j == processors) {
                pthread_kill(pthrs[0]->thr, SIGINT);
            } else if (i != processors) {
                pthread_kill(pthrs[i]->thr, SIGINT);
            } else if (j != processors) {
                pthread_kill(sthrs[j]->thr, SIGINT);
            }
            printf("Thread #%x(s) is killed.\n", mtus_thread_self());
            pthread_exit((void *)0);
        }
#else 
        if (i == processors)
            fprintf(stderr, "I'm already killed???\n");
        pthrs[i]->thr = NULL;
        /* detect other still-alive processors */
        for (i = 1; (i < processors) && (pthrs[i]->thr == NULL); i++)
            ;
        if (i == processors) {
            pthread_kill(pthrs[0]->thr, SIGINT);
        } else {
            pthread_kill(pthrs[i]->thr, SIGINT);
        }
        printf("Thread #%x is killed.\n", mtus_thread_self());
        pthread_exit((void *)0);
#endif  
    }
        
#else /* not LTC */
# ifdef CTHREAD
    mutex_unlock(addsymbol_lock);
    mutex_unlock(memget_lock);
    mutex_unlock(cthread_fork_lock);
    mutex_unlock(print_lock);
    mutex_unlock(read_lock);
# endif
    back_trace();               /* By N.Watanabe 92/11/29 */
#if defined(__CYGWIN32__) || defined(linux)
    siglongjmp(toplevel, 1);
#else
    longjmp(toplevel, 1);
#endif
#endif /* LTC */
}

void proc_terminate_handler()
{
    void proc_terminate_handler_aux(int);
    proc_terminate_handler_aux(1);
}

void proc_terminate_handler_aux(flag)
int flag;
{
#ifndef WAITPID
    struct rusage rus;
#endif
    static int in_handler = 0;

#ifndef SUNSOL2
    signal(SIGCHLD, proc_terminate_handler);
#endif
    if (flag && in_handler++) return;

    if (sigchld_enable)
        do {
            int status;

#ifdef WAITPID
            while (waitpid(-1, &status, WNOHANG) > 0)
#else
            while (wait3(&status, WNOHANG, &rus) > 0)
#endif
                ;
        } while (--in_handler > 0);
}

#ifdef PVM3      /* By A243 */
int kill_pvm_family_process()
{
    struct pvmtaskinfo *taskp;
    int i, ntask, info, mytid, *kill_process, kill_num = 0, inum, bytes;
    int err_tid, bufid;
    char *sp, *host;

    mytid = pvm_mytid();        
    info = pvm_tasks(0, &ntask, &taskp);

    if(info >= 0){
        kill_process = (int *)malloc(sizeof(int) * ntask);
        pvm_gsize("pvm_error_group");
        inum = pvm_joingroup("pvm_error_group");
        if(inum != 0){
            inum = pvm_gettid("pvm_error_group", 0);
            bufid = pvm_mkbuf(0);
            pvm_setrbuf(bufid);
            pvm_initsend(0);
            pvm_pkint(&mytid, 1, 1);
            pvm_send(inum, PVM_ERROR_TAG);
            pvm_recv(inum, PVM_ERROR_TAG);
            pvm_bufinfo(pvm_getrbuf(), &bytes, (int *)0, (int *)0);
            host = (char *)malloc(sizeof(char) * bytes);
            sp = (char *)malloc(sizeof(char) * bytes);
            pvm_upkint(&err_tid, 1, 1);
            pvm_upkstr(host);
            pvm_upkstr(sp);
            printf("\nSignaled by process #x%x on host %s.\n\n%s\n", err_tid, host, sp);
            pvm_freebuf(bufid);
            free(sp);
        }
        pvm_lvgroup("pvm_error_group");
        serch_kill_process(taskp, ntask, kill_process, &kill_num,
                           &mytid, 1);
        if(kill_num != 0){
            for(i = 0; i < kill_num; i++){
                pvm_kill(*(kill_process + i));
                printf("#x%x ",*(kill_process + i));
            }
            if(kill_num == 1)
                printf("is killed.\n");
            else
                printf("are killed.\n");
        } 
        free(kill_process);
        pvm_delete("pvmgs", 0);
        pvm_kill(gstid);
        gstid = -1;
    }
}

int *serch_kill_process(taskp, ntask, kill_process, kill_num, 
                        tmp_process, tmp_num)
struct pvmtaskinfo *taskp;
int ntask, *kill_process, *kill_num, *tmp_process, tmp_num;
{
    int i, j, *new_process, new_num = 0;
    new_process = (int *)malloc(sizeof(int) * ntask);

    for(i = 0; i < tmp_num; i++){
        for(j = 0; j < ntask; j++){
            if(((*(taskp + j)).ti_ptid == *(tmp_process + i)) &&
                ((strcmp((*(taskp + j)).ti_a_out , "pvmgs")) != 0)){ 
                *(kill_process + *kill_num) = (*(taskp + j)).ti_tid;
                (*kill_num)++;
                *(new_process + new_num) = (*(taskp + j)).ti_tid;
                new_num++;
            }
        }
    }

    if(new_num != 0)
        serch_kill_process(taskp, ntask, kill_process, kill_num,
                           new_process, new_num);
    free(new_process);    

}               
#endif

#ifdef TCPIP
void
# ifdef SIGACTION
timeout_err()
# else
timeout_err (sig, code, scp, addr)
     int sig, code;
     struct sigcontext *scp;
     char *addr;
# endif
{
#if defined(__CYGWIN32__) || defined(linux)
  extern sigjmp_buf timeout_env;
#else
  extern jmp_buf timeout_env;
#endif

#if defined(__CYGWIN32__) || defined(linux)
  siglongjmp (timeout_env, 1);
#else
  longjmp (timeout_env, 1);
#endif
}

void
pipe_broken (sig)
     int sig;
{
  signal (SIGPIPE, pipe_broken);

  /*
   * We can no longer write to a socket (or pipe, FIFO).
   * Because there is no process to read it.
   */
//  bokkan ("Broken Pipe.");
  err_error("Broken Pipe.");
}
#endif


usage_err()
{
    printf("usage: raw_tus [-l<path>][-h<max_heap_size>][-a][-e<exps>]\n");
    printf("            real_max_heap_size = max_heap_size * 64Kbyte\n");
    exit(1);
}


#ifdef CTHREAD
#ifdef USE_REG
   static REGS *main_thread_regs;
#endif
#endif

void make_faslib_path(char *dest, char *src)
{
    strcpy(dest, faspath);
    strcat(dest, src);
}

init_tutscheme(ac, av)
int ac;
char    *av[];
{
    char    fname[100], *scm_dir;
#ifdef CTHREAD
    REGS *regsp;
#endif
#ifndef SAVE_MEM
    FILE *img_fp;
#endif

    fname[0] = '\0';
#ifdef THINK_C
    {
        Size s, t;
        t = MaxMem(&s);
        max_entity_area_size = (s + t - (Size)(436 * 1024)) / 8;
        max_entity_area_size *= 8;
        entity_area_size = max_entity_area_size;
    }
#else
    max_entity_area_size = entity_area_size = DEFAULT_MAX_HEAP_SIZE;
#endif

    system_function_loading_flag = ON;
    fas_lib_loading_mess_flag = ON;

    for (ac--; ac > 0; ac--)
        if (av[ac][0] == '-' && av[ac][1] == 'l') {
            strcpy(fname, av[ac]+2);
            continue;
        } else if (av[ac][0] == '-' && av[ac][1] == 'h') {
            max_entity_area_size
                = entity_area_size
                = strtonum(av[ac]+2) * HEAP_SIZE_UNIT;
            if (max_entity_area_size == 0)
                usage_err();
            continue;
        } else if (av[ac][0] == '-' && av[ac][1] == 'L') {
            strcpy(faspath, av[ac]+2);
            fas_lib_loading_mess_flag = OFF;
        }

#ifdef USE_REG
    main_thread_regs = (REGS *)malloc(sizeof(REGS));
#else /* not USE_REG */
# ifdef CTHREAD
    regsp = (REGS *)malloc(sizeof(REGS));
#  ifdef PTHREAD
    pthread_key_create(&pthread_key, NULL);
#  endif
    cthread_set_data(cthread_self(), (any_t)regsp);
# endif
#endif /* USE_REG */

    meminit();
#ifdef SHE
    printf("1:thr_data(0) = %x\n", thr_data);
#endif

#ifdef SHE
    /*    init_sst_frame_free_list(); */
    init_she_scheduler(0);
    print_stat(pthrs, 0);
#endif

#ifdef THINK_C
/*  if (!doFinderInfo()) { *//*
  if ((img_fp = fopen("TUS.img", "rb")) != NULL)
          load_img(img_fp);
  else {
*/
#endif
#ifdef SHE
    printf("2:thr_data(0) = %x\n", thr_data);
#endif
    sfuninit();
    sfuninit2();
    define_special_object();
    define_symbol();
#ifdef SAVE_MEM
    printmemmap();
#endif

#if (!defined(SAVE_MEM) && !defined(THINK_C))
    sprintf(fname, "%ssaved_image", fname);
    if ((img_fp = fopen(fname, "rb")) == NULL || !load_img(img_fp)) {
#endif /* } */
#if (!defined(SAVE_MEM) && defined(THINK_C))
    if ((img_fp = fopen("saved_image", "rb")) == NULL || !load_img(img_fp)) {
#endif

    /* system program load */

        if (fas_lib_loading_mess_flag == ON)
            print_con("loading macro library: ");
        make_faslib_path(fname, "/maclib.fas");
        system_program_loader(fname, strlen(fname));
        if (fas_lib_loading_mess_flag == ON)
            print_con("\n");

        if (fas_lib_loading_mess_flag == ON)
            print_con("loading scheme library: ");
        make_faslib_path(fname, "/sclib.fas");
        system_program_loader(fname, strlen(fname));
        if (fas_lib_loading_mess_flag == ON)
            print_con("\n");

#ifdef PVM3
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading top level function(pvm-version): ");

        make_faslib_path(fname, "/topfun_pvm.fas");
#else
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading top level function: ");
        make_faslib_path(fname, "/topfun.fas");
#endif
        system_program_loader(fname, strlen(fname));
        if (fas_lib_loading_mess_flag == ON)
            print_con("\n");

        if (fas_lib_loading_mess_flag == ON)
            print_con("loading debug function: ");
        make_faslib_path(fname, "/debugfun.fas");
        system_program_loader(fname, strlen(fname));
        if (fas_lib_loading_mess_flag == ON)
            print_con("\n");

#ifdef CTHREAD
# ifdef SHE
#  ifdef TOUCH_ARG
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading mcomp-she-nt: ");
        make_faslib_path(fname, "/mcomp-she-nt.fas");
#  else
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading mcomp-she-t: ");
        make_faslib_path(fname, "/mcomp-she-t.fas");
#  endif
# elif defined(LTC)
#  ifdef TOUCH_ARG
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading mcomp-ltc-nt: ");
        make_faslib_path(fname, "/mcomp-ltc-nt.fas");
#  else
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading mcomp-ltc-t: ");
        make_faslib_path(fname, "/mcomp-ltc-t.fas");
#  endif
# else
#  ifdef TOUCH_ARG
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading mcomp-nt: ");
        make_faslib_path(fname, "/mcomp-nt.fas");
#  else
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading mcomp-t: ");
        make_faslib_path(fname, "/mcomp-t.fas");
#  endif
# endif
        system_program_loader(fname, strlen(fname));
    print_con("\n");
#else  /* CTHREAD */
# ifdef PVM3
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading compiler(pvm-version): ");
        make_faslib_path(fname, "/compiler_pvm.fas");
# else
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading compiler: ");
        make_faslib_path(fname, "/compiler.fas");
# endif
        system_program_loader(fname, strlen(fname));
        if (fas_lib_loading_mess_flag == ON)
            print_con("\n");
#endif
        if (fas_lib_loading_mess_flag == ON)
            print_con("loading re-compiler: ");    /* By A243  Dec, 1992 */
        make_faslib_path(fname, "/re-compiler.fas");
        system_program_loader(fname, strlen(fname));
        if (fas_lib_loading_mess_flag == ON)
            print_con("\n");

#ifndef SAVE_MEM
    }
#endif

    system_function_loading_flag = OFF;

    /* clear the symbol which has the function written by scheme. */
    /* clear_ordinary_symbol(); */
#ifdef THINK_C
/*  } */
#endif
}

/*
** for Lazy Task Creation
**   06 Aug 1997
*/
#ifdef LTC
extern void primary_steal();
#endif

#ifdef TK
# ifdef __CYGWIN32__
int Tk_AppInit(Tcl_Interp *intr)
{

    if (Tcl_Init(intr) == TCL_ERROR)
        return TCL_ERROR;
    if (Tk_Init(intr) == TCL_ERROR)
        return TCL_ERROR;
    interp = intr;
    longjmp(jb, 1);
}
# endif
#endif

main(argc, argv)
int argc;
char    *argv[];
{
    int i, obj, from_outside;
#ifdef THINK_C
    KeyMap mykey;
#endif

#ifdef TCPIP
# ifdef SIGACTION
    struct sigaction act;
# else
    struct sigvec svec;
# endif
#endif

#ifdef CONDITIONAL_FUTURE
    int ppriority, spriority;
#endif

    strcpy(faspath, "../../lsp");
#ifdef CTHREAD
    cthread_init();
    current_thread_no = 1;
    main_cthread_id = cthread_self();
/*    cthread_set_limit(4); */

# ifdef SUNSOL2
#  if defined(THREAD) && !defined(LTC)
    thr_setconcurrency(sysconf(_SC_NPROCESSORS_ONLN));
#  endif
# endif

#ifdef PTHREAD
    pthread_attr_init(&pattr);
#  ifdef CONDITIONAL_FUTURE
    pthread_attr_init(&sattr);
#endif
    
# if defined(sparc) && defined(USE_REG)
    pthread_attr_setscope(&pattr, PTHREAD_SCOPE_SYSTEM);
#  ifdef CONDITIONAL_FUTURE
    pthread_attr_setscope(&sattr, PTHREAD_SCOPE_SYSTEM);
#  endif
# endif

# ifdef CONDITIONAL_FUTURE
    pthread_attr_getschedparam(&pattr, &pschedparam);
    pthread_attr_getschedparam(&sattr, &sschedparam);
    pschedparam.sched_priority += 10;
/*    pschedparam.sched_priority += 0; */
    sschedparam.sched_priority += 0;
    if (pthread_attr_setschedparam(&pattr, &pschedparam))
        fprintf(stderr, "priority cannot be changed.\n");
    if (pthread_attr_setschedparam(&sattr, &sschedparam))
        fprintf(stderr, "priority cannot be changed.\n");
# endif
#endif

    addsymbol_lock = mutex_alloc();
    memget_lock = mutex_alloc();
    cthread_fork_lock = mutex_alloc();
    print_lock = mutex_alloc();
    read_lock = mutex_alloc();
#endif /* CTHREAD */

#ifdef TK
    Tk_initialized = 0;
#endif

/*
    setbuf(stdin, stdin_buf);
    setbuf(stdout, stdout_buf);
    setlinebuf(stdout);
*/
    symbol_sensitive = 0;

    /*  freeze file ? */
    if (freeze_flag == 1) {
#ifndef THINK_C
#ifdef CTHREAD
        cthread_set_data(cthread_self(), (any_t)main_thread_regs);
#endif
        meminit_for_freeze();
        goto freeze_start;
#endif
    } else
        freeze_flag = 1;

#ifdef X
    openwin();
#endif

#ifndef SAVE_MEM
    sprintf(tus_self, "%s", argv[0]);
#else
    sprintf(tus_self, "%s/%s", SCM_DIR, argv[0]);
#endif

    init_tutscheme(argc, argv);

#ifdef TK
# ifdef __CYGWIN32__
    if (setjmp(jb) == 0)
        Tk_Main(argc, argv, Tk_AppInit);
    {
        pthread_t pth;
        pthread_create(&pth, NULL, (void *)Tk_MainLoop, NULL);
        pthread_detach(pth);
    }
#else
    interp = Tcl_CreateInterp();
//    TkCreateMainWindow(interp, NULL, "raw_dtus", "Tk");
 
    if (Tcl_Init(interp) == TCL_ERROR) {
        fprintf(stderr, "Tcl_Init failed: %s\n", interp->result);
    }
    if (Tk_Init(interp) == TCL_ERROR) {
        fprintf(stderr, "Tk_Init failed: %s\n", interp->result);
    }
    Tk_MainWindow(interp);
#endif
    Tk_initialized = 1;
    Tcl_CreateCommand(interp, "apply_sfun", apply_fun,
                      (ClientData*)NULL,
                      (Tcl_CmdDeleteProc *)NULL);
    Tcl_CreateCommand(interp, "do_one_event", do_one_event,
                      (ClientData*)NULL,
                      (Tcl_CmdDeleteProc *)NULL);
/*
    Tcl_CreateChannelHandler(Tcl_GetStdChannel(TCL_STDIN), TK_READABLE, 
                             StdinProc, (ClientData)0);
*/
#endif

    /* main loop */
 freeze_start:
    for (i = argc-1; i > 0; i--) /* T.Komiya */
        if (argv[i][0] == '-' && argv[i][1] == 'a') {
            set_sval(recomlevel, SCOBJ_2);
            continue;
        } else if (argv[i][0] == '-' && argv[i][1] == 'e') {
            c_in_port.obj = mk_strport(STRING_INPUT,
                                       mk_str(ORDINARY_STRING,
                                              A_STRING,
                                              strlen(argv[i]+2),
                                              (char *)argv[i]+2));
            sctopsym = define_a_ordinary_symbol("eosctop");
            continue;
        }
    if (c_in_port.obj == stdin_port.obj) /* T.Komiya */
        print_welcomsg();
#ifndef THINK_C
#ifdef HP
    setbuf(stdout, stdout_buf);
#else
    setlinebuf(stdout);
#endif /* HP */
#endif
#if defined(__CYGWIN32__) || defined(linux)
    if ((from_outside = sigsetjmp(toplevel, 1)) != 0) {
#else
    if ((from_outside = setjmp(toplevel)) != 0) {
#endif
#ifdef linux                    /* bug of linux ?? by T.Komiya */
/*
        char tmpbuf[256];
        gets(tmpbuf);
*/
#endif
        set_a2(LTAG_NIL);
        c_in_port.obj  = stdin_port.obj;
        c_out_port.obj = stdout_port.obj;
#ifdef THINK_C
        GetKeys(mykey);
        while ((mykey[1] & 0x8000) && (mykey[1] & 0x800000))
            GetKeys(mykey);
        FlushEvents(keyDownMask, 0);
#endif
/*  By N.Watanabe  *********************************
        obj = get_sval(debugsym);
        if (obj == SCOBJ_0) {
        } else if (obj == SCOBJ_1) {
            back_trace();
        } else if (obj == SCOBJ_2) {
            print_con("Sorry, debugger is not implemented yet.");
        } else {
            print_con("*debug-level* has the invalid value(");
            printobj_con(obj, PRINT, 5);
            print_con(").");
        }
********************************************/
    } 
#ifdef CALLIOC
     else {
         ((unsigned *)stack_bottom)[0] = LTAG_NIL;
         ((unsigned *)stack_bottom)[2] = 0;
         ((unsigned *)stack_bottom)[3] = NULL;
         ((unsigned *)stack_bottom)[4] = NULL;
         active_stacks = (unsigned *)stack_bottom;
         free_stacks = NULL;
    }
#endif
#ifdef CALLIOC
    ((unsigned *)stack_bottom)[5] = 0;
#endif
    set_top(stack_bottom);
/*
** for Lazy Task Creation
**   06 Aug 1997
*/
#ifdef LTC
    s_push(primary_steal);
#else
    s_push(top);
#endif
    s_push(0);
    s_push(0);
    set_base(top_val());
    set_active(top_val());
    set_a2(LTAG_NIL);
    set_reg(LTAG_NIL);
    more.obj = LTAG_NIL;
#ifdef CTHREAD
    optcnt.obj = LTAG_NIL;
#endif

/*
** for Lazy Task Creation
**   06 Aug 1997
*/
#ifdef LTC
    ((unsigned int **)stack_top)[STACK_LOCK] = mutex_alloc();
    ((unsigned int **)stack_top)[STACK_BASE] = (unsigned int)sp_reg;
    ((unsigned int **)stack_top)[FUTURE_PH] = (unsigned)LTAG_NIL;
    ((unsigned int **)stack_top)[STACK_INFO_END] = 0;
    ((unsigned int **)stack_top)[LTQ_HEAD]
         = ((unsigned int **)stack_top)[LTQ_TAIL]
         = ((unsigned int **)stack_top + (STACK_INFO_END + 1));
#endif

    /*  By T.Komiya*/
    old_top = (unsigned *) NULL;
    old_top1 = old_top;
    bcraw.obj = LTAG_NIL;
    bcraw2.obj = LTAG_NIL;
    bcbody.obj = LTAG_NIL;
    tempp.obj = LTAG_NIL;
    cdrbody.obj = LTAG_NIL;
    acode.obj = LTAG_NIL;
    bcf.obj = LTAG_NIL;
    bcconst.obj = LTAG_NIL;
    xfunconst.obj = LTAG_NIL;
    bcfhead.obj = LTAG_NIL;
    reg1 = LTAG_NIL;
    reg2 = LTAG_NIL;
    reg3 = LTAG_NIL;
    call_cc_arg.obj = LTAG_NIL;
    top_of_aflk.obj = LTAG_NIL;
    end_of_aflk.obj = LTAG_NIL;
    af_tmp.obj = LTAG_NIL;
    sharp_eq_context_max = 0;
    /***************/

    s_push(get_sval(sctopsym));
    if (c_in_port.obj != stdin_port.obj) { /* T.Komiya */
        s_push(c_in_port.obj);
        c_in_port.obj = stdin_port.obj;
    }
#ifdef TCPIP
# ifdef SIGACTION
    act.sa_handler = (void (*)())timeout_err;
    sigemptyset(&(act.sa_mask));
    sigaddset(&(act.sa_mask), SIGINT);
    act.sa_flags = 0;
    sigaction(SIGALRM, &act, (struct sigaction *)NULL);
# else
    svec.sv_handler = (void (*) ()) timeout_err;
    svec.sv_mask = SIGINT;
    svec.sv_flags = SV_INTERRUPT;
    sigvec (SIGALRM, &svec, (struct sigvec *) 0);
# endif
    signal (SIGPIPE, pipe_broken);
#endif

    interrupt_enable = TRUE;
#ifdef THINK_C
    signal(SIGINT, (void(*)())interr);
#else
    signal(SIGINT, interr);
    sigchld_enable = TRUE;
    /* Solaris 2 has problem with it */
# ifdef SUNSOL2
{
    struct sigaction sigact;

    sigemptyset(&(sigact.sa_mask));
    sigact.sa_handler = proc_terminate_handler;
    sigact.sa_flags = SA_NOCLDSTOP | SA_RESTART;
    sigaction(SIGCHLD, &sigact, NULL);
}
# else
    signal(SIGCHLD, proc_terminate_handler);
# endif
#endif
#ifdef PVM3
    signal(SIGUSR1, pvm_on_demand);
#endif
#ifdef HAVE_LIBREADLINE
    using_history();
    stifle_history(30);
#endif
    if (!from_outside) {
        CARCDR *x, *y;
        unsigned s;

        for (i = argc-1; i >= 0; i--)
            s_push(mk_str(ORDINARY_STRING, A_STRING, strlen(argv[i]), argv[i]));
        x = y = (CARCDR *)mk_nconsp(argc);
        for (i = argc; i > 0; i--)
            y++->car.obj = s_pop();
        s_push(LTAG_CONS | SSERDDA(x));
        s = define_a_ordinary_symbol("*argv*");
        set_sval(s, s_pop());
    }
    obj = bc_interpreter();
}

void system_program_loader(filename, namelength)
char    *filename;
int namelength;
{
    unsigned tmp;

    set_top(stack_bottom);
    s_push(top);
    s_push(0);
    s_push(0);
    set_base(top_val());
    set_active(top_val());
    set_a2(LTAG_NIL);
    set_reg(LTAG_NIL);
    s_push(get_sval(bcloadsym));
    tmp = mk_str(ORDINARY_STRING, A_STRING, namelength, filename);
    s_push(tmp);
    /* restore_c_stack(); */
    xcall();
    bc_interpreter();
    set_a2(LTAG_NIL);
}


clear_ordinary_symbol()
{
    int i;
    unsigned *hashlink;
    
    for (i = 0; i < hash_table_size; i++) {
        hashlink = hash_table + i;
        while (*hashlink != LTAG_NIL) {
            if (GETtag(*hashlink) == LTAG_SPECIAL)
                hashlink = ((unsigned *) ADDRESS(*hashlink)) + 3;
            else {
                if (protect_symbol_p(*hashlink))
                    hashlink = ((unsigned *) ADDRESS(*hashlink)) + 3;
                else
                    *hashlink = *(((unsigned *) ADDRESS(*hashlink)) + 3);
            }
        }
    }
}


protect_symbol_p(sym)
int sym;
{
    unsigned *pointer;
    pointer = (unsigned *)get_sval(protect_symbol_listsym);
    while (pointer != (unsigned *)LTAG_NIL) {
        if (CARobj(pointer) == sym)
            return (TRUE);
        pointer = (unsigned *)CDRobj(pointer);
    }
    return (FALSE);
}


/* global break function */

bokkan(errmessage)
char    *errmessage;
{
#ifdef PVM3
    SOBJ pvm_string_port;
    unsigned new_string;
    PORT * portobj;
    char *strobj_to_str();
#endif

    error_prologue();
    c_in_port.obj  = stdin_port.obj;
    c_out_port.obj = stdout_port.obj;
    print_con("Error: ");
    print_con(errmessage);
    if (get_sval(user_exec_sym) == LTAG_T) {
        print_con("\n       in ");
        print_last_funcall(-1);
    }

#ifdef PVM3
    new_string = SSERDDA(Pgetblock2(128)) | LTAG_STRING;
    pvm_string_port.obj = mk_strport(STRING_OUTPUT, new_string);
    c_out_port.obj = pvm_string_port.obj;

    printf2("Error: ");
    printf2(errmessage);
    if (get_sval(user_exec_sym) == LTAG_T) {
        printf2("\n       in ");
        print_last_funcall2(-1);
    }
    printf2("\n");
    portobj = (PORT *) ADDRESS(pvm_string_port.obj);
    c_out_port.obj = stdout_port.obj;

    sigint_root_process(strobj_to_str(mk_string(ORDINARY_STRING,
                           (int)portobj->fp - 4,
                           (char *)ADDRESS(portobj->slot.obj),
                                  4)));
#endif

    error_epilogue();
}

#ifdef PVM3                          /* By A243 */
int sigint_root_process(errmessage)
char *errmessage;
{
    struct pvmtaskinfo *taskp;
    int i, ntask, info, mytid, root_tid, inum, bufid;

    mytid = pvm_mytid();
    info = pvm_tasks(0, &ntask, &taskp);
    root_tid = mytid;
 
    if(info >= 0){
        if(pvm_parent() != -23){
            pvm_gsize("pvm_error_group");
            pvm_joingroup("pvm_error_group");
            bufid = pvm_mkbuf(0);
            pvm_setrbuf(bufid);
            serch_root_process(taskp, ntask, &root_tid);
            pvm_sendsig(root_tid, SIGINT);
            pvm_recv(root_tid, PVM_ERROR_TAG);
            pvm_upkint(&inum, 1, 1);
            pvm_freebuf(bufid);
            pvm_initsend(0);
            pvm_pkint(&mytid, 1, 1);
            pvm_pkstr(ourhostname);
            pvm_pkstr(errmessage);
            pvm_send(root_tid, PVM_ERROR_TAG);
            pvm_lvgroup("pvm_error_group");
            while(1);
        }
        else    
            kill_pvm_family_process();
    }
}

int serch_root_process(taskp, ntask, num)
struct pvmtaskinfo *taskp;
int ntask, *num;
{
    int i, new = 0;

    for(i = 0; i < ntask; i++)
        if((*(taskp + i)).ti_tid == *num)
            new = (*(taskp + i)).ti_ptid;

    if(new != 0){
        *num = new;
        serch_root_process(taskp, ntask, num);
    }
}
#endif

sbokkan1(funname, arg, errmessage)
char    *funname;
int arg;
char    *errmessage;
{
#ifdef PVM3
    SOBJ pvm_string_port;
    unsigned new_string;
    PORT * portobj;
    char *strobj_to_str();
#endif

    error_prologue();
    c_in_port.obj = stdin_port.obj;
    c_out_port.obj = stdout_port.obj;

    print_con("Error: ");
    print_con(errmessage);
    print_con("\n     in (");
    print_con(funname);
    print_con(" ");
    printobj_con(arg, PRINT, 100);
    print_con(")");

#ifdef PVM3
    new_string = SSERDDA(Pgetblock2(128)) | LTAG_STRING;
    pvm_string_port.obj = mk_strport(STRING_OUTPUT, new_string);
    c_out_port.obj = pvm_string_port.obj;

    printf2("Error: ");
    printf2(errmessage);
    printf2("\n     in (");
    printf2(funname);
    printf2(" ");
    printobj(arg, PRINT, 100);
    printf2(")\n");
    portobj = (PORT *) ADDRESS(pvm_string_port.obj);
    c_out_port.obj = stdout_port.obj;

    sigint_root_process(strobj_to_str(mk_string(ORDINARY_STRING,
                           (int)portobj->fp - 4,
                           (char *)ADDRESS(portobj->slot.obj),
                                  4)));
#endif
    error_epilogue();
}

sbokkan2(funname, arg1, arg2, errmessage) /* T.Komiya */
char    *funname;
int arg1, arg2;
char    *errmessage;
{
#ifdef PVM3
    SOBJ pvm_string_port;
    unsigned new_string;
    PORT * portobj;
    char *strobj_to_str();
#endif

    error_prologue();
    c_in_port.obj = stdin_port.obj;
    c_out_port.obj = stdout_port.obj;

    print_con("Error: ");
    print_con(errmessage);
    print_con("\n     in (");
    print_con(funname);
    print_con(" ");
    printobj_con(arg1, PRINT, 100);
    print_con(" ");
    printobj_con(arg2, PRINT, 100);
    print_con(")");

#ifdef PVM3
    new_string = SSERDDA(Pgetblock2(128)) | LTAG_STRING;
    pvm_string_port.obj = mk_strport(STRING_OUTPUT, new_string);
    c_out_port.obj = pvm_string_port.obj;

    printf2("Error: ");
    printf2(errmessage);
    printf2("\n     in (");
    printf2(funname);
    printf2(" ");
    printobj(arg1, PRINT, 100);
    printf2(" ");
    printobj(arg2, PRINT, 100);
    printf2(")\n");
    portobj = (PORT *) ADDRESS(pvm_string_port.obj);
    c_out_port.obj = stdout_port.obj;

    sigint_root_process(strobj_to_str(mk_string(ORDINARY_STRING,
                           (int)portobj->fp - 4,
                           (char *)ADDRESS(portobj->slot.obj),
                                  4)));
#endif
    error_epilogue();
}

sbokkan3(funname, arg1, arg2, arg3, errmessage) /* T.Komiya */
char    *funname;
int arg1, arg2, arg3;
char    *errmessage;
{
#ifdef PVM3
    SOBJ pvm_string_port;
    unsigned new_string;
    PORT * portobj;
    char *strobj_to_str();
#endif

    error_prologue();
    c_in_port.obj = stdin_port.obj;
    c_out_port.obj = stdout_port.obj;

    print_con("Error: ");
    print_con(errmessage);
    print_con("\n     in (");
    print_con(funname);
    print_con(" ");
    printobj_con(arg1, PRINT, 100);
    print_con(" ");
    printobj_con(arg2, PRINT, 100);
    print_con(" ");
    printobj_con(arg3, PRINT, 100);
    print_con(")");

#ifdef PVM3
    new_string = SSERDDA(Pgetblock2(128)) | LTAG_STRING;
    pvm_string_port.obj = mk_strport(STRING_OUTPUT, new_string);
    c_out_port.obj = pvm_string_port.obj;

    printf2("Error: ");
    printf2(errmessage);
    printf2("\n     in (");
    printf2(funname);
    printf2(" ");
    printobj(arg1, PRINT, 100);
    printf2(" ");
    printobj(arg2, PRINT, 100);
    printf2(" ");
    printobj(arg3, PRINT, 100);
    printf2(")\n");
    portobj = (PORT *) ADDRESS(pvm_string_port.obj);
    c_out_port.obj = stdout_port.obj;

    sigint_root_process(strobj_to_str(mk_string(ORDINARY_STRING,
                           (int)portobj->fp - 4,
                           (char *)ADDRESS(portobj->slot.obj),
                                  4)));
#endif
    error_epilogue();
}


bokkan_stack_overflow()
{
    int fname;
#ifdef PVM3
    SOBJ pvm_string_port;
    unsigned new_string;
    PORT * portobj;
    char *strobj_to_str();
#endif

    error_prologue();
    c_in_port.obj = stdin_port.obj;
    c_out_port.obj = stdout_port.obj;
    print_con("Error: ");
    print_con("Stack overflow");
    print_con("\n       in ");
/*    print_last_funcall(-1); *//* T.Komiya */

    /* T.Komiya */
    print_con("(");
    if ((fname = cfname()) == 0) {
        /* (  a paren to balance */
        print_con(") empty frame\n");
#if defined(__CYGWIN32__) || defined(linux)
        siglongjmp(toplevel, 1);
#else
        longjmp(toplevel, 1);
#endif
    } else if (fname == LTAG_NIL)
        print_con("<non-named-function>");
    else
        printobj_con(fname, PRINC, -1);
    print_con(" ... )");

#ifdef PVM3
    new_string = SSERDDA(Pgetblock2(128)) | LTAG_STRING;
    pvm_string_port.obj = mk_strport(STRING_OUTPUT, new_string);
    c_out_port.obj = pvm_string_port.obj;

    printf2("Error: ");
    printf2("Stack overflow");
    printf2("\n       in ");

    printf2("(");
    if ((fname = cfname()) == 0) {
        printf2(") empty frame\n");
#if defined(__CYGWIN32__) || defined(linux)
        siglongjmp(toplevel, 1);
#else
        longjmp(toplevel, 1);
#endif
    } else if (fname == LTAG_NIL)
        printf2("<non-named-function>");
    else
        printobj(fname, PRINC, -1);
    printf2(" ... )\n");
    portobj = (PORT *) ADDRESS(pvm_string_port.obj);
    c_out_port.obj = stdout_port.obj;

    sigint_root_process(strobj_to_str(mk_string(ORDINARY_STRING,
                           (int)portobj->fp - 4,
                           (char *)ADDRESS(portobj->slot.obj),
                                  4)));
#endif
    error_epilogue();
}

#ifdef CALLIOC

bokkan_stack_underflow()
{
    error_prologue();
    c_in_port.obj  = stdin_port.obj;
    c_out_port.obj = stdout_port.obj;
    print_con("Error: Invalid invocation of IOC (stack underflow).\n");
    error_epilogue();
}

#endif

Serror_break()
{
    int obj;        /* By N.Watanabe */
    
    if (old_top != old_top1) {
        set_top(old_top);
        old_top1 = old_top;
        c_strbuf = strbuf;
    }
    fflush(stdin);
    fflush(stdout);
    c_in_port.obj = stdin_port.obj;
    c_out_port.obj = stdout_port.obj;
    
    obj = get_sval(debugsym);    /*  By N.Watanabe  */
    if (obj == SCOBJ_0) {
    } else if (obj == SCOBJ_1) {
        back_trace();
    } else if (obj == SCOBJ_2) {
        print_con("Sorry, debugger is not implemented yet.");
    } else {
        print_con("*debug-level* has the invalid value(");
        printobj_con(obj, PRINT, 5);
        print_con(").");
    }
    error_epilogue();
    set_reg(LTAG_NIL);
}

error_prologue()
{
    if (old_top != old_top1) {
        set_top(old_top);
        old_top1 = old_top;
        c_strbuf = strbuf;
    }
    fflush(stdin);
/*                      By N.Watanabe Nov. 28, 1992
    c_in_port.obj = stdin_port.obj;
    c_out_port.obj = stdout_port.obj;
*/
}

error_epilogue()
{
#ifdef LTC
    interr(2);
#else
#if defined(__CYGWIN32__) || defined(linux)
    siglongjmp(toplevel, 1);
#else
    longjmp(toplevel, 1);
#endif
#endif
}


cfname()
{
    if (FUNp(*(active_reg - 1))) {
        return (*(((unsigned *) ADDRESS(*(active_reg - 1))) + 2));
    } else
        return (0);
}


print_last_funcall(argnum)  /* This function depend on
                             * error_initialize(). */
int argnum;         /* argnum >= 0 : argnum is real argument
                     * number. */
/* argnum <  0 : argnum is unknown. */
{
    int fname;
    unsigned *active_p;
    int i;
    print_con("(");
    if ((fname = cfname()) == 0) {
        /* (  a parlen to balance */
        print_con(") empty frame\n");
#if defined(__CYGWIN32__) || defined(linux)
        siglongjmp(toplevel, 1);
#else
        longjmp(toplevel, 1);
#endif
    } else if (fname == LTAG_NIL)
        print_con("<non-named-function>");
    else
        printobj_con(fname, PRINC, -1);
    if (argnum < 0)
/*        argnum = getargnum2(); */
        argnum = getargnum();
    active_p = (unsigned *) active_val();
    for (i = 0; i < argnum; ++i) {
        print_con(" ");
        printobj_con(*(active_p - (i + 2)), PRINT, 5);
    }
    print_con(")");
}

#ifdef PVM3
print_last_funcall2(argnum)  /* This function depend on
                             * error_initialize(). */
int argnum;         /* argnum >= 0 : argnum is real argument
                     * number. */
/* argnum <  0 : argnum is unknown. */
{
    int fname;
    unsigned *active_p;
    int i;
    printf2("(");
    if ((fname = cfname()) == 0) {
        /* (  a parlen to balance */
        printf2(") empty frame\n");
#if defined(__CYGWIN32__) || defined(linux)
        siglongjmp(toplevel, 1);
#else
        longjmp(toplevel, 1);
#endif
    } else if (fname == LTAG_NIL)
        printf2("<non-named-function>");
    else
        printobj(fname, PRINC, -1);
    if (argnum < 0)
/*        argnum = getargnum2(); */
        argnum = getargnum();
    active_p = (unsigned *) active_val();
    for (i = 0; i < argnum; ++i) {
        printf2(" ");
        printobj(*(active_p - (i + 2)), PRINT, 5);
    }
    printf2(")");
}
#endif

Sstack_dump()
{
    int depth;
#ifdef TOUCH_ARG
    depth = touch_obj(getsv(-2));
#else
    depth = getsv(-2);
#endif
    if (GETfix_tag(depth) == LTAG_FIXNUM) {
        stackdump(fix_to_int(depth), LTAG_NIL);
    }
    set_reg(LTAG_T);
}



void
stackdump(depth, option)
int depth;
int option;
{
    unsigned *active;
    unsigned *ractive;
    unsigned *base;
    unsigned *rbase;
    unsigned *top;
    int flag;
    int i;
    
    flag = SPRINT;
    if (option == LTAG_NIL)
        flag = PRINT;
    ractive = (unsigned *) active_val();
    active = ractive;
    rbase = (unsigned *) base_val();
    base = rbase;
    top = (unsigned *) top_val();
    while ((depth-- > 0) && (top < ((unsigned *) stack_bottom))) {
        if (top == base) {
            base = (unsigned *)*top;
            for (i = 0; i < 4; i++) {
                printf("%8x: ", top);
                printf("%8x :*", *top);
                if ((top == rbase) || (top == ractive) || 
                    (top == ((unsigned *) stack_bottom - 3))) {
                    if (top == rbase)
                        printf("  <== base");
                    if (top == ractive)
                        printf("  <== active");
                    if (top == ((unsigned *) stack_bottom - 
                                3))
                        printf("  <== bottom");
                }
                printf("\n");
                top++;
            }
        } else {
            printf("%8x: ", top);
            printf("%8x : \n", *top);
#ifdef PVM3            
            printobj(*top, flag, 10, OFF);
#else
            printobj(*top, flag, 10);
#endif
            printf("\n");
            ++top;
        }
    }
}


stackdump2(depth)
int depth;
{
    unsigned *active;
    unsigned *ractive;
    unsigned *base;
    unsigned *rbase;
    unsigned *top;
    int i;
    
    printf("old_top = %x, old_top1 = %x\n", old_top, old_top1);
    ractive = (unsigned *) active_val();
    active = ractive;
    rbase = (unsigned *) base_val();
    base = rbase;
    top = (unsigned *) top_val();
    while ((depth-- > 0) && (top < ((unsigned *) stack_bottom))) {
        if (top == base) {
            base = (unsigned *)*top;
            for (i = 0; i < 4; i++) {
                printf("%8x: ", top);
                printf("%8x :*", *top);
                if ((top == rbase) || (top == ractive) || 
                    (top == ((unsigned *) stack_bottom - 3))) {
                    if (top == rbase)
                        printf("  <== base");
                    if (top == ractive)
                        printf("  <== active");
                    if (top == ((unsigned *) stack_bottom - 
                                3))
                        printf("  <== bottom");
                }
                printf("\n");
                top++;
            }
        } else {
            printf("%8x: ", top);
            printf("%8x : ", *top);
            printf("\n");
            ++top;
        }
    }
}


printmemmap()
{
    char    s[150];
    
    sprintf(s, "stack_top=%x, stack_bottom=%x, stack_size=%x b, hash_size=%x b\n", (int) stack_top, (int) stack_bottom,
            STACK_SIZE, HASH_SIZE);
    print_con(s);
    sprintf(s, "heap_top= %x, heap_bottom=%x, heap_size=%x b\n",
            (int) entity_base, (int) entity_fin, max_entity_area_size);
    print_con(s);
    sprintf(s, "initial_heap_size = %x\n", (int) (entity_area_size / 2));
    print_con(s);
}

print_welcomsg()
{
    char s[100];

/* #ifdef TCPIP
    sprintf (s, "Distributed TUTScheme (alpha version).\n");
#else */
    sprintf(s, "TUTScheme version %s\n", TUS_VERSION);
/* #endif */
    print_con(s);
    print_con("(C) Copyright Taiichi Yuasa, 1991-2016.  All rights reserved.\n");
    print_con("\nIf you find any bug, please send bug reports to komiya@spa.is.uec.ac.jp.\n");
}

#ifdef THINK_C
doFinderInfo()
{
    short       i, count, fref, msg;
    AppFile     file;
    int c, *bufp, *bufend, *heap;
    char buf[512];
    
    CountAppFiles(&msg, &count);
    if (count > 0) {
        GetAppFiles(i, &file);
        if (FSOpen(file.fName, file.vRefNum, &fref) != noErr) {
            FileError("\pError opening ", file.fName);
            ExitToShell();
        }
        heap = entity_start;
        do {
            bufp = (unsigned *)buf;
            c = 512;
            i = FSRead(fref, &c, &buf);
            bufend = (unsigned *)((char *)bufp + c);
            while (bufp < bufend)
                *heap++ = *bufp++;
        } while (i != eofErr);
        last_entity = heap;
        FSClose(fref);
    }
    return((int)count);
}

FileError(s, f)
Str255 s, f;
{
    ParamText(s, f,"\p", "\p");
    Alert( ErrorAlert, 0L );
}

#endif

#ifndef SAVE_MEM

extern int memget_mode;

load_img(fp)
FILE *fp;
{
    unsigned *old_entity_base, *old_entity_start, *old_entity_end, *old_last_entity;

    fread(&old_entity_start, sizeof(unsigned *), 1, fp);
    fread(&old_entity_end, sizeof(unsigned *), 1, fp);
    fread(&old_last_entity, sizeof(unsigned *), 1, fp);
    fread(&old_entity_base, sizeof(unsigned *), 1, fp);
    if (old_entity_base != entity_base) {
        printf("Can not load saved_image file.\n");
        fclose(fp);
        return(0);
    }
    entity_start = old_entity_start;
    entity_end = old_entity_end;
    last_entity = old_last_entity;
    gctype = fgetc(fp);
    memget_mode = fgetc(fp);
    entity_position = fgetc(fp);
    fread(&quotesym.obj, sizeof(unsigned *), 1, fp);
    fread(&qquotesym.obj, sizeof(unsigned *), 1, fp);
    fread(&atsym.obj, sizeof(unsigned *), 1, fp);
    fread(&commasym.obj, sizeof(unsigned *), 1, fp);
    
    fread(&definesym.obj, sizeof(unsigned *), 1, fp);
    fread(&macrosym.obj, sizeof(unsigned *), 1, fp);
    fread(&xreturnsym.obj, sizeof(unsigned *), 1, fp);
    fread(&lreturnsym.obj, sizeof(unsigned *), 1, fp);
    fread(&xcallsym.obj, sizeof(unsigned *), 1, fp);
    fread(&c_f_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&mkncsym.obj, sizeof(unsigned *), 1, fp);
    fread(&sctopsym, sizeof(unsigned *), 1, fp);
    fread(&errsym, sizeof(unsigned *), 1, fp);     /* By N.Watanabe 92/11/19 */
    fread(&recomsym, sizeof(unsigned *), 1, fp);   /* By A243 */
    fread(&recomlevel, sizeof(unsigned *), 1, fp); /* By A243 */
#ifdef PVM3
    fread(&bctran_exec_form_sym, sizeof(unsigned *), 1, fp); /* By A243 */    
    fread(&sharp_closure_pvm, sizeof(unsigned *), 1, fp); /* By A243 */    
    fread(&pvm_recv_function, sizeof(unsigned *), 1, fp); /* By A243 */
    fread(&pvm_reduce_function, sizeof(unsigned *), 1, fp); /* By A243 */
    fread(&pvm_get_unbound_obj, sizeof(unsigned *), 1, fp); /* By A243 */
    fread(&pvm_put_unbound_obj, sizeof(unsigned *), 1, fp); /* By A243 */
#endif
    fread(&debugsym, sizeof(unsigned *), 1, fp);
    fread(&user_exec_sym, sizeof(unsigned *), 1, fp);
    fread(&sys_debug_sym, sizeof(unsigned *), 1, fp);
    fread(&bcloadsym, sizeof(unsigned *), 1, fp);
    fread(&bc_trace_proc_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&sysfun_tracer_body.obj, sizeof(unsigned *), 1, fp);
    fread(&max_back_trace_depth_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&gc_message_level_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&msgc_to_copygc_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&struct_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&l_t_e_sym.obj, sizeof(unsigned *), 1, fp);
#ifdef PVM3
    fread(&pvm_l_t_e_sym.obj, sizeof(unsigned *), 1, fp);   /* By A243 */
#endif
    fread(&local_fun_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&beginsym.obj, sizeof(unsigned *), 1, fp);
    fread(&bcload_with_cip_sym.obj, sizeof(unsigned *), 1, fp);

    /* 2nd: the other object which the scheme system hold */
    fread(&unbound_obj.obj, sizeof(unsigned *), 1, fp);
    fread(&eof_obj.obj, sizeof(unsigned *), 1, fp);
    fread(&reg1, sizeof(unsigned *), 1, fp);
    fread(&reg2, sizeof(unsigned *), 1, fp);
    fread(&reg3, sizeof(unsigned *), 1, fp);
    fread(&stdin_port.obj, sizeof(unsigned *), 1, fp);
    fread(&stdout_port.obj, sizeof(unsigned *), 1, fp);
    fread(&c_in_port.obj, sizeof(unsigned *), 1, fp);
    fread(&c_out_port.obj, sizeof(unsigned *), 1, fp);
    fread(&bcconst.obj, sizeof(unsigned *), 1, fp);
    fread(&xfunconst.obj, sizeof(unsigned *), 1, fp);
    fread(&bcraw.obj, sizeof(unsigned *), 1, fp);
    fread(&bcraw2.obj, sizeof(unsigned *), 1, fp);
    fread(&bcbody.obj, sizeof(unsigned *), 1, fp);
    fread(&acode.obj, sizeof(unsigned *), 1, fp);
    fread(&tempp.obj, sizeof(unsigned *), 1, fp);
    fread(&cdrbody.obj, sizeof(unsigned *), 1, fp);
    fread(&bcf.obj, sizeof(unsigned *), 1, fp);
    fread(&bcfhead.obj, sizeof(unsigned *), 1, fp);
    fread(&call_cc_arg.obj, sizeof(unsigned *), 1, fp);
    fread(&top_of_aflk.obj, sizeof(unsigned *), 1, fp);
    fread(&end_of_aflk.obj, sizeof(unsigned *), 1, fp);
    fread(&af_tmp.obj, sizeof(unsigned *), 1, fp);
    fread(&more.obj, sizeof(unsigned *), 1, fp);
#ifdef CTHREAD
    fread(&optcnt.obj, sizeof(unsigned *), 1, fp);
#endif
    fread(&portlist.obj, sizeof(unsigned *), 1, fp);
    fread(&ncclist.obj, sizeof(unsigned *), 1, fp);
    fread(&wplist.obj, sizeof(unsigned *), 1, fp);
    
#ifdef TCPIP
    fread(&conn_timeout.obj, sizeof(unsigned *), 1, fp);
    fread(&tcp_timeout.obj, sizeof(unsigned *), 1, fp);
    fread(&udp_timeout.obj, sizeof(unsigned *), 1, fp);
    fread(&accept_timeout.obj, sizeof(unsigned *), 1, fp);
    fread(&localhostname.obj, sizeof(unsigned *), 1, fp);
#endif
    
#ifdef NIHONGO
    fread(&default_kanji_input_code_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&default_kanji_output_code_sym.obj, sizeof(unsigned *), 1, fp);
#ifdef TCPIP
    fread(&default_network_kanji_input_code_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&default_network_kanji_output_code_sym.obj, sizeof(unsigned *), 1, fp);
#endif
    fread(&euc_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&jis_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&sjis_sym.obj, sizeof(unsigned *), 1, fp);
    fread(&noc_sym.obj, sizeof(unsigned *), 1, fp);
#endif
    fread(&print_circle_sym.obj, sizeof(unsigned *), 1, fp);
    fread(hash_table, hash_table_size, 4, fp);
    fread(entity_start, last_entity - entity_start, 4, fp);
/*
    relocate_img(hash_table,
                 hash_table + hash_table_size,
                 (unsigned)hash_table - (unsigned)old_hash_table);
     relocate_img(entity_base,
                  entity_end,
                  (unsigned)entity_base - (unsigned)old_entity_base);
*/  
    fclose(fp);
    meminit_for_freeze();
    return(1);
}
/*
relocate_img(start, end, dist)
unsigned *start, *end;
int dist;
{
    while (start < end) {
        switch (GETtag(*start)) {
        case LTAG_BLOCK2 :
            start += (IMM(*start) + 1);
            break;
        case LTAG_BLOCK1 :
            if (0xe0000000 & *start && !(0x80000000 & *start))
  
  
    add_dis_pobj(vacancy_size);
}
*/

#endif

search_heap(x)
unsigned x;
{
    unsigned *h;

    for (h = entity_start; h < last_entity; h++)
        if (ADDRESS(*h) == x)
            printf("%x: %x\n", h, x);
    return(0);
}

search_stack(x)
unsigned x;
{
    unsigned *s;

    for (s = (unsigned *)stack_top; s < (unsigned *)stack_bottom; s++)
        if (ADDRESS(*s) == x)
            printf("%x: %x\n", s, x);
    return(0);
}
